@@ -35,7 +35,7 @@
 /* RCS info. */
 
 #ifndef lint
-static char *RcsId = "$Id: regex.c,v 2.0 1997/09/06 11:56:40 malc Rel $";
+static char *RcsId = "$Id: regex.c,v 2.1 1997/09/07 00:35:35 malc Exp $";
 static char *RegexId = REGEXID;
 #endif /* ! lint */
 
@@ -123,8 +123,8 @@ static regex_t *regex = NULL;
 /****************************************************************************/
 /* Variables used while compiling a pattern */
 
-static __regatom_t *atoms = NULL;
-static __regatom_t *last = NULL;
+static __regatom_t *catoms = NULL;
+static __regatom_t *clast = NULL;
 static int no_subexpressions = 0;
 static int no_repeats = 0;
 static int reg_error = 0;
@@ -158,7 +158,7 @@ int flags;
 
 	/* Now initialise the static values */
 
-	atoms = last = NULL;
+	catoms = clast = NULL;
 	no_subexpressions = 0;
 	no_repeats = 0;
 
@@ -173,7 +173,7 @@ int flags;
 
 	/* Set the expression's atom list and item counts */
 
-	expr->re_atoms = atoms;
+	expr->re_atoms = catoms;
 	expr->re_nsub = no_subexpressions;
 	expr->re_nrpt = no_repeats;
 
@@ -339,18 +339,18 @@ int flags;
 		/* Generate a class containing the character */
 
 		(void) sprintf(class, "%c%c\n", *pos, toupper(*pos));
-		atoms = add_atom(atoms, match_class, class, 2, 1);
+		catoms = add_atom(catoms, match_class, class, 2, 1);
 	} else if (flags & REG_ICASE && isascii(*pos) && isupper(*pos)) {
 		/* Generate a class containing the character */
 
 		(void) sprintf(class, "%c%c\n", tolower(*pos), *pos);
-		atoms = add_atom(atoms, match_class, class, 2, 1);
+		catoms = add_atom(catoms, match_class, class, 2, 1);
 	} else {
 		/* Simply add the character to be matched */
 
-		atoms = add_atom(atoms, match_char, pos, 1, 1);
+		catoms = add_atom(catoms, match_char, pos, 1, 1);
 	}
-	return((atoms != NULL) ? pos + 1 : NULL);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 static char *compile_any(pos, flags)
@@ -359,9 +359,9 @@ int flags;
 {
 	/* Compile "match any character" into the expression */
 
-	atoms = add_atom(atoms, !(flags & REG_NEWLINE) ? match_any :
-			 match_any_newline, pos, 1, 1);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, !(flags & REG_NEWLINE) ? match_any
+			  : match_any_newline, pos, 1, 1);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -468,9 +468,9 @@ int flags;
 
 	/* And add the class to the expression */
 
-	atoms = add_atom(atoms, (negated) ? match_neg_class :
-			 match_class, cstring, strlen(cstring), 1);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, (negated) ? match_neg_class :
+			  match_class, cstring, strlen(cstring), 1);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -483,9 +483,9 @@ int flags;
 	 * This will need to be rewritten to handle locales.
 	 */
 
-	atoms = add_atom(atoms, match_class, WORD_CHARS,
-			 strlen(WORD_CHARS), 1);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, match_class, WORD_CHARS,
+			  strlen(WORD_CHARS), 1);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 static char *compile_non_word_class(pos, flags)
@@ -498,10 +498,10 @@ int flags;
 	 * locales properly.
 	 */
 
-	atoms = add_atom(atoms, !(flags & REG_NEWLINE) ? match_neg_class :
-			 match_neg_class_newline, WORD_CHARS,
-			 strlen(WORD_CHARS), 1);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, !(flags & REG_NEWLINE) ? match_neg_class
+			  : match_neg_class_newline, WORD_CHARS,
+			  strlen(WORD_CHARS), 1);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 static char *compile_beginning(pos, flags)
@@ -510,8 +510,8 @@ int flags;
 {
 	/* Compile beginning of line into the expression */
 
-	if (!(flags & REG_EXTENDED) && last != NULL
-	    && last->ra_matchfunc != match_start_subexpr) {
+	if (!(flags & REG_EXTENDED) && clast != NULL
+	    && clast->ra_matchfunc != match_start_subexpr) {
 		/* Treat this character as a literal match */
 
 		return(compile_char(pos, flags));
@@ -519,9 +519,9 @@ int flags;
 
 	/* Actually match the beginning of the line */
 
-	atoms = add_atom(atoms, !(flags & REG_NEWLINE) ? match_beginning :
-			 match_beginning_newline, NULL, 0, 0);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, !(flags & REG_NEWLINE) ? match_beginning
+			  : match_beginning_newline, NULL, 0, 0);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 static char *compile_end(pos, flags)
@@ -539,9 +539,9 @@ int flags;
 
 	/* Actually match the end of the line */
 
-	atoms = add_atom(atoms, !(flags & REG_NEWLINE) ? match_end :
-			 match_end_newline, NULL, 0, 0);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, !(flags & REG_NEWLINE) ? match_end
+			  : match_end_newline, NULL, 0, 0);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -551,8 +551,8 @@ int flags;
 {
 	/* Compile start of word into the expression */
 
-	atoms = add_atom(atoms, match_word_start, NULL, 0, 0);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, match_word_start, NULL, 0, 0);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -562,8 +562,8 @@ int flags;
 {
 	/* Compile end of word into the expression */
 
-	atoms = add_atom(atoms, match_word_end, NULL, 0, 0);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, match_word_end, NULL, 0, 0);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -573,8 +573,8 @@ int flags;
 {
 	/* Compile word boundary into the expression */
 
-	atoms = add_atom(atoms, match_word_boundary, NULL, 0, 0);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, match_word_boundary, NULL, 0, 0);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -584,8 +584,8 @@ int flags;
 {
 	/* Compile not word boundary into the expression */
 
-	atoms = add_atom(atoms, match_non_word_boundary, NULL, 0, 0);
-	return((atoms != NULL) ? pos + 1 : NULL);
+	catoms = add_atom(catoms, match_non_word_boundary, NULL, 0, 0);
+	return((catoms != NULL) ? pos + 1 : NULL);
 }
 /****************************************************************************/
 static char *compile_subexpr(pos, flags)
@@ -603,8 +603,8 @@ int flags;
 
 	/* Add the start of subexpression as a new atom */
 
-	if ((atoms = add_atom(atoms, match_start_subexpr,
-			      NULL, 0, 0)) == NULL) {
+	if ((catoms = add_atom(catoms, match_start_subexpr,
+			       NULL, 0, 0)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
@@ -612,7 +612,7 @@ int flags;
 
 	/* Store the start of the subexpression */
 
-	start = last;
+	start = clast;
 
 	/* Skip the start of the subexpression */
 
@@ -638,8 +638,8 @@ int flags;
 
 	/* Add the end of subexpression as a new atom */
 
-	if ((atoms = add_atom(atoms, match_end_subexpr,
-			      NULL, 0, 0)) == NULL) {
+	if ((catoms = add_atom(catoms, match_end_subexpr,
+			       NULL, 0, 0)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
@@ -648,9 +648,9 @@ int flags;
 	/* Now set up the subexpression's details */
 
 	start->ra_subexpr = subexpr;
-	start->ra_ref = last;
-	last->ra_subexpr = subexpr;
-	last->ra_ref = start;
+	start->ra_ref = clast;
+	clast->ra_subexpr = subexpr;
+	clast->ra_ref = start;
 
 	/* And return the position in the expression */
 
@@ -664,21 +664,21 @@ int flags;
 {
 	/* Compile a back reference into the expression */
 
-	__regatom_t *atom = atoms;
+	__regatom_t *atom = catoms;
 
 	/* First add the back reference as an atom */
 
-	if ((atoms = add_atom(atoms, match_backref, NULL, 0, -1)) == NULL) {
+	if ((catoms = add_atom(catoms, match_backref, NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}		
-	last->ra_subexpr = (*pos - '0');
+	clast->ra_subexpr = (*pos - '0');
 
 	/* Check the subexpression can be back referenced */
 
 	while (atom != NULL && (atom->ra_matchfunc != match_end_subexpr ||
-				atom->ra_subexpr != last->ra_subexpr)) {
+				atom->ra_subexpr != clast->ra_subexpr)) {
 		/* Move on to the next atom */
 
 		atom = atom->ra_next;
@@ -704,7 +704,7 @@ int flags;
 
 	/* Find the start of the first alternative */
 
-	atom = last;
+	atom = clast;
 	while (atom != NULL && atom->ra_matchfunc != match_start_subexpr
 	       && atom->ra_matchfunc != match_start_alternative) {
 		/* Move to the previous logical atom in the expression */
@@ -729,27 +729,27 @@ int flags;
 
 	/* Check if we've found the start of the alternative */
 
-	start = (atom != NULL) ? atom : atoms;
+	start = (atom != NULL) ? atom : catoms;
 
 	/* Insert a 'start of alternative' before the LHS */
 
-	if ((atoms = insert_atom(atoms, start, match_start_alternative,
-				 NULL, 0, -1)) == NULL) {
+	if ((catoms = insert_atom(catoms, start, match_start_alternative,
+				  NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}		
-	start = (start != NULL) ? start->ra_prev : atoms;
+	start = (start != NULL) ? start->ra_prev : catoms;
 
 	/* Add and set up the end of the first alternative */
 
-	if ((atoms = add_atom(atoms, match_end_alternative,
-			     NULL, 0, -1)) == NULL) {
+	if ((catoms = add_atom(catoms, match_end_alternative,
+			       NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}		
-	start->ra_ref = alternative = last;
+	start->ra_ref = alternative = clast;
 
 	/* Now skip to the start of the next expression */
 
@@ -766,13 +766,13 @@ int flags;
 
 	/* Add and set up the end of the second alternative */
 
-	if ((atoms = add_atom(atoms, match_end_alternative,
-			      NULL, 0, -1)) == NULL) {
+	if ((catoms = add_atom(catoms, match_end_alternative,
+			       NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}
-	alternative->ra_ref = last;
+	alternative->ra_ref = clast;
 
 	/* And return the updated position */
 
@@ -789,13 +789,13 @@ int flags;
 
 	/* Check that this repeat is valid */
 
-	if (last == NULL || last->ra_length <= 0
-	    && last->ra_matchfunc != match_end_subexpr) {
+	if (clast == NULL || clast->ra_length <= 0 &&
+	    clast->ra_matchfunc != match_end_subexpr) {
 		/* An error for extended, or a char match for basic REs */
 
 		reg_error = REG_BADRPT;
 		return((!(flags & REG_EXTENDED) &&
-			(last == NULL || last->ra_matchfunc
+			(clast == NULL || clast->ra_matchfunc
 			 != match_end_repeat))
 		       ? compile_char(pos, flags) : NULL);
 	}
@@ -806,32 +806,32 @@ int flags;
 
 	/* Extract the last subexpresssion from the expression */
 
-	start = (last->ra_matchfunc == match_end_subexpr)
-		? last->ra_ref : last;
+	start = (clast->ra_matchfunc == match_end_subexpr)
+		? clast->ra_ref : clast;
 
 	/* Add and set up the start of the repeat */
 
-	if ((atoms = insert_atom(atoms, start, match_start_repeat,
-				 NULL, 0, -1)) == NULL) {
+	if ((catoms = insert_atom(catoms, start, match_start_repeat,
+				  NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}		
-	start = (start != NULL) ? start->ra_prev : atoms;
+	start = (start != NULL) ? start->ra_prev : catoms;
 	start->ra_min = (*pos == '+') ? 1 : 0;
 	start->ra_max = (*pos == '?') ? 1 : -1;
 
 	/* Add and set up the end of the repeat */
 
-	if ((atoms = add_atom(atoms, match_end_repeat,
-			      NULL, 0, -1)) == NULL) {
+	if ((catoms = add_atom(catoms, match_end_repeat,
+			       NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}
-	last->ra_min = last->ra_max = 0;
-	last->ra_ref = start;
-	start->ra_ref = last;
+	clast->ra_min = clast->ra_max = 0;
+	clast->ra_ref = start;
+	start->ra_ref = clast;
 
 	/* And return the updated position */
 
@@ -887,8 +887,8 @@ int flags;
 
 	/* We have a repeat, but is it valid? */
 
-	if (last == NULL || last->ra_length <= 0 &&
-	    last->ra_matchfunc != match_end_subexpr) {
+	if (clast == NULL || clast->ra_length <= 0 &&
+	    clast->ra_matchfunc != match_end_subexpr) {
 		/* This is an error */
 
 		reg_error = REG_BADRPT;
@@ -920,39 +920,39 @@ int flags;
 
 	/* Extract the last subexpresssion from the expression */
 
-	repeat = (last->ra_matchfunc == match_end_subexpr)
-		? last->ra_ref : last;
-	old_last = last;
-	last = repeat->ra_prev;
+	repeat = (clast->ra_matchfunc == match_end_subexpr)
+		? clast->ra_ref : clast;
+	old_last = clast;
+	clast = repeat->ra_prev;
 	repeat->ra_prev = NULL;
-	atoms = (last != NULL) ? atoms : NULL;
+	catoms = (clast != NULL) ? catoms : NULL;
 
 	/* Add and set up the start of the repeat */
 
-	if ((atoms = add_atom(atoms, match_start_repeat,
-			      NULL, 0, -1)) == NULL) {
+	if ((catoms = add_atom(catoms, match_start_repeat,
+			       NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}		
-	last->ra_min = min;
-	last->ra_max = max;
-	last->ra_next = repeat;
-	repeat->ra_prev = last;
-	repeat = last;
-	last = old_last;
+	clast->ra_min = min;
+	clast->ra_max = max;
+	clast->ra_next = repeat;
+	repeat->ra_prev = clast;
+	repeat = clast;
+	clast = old_last;
 
 	/* Add and set up the end of the repeat */
 
-	if ((atoms = add_atom(atoms, match_end_repeat,
-			      NULL, 0, -1)) == NULL) {
+	if ((catoms = add_atom(catoms, match_end_repeat,
+			       NULL, 0, -1)) == NULL) {
 		/* Out of memory compiling the expression */
 
 		return(NULL);
 	}		
-	last->ra_min = last->ra_max = 0;
-	last->ra_ref = repeat;
-	repeat->ra_ref = last;
+	clast->ra_min = clast->ra_max = 0;
+	clast->ra_ref = repeat;
+	repeat->ra_ref = clast;
 
 	/* And return the updated position */
 
@@ -1047,11 +1047,11 @@ int vlen, len;
 
 	/* Add the atom to the list */
 
-	if (atoms != NULL && last != NULL) {
-		last->ra_next = node;
+	if (atoms != NULL && clast != NULL) {
+		clast->ra_next = node;
 	}
-	node->ra_prev = last;
-	last = node;
+	node->ra_prev = clast;
+	clast = node;
 
 	/* And return the updated list */
 
@@ -1977,7 +1977,7 @@ int subexpr;
 
 	/* Search the atom list for the subexpression */
 
-	for (atom = atoms; atom != NULL; atom = atom->ra_next) {
+	for (atom = catoms; atom != NULL; atom = atom->ra_next) {
 		/* Is this the subexpression we want? */
 
 		if (atom->ra_subexpr == subexpr) {
