@@ -1,5 +1,5 @@
 /* Af.c - Main module for af.
-   Copyright (C) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1990 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -37,7 +37,7 @@
  */
 
 #ifndef lint
-static char *RcsId = "$Id: af.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: af.c,v 2.3 2002/08/21 23:54:48 malc Exp $";
 static char *HeaderId = HEADERID;
 static char *ConfigId = CONFIGID;
 static char *VersionId = VERSIONID;
@@ -50,16 +50,16 @@ extern char *xmalloc(), *xrealloc(), *xstrdup();
 extern char *vstrcat(), *get_user(), *get_home();
 extern char *get_mailbox(), *get_vtext(), *expand();
 extern char *fullpath(), *alias(), *a_strerror();
-extern int getopt(), isatty(), nonempty();
-extern int count_buffers(), count_windows();
+extern int getopt(), isatty(), nonempty(), send_mail();
+extern int quiet_mail(), count_buffers(), count_windows();
+extern int read_mime_config();
 extern void free(), free_forms(), exit(), init_keymaps();
 extern void init_history(), init_vars(), reset_var();
-extern void read_aliases(), send_mail(), quiet_mail();
-extern void init_tcntrl(), init_tmodes(), init_vterm();
-extern void end_tmodes(), interactive(), silent();
-extern void errors_to_typeout(), msg(), msgl(), cmsg();
-extern void emsgl(), show_initial_msg(), tbeep();
-extern void show_buffer(), first_display(), hide_cursor();
+extern void read_aliases(), init_tcntrl(), init_tmodes();
+extern void init_vterm(), end_tmodes(), errors_to_typeout();
+extern void interactive(), silent(), hide_cursor(), tbeep();
+extern void msg(), msgl(), cmsg(), emsgl(), show_buffer();
+extern void first_display(), show_initial_msg();
 extern void do_commands(), run_script();
 extern RETSIGTYPE (*signal())();
 extern WINDOW *init_windows();
@@ -100,7 +100,7 @@ extern WINDOW *cwin;
 /****************************************************************************/
 /* The user options are file-scoped */
 
-static USEROPTS *options;
+static USEROPTS *user_options;
 
 /****************************************************************************/
 int main(argc, argv)
@@ -109,8 +109,10 @@ char **argv;
 {
 	/* Process and handle the arguments */
 
-	char *users = NULL;
-	int sending;
+	char *addrs = NULL;
+	char *cc = NULL;
+	char *bcc = NULL;
+	int sending, status;
 
 #ifndef READ_VIA_PIPES
 	/* Revert to the real uid and gid */
@@ -130,23 +132,24 @@ char **argv;
 
 	/* Parse the command line arguments */
 
-	options = getargs(argc, argv);
+	user_options = getargs(argc, argv);
 
 	/* Handle just printing the version */
 
-	if (options->version) {
+	if (user_options->version) {
 		show_version();
 		return(0);
 	}
 
 	/* Are we in sending or interactive mode? */
 
-	sending = (options->users != NULL || options->edit);
+	sending = (user_options->users != NULL || user_options->cc != NULL
+		   || user_options->bcc != NULL || user_options->edit);
 
 	/* Initialise the terminal */
 
 	init_tcntrl();
-	init_tmodes();
+	init_tmodes(TRUE);
 	init_vterm();
 
 	/* Check that stdin & stdout are ok */
@@ -154,22 +157,22 @@ char **argv;
 	if (!isatty(fileno(stdin)) && !sending) {
 		panic("Stdin not a tty");
 	}
-	if (!isatty(fileno(stdout)) && (!sending || options->edit)) {
+	if (!isatty(fileno(stdout)) && (!sending || user_options->edit)) {
 		panic("Stdout not a tty");
 	}
 
 	/* Handle the 'does mail exist' and 'only if exists' options */
 
-	if (!sending && (options->exists || options->zero)) {
-		chk_exists(options);
+	if (!sending && (user_options->exists || user_options->zero)) {
+		chk_exists(user_options);
 	}
 
 	/* Become interactive if required */
 
-	if (!sending && options->script == NULL) {
+	if (!sending && user_options->script == NULL) {
 		interactive();
 		show_initial_msg();
-	} else if (options->script != NULL) {
+	} else if (user_options->script != NULL) {
 		silent();
 	}
 
@@ -179,8 +182,8 @@ char **argv;
 
 	/* Read the startup file, verbosely unless sending from stdin */
 
-	if (!read_startup(options->loadfile, options->nostartup,
-			  isatty(fileno(stdin)) || options->edit)) {
+	if (!read_startup(user_options->loadfile, user_options->nostartup,
+			  isatty(fileno(stdin)) || user_options->edit)) {
 		/* Exit in startup file; honour it */
 
 		end_tmodes();
@@ -196,24 +199,32 @@ char **argv;
 	if (sending) {
 		/* Alias and expand the user list */
 
-		if (options->users != NULL &&
-		    (users = alias(options->users)) == NULL) {
+		if (user_options->users != NULL &&
+		    (addrs = alias(user_options->users)) == NULL
+		    || user_options->cc != NULL &&
+		    (cc = alias(user_options->cc)) == NULL
+		    || user_options->bcc != NULL &&
+		    (bcc = alias(user_options->bcc)) == NULL) {
 			/* Managed to fail the alias */
 
 			panic(a_strerror());
 		}
 
-		if (options->edit || isatty(fileno(stdin))) {
-			send_mail(users, options->subject);
+		if (user_options->edit || isatty(fileno(stdin))) {
+			status = send_mail(addrs, cc, bcc,
+					   user_options->subject,
+					   user_options->contype, FALSE);
 		} else {
-			quiet_mail(users, options->subject);
+			status = quiet_mail(addrs, cc, bcc,
+					    user_options->subject,
+					    user_options->contype);
 		}
 
 		/* Clean up the display before we exit */
 
 		end_tmodes();
 		(void) putchar('\n');
-		return(0);
+		return((status) ? 0 : 1);
 	}
 
 	/* Start up buffers and windows */
@@ -222,8 +233,8 @@ char **argv;
 
 	/* Now go and handle commands, from the keyboard or a script file */
 
-	if (options->script != NULL) {
-		run_script(options->script);
+	if (user_options->script != NULL) {
+		run_script(user_options->script);
 	} else {
 		do_commands();
 	}
@@ -231,7 +242,7 @@ char **argv;
 #ifdef READ_VIA_POP3
 	/* Close any open POP3 folders */
 
-	kill_pop3();
+	kill_pop3(NULL);
 #endif /* READ_VIA_POP3 */
 
 	/* Clean up the terminal and exit */
@@ -256,19 +267,23 @@ int finish_startup()
 
 	/* Are we in sending or interactive mode? */
 
-	sending = (options->users != NULL || options->edit);
+	sending = (user_options->users != NULL || user_options->edit);
 
 	/* Read the alias files */
 
-	read_aliases(isatty(fileno(stdin)) || options->edit);
+	read_aliases(isatty(fileno(stdin)) || user_options->edit);
+
+	/* Read the MIME configuration */
+
+	(void) read_mime_config();
 
 	/* Normalise errors unless required later */
 
-	errors_to_typeout(!sending && options->folders[1] != NULL);
+	errors_to_typeout(!sending && user_options->folders[1] != NULL);
 
 	/* Handle the 'edit headers' option */
 
-	if (options->hdrs) {
+	if (user_options->hdrs) {
 		(void) reset_var(V_EDIT_IHDRS, VI_TRUE);
 	}
 
@@ -290,7 +305,8 @@ int startup_display()
 
 	/* Check if we've already started the display */
 
-	if (cwin != NULL || options->users != NULL || options->edit) {
+	if (cwin != NULL || user_options->users != NULL
+	    || user_options->edit) {
 		/* Success if already done; failed otherwise */
 
 		return(cwin != NULL);
@@ -298,7 +314,7 @@ int startup_display()
 
 	/* Check if we have a pending folder */
 
-	chk_pending(options);
+	chk_pending(user_options);
 
 	/* Initialise the scratch buffer */
 
@@ -306,15 +322,15 @@ int startup_display()
 
 	/* Make and read the required buffers */
 
-	for (folder = 0; options->folders[folder] != NULL; folder++) {
+	for (folder = 0; user_options->folders[folder] != NULL; folder++) {
 		/* Is there a pending folder for this buffer? */
 
-		pfolder = (options->pending) ?
+		pfolder = (user_options->pending) ?
 			fullpath(get_mailbox(NULL)) : NULL;
 
 		/* Make and read the buffer */
 
-		buf = add_buffer(buf, NULL, options->folders[folder],
+		buf = add_buffer(buf, NULL, user_options->folders[folder],
 				 pfolder, M_MAIL);
 	}
 
@@ -325,7 +341,7 @@ int startup_display()
 
 	/* Set up as many windows as we need */
 
-	cwin = winlist = init_windows((options->windows) ? nbuffers : 1);
+	cwin = winlist = init_windows((user_options->windows) ? nbuffers : 1);
 	nwindows = count_windows(winlist);
 
 	/* Show each buffer in a window */
@@ -421,41 +437,27 @@ char **argv;
 	options.pending = options.version = FALSE;
 	options.folders = NULL;
 	options.loadfile = options.script = NULL;
-	options.subject = options.users = NULL;
+	options.subject = options.contype = NULL;
+	options.users = options.cc = options.bcc = NULL;
 
 	/* Now parse the command line */
 
 	while ((opt = getopt(argc, argv, OPTS)) != EOF) {
 		switch (opt) {
-		case 'e':
-			options.exists = TRUE;
-			break;
-		case 'z':
-			options.zero = TRUE;
+		case 'C':
+			if (options.contype != NULL) {
+				usage(argv[0]);
+			}
+			options.contype = xstrdup(optarg);
 			break;
 		case 'E':
 			options.edit = TRUE;
 			break;
-		case 'H':
-			options.hdrs = TRUE;
-			break;
-		case 'n':
-			options.nostartup = TRUE;
-			break;
-		case 'w':
-			options.windows = TRUE;
-			break;
-		case 'l':
-			options.loadfile = expand(optarg);
-			break;
-		case 'f':
-			add_folder(&options, fullpath(optarg));
-			break;
 		case 'F':
 			all_folders = TRUE;
 			break;
-		case 'u':
-			add_folder(&options, fullpath(get_mailbox(optarg)));
+		case 'H':
+			options.hdrs = TRUE;
 			break;
 		case 'S':
 			if (options.script != NULL) {
@@ -463,15 +465,48 @@ char **argv;
 			}
 			options.script = expand(optarg);
 			break;
+		case 'b':
+			if (options.bcc != NULL) {
+				usage(argv[0]);
+			}
+			options.bcc = xstrdup(optarg);
+			break;
+		case 'c':
+			if (options.cc != NULL) {
+				usage(argv[0]);
+			}
+			options.cc = xstrdup(optarg);
+			break;
+		case 'e':
+			options.exists = TRUE;
+			break;
+		case 'f':
+			add_folder(&options, fullpath(optarg));
+			break;
+		case 'l':
+			options.loadfile = expand(optarg);
+			break;
+		case 'n':
+			options.nostartup = TRUE;
+			break;
 		case 's':
 			if (options.subject != NULL) {
 				usage(argv[0]);
 			}
 			options.subject = xstrdup(optarg);
 			break;
+		case 'u':
+			add_folder(&options, fullpath(get_mailbox(optarg)));
+			break;
 		case 'v':
 			options.version = TRUE;
 			break;
+		case 'w':
+			options.windows = TRUE;
+			break;
+		case 'z':
+			options.zero = TRUE;
+			break;
 		case '?':
 			/* Give a usage message and quit */
 
@@ -696,7 +731,7 @@ USEROPTS *options;
 			(void) printf("%s mail in %s.\n", (exists)
 				      ? "There is" : "No",
 				      options->folders[folder]);
-			init_tmodes();
+			init_tmodes(FALSE);
 		}
 
 		/* Delete the folder from the list if required */
@@ -766,9 +801,9 @@ char *prognam;
 {
 	/* Print a usage message and exit abnormally */
 
-	(void) fprintf(stderr, "Usage : %s [ -envwzEFHX ][ -l loadfile ]\
-[ -S scriptfile ]\n\t[ -f folder ] [ -u user ][ -s subject ] [ user ... ]\n",
-		       prognam);
+	(void) fprintf(stderr, "Usage : %s [ -envwzEFHX ][ -f folder ]\
+[ -u user ][ -l loadfile ][ -S scriptfile ]\n\t[ -c cc ][ -b bcc ]\
+[ -s subject ]\n\t[ -C content-type ] [ user ... ]\n", prognam);
 	exit(2);
 }
 /****************************************************************************/
