@@ -1,5 +1,5 @@
 /* Atom.c - Code to handle atoms of RFC 822 headers.
-   Copyright (C) 1992, 1993, 1994, 1996, 1997 Malc Arnold.
+   Copyright (C) 1992 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: atom.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: atom.c,v 2.4 2002/08/21 23:54:48 malc Exp $";
 static char *AtomId = ATOMID;
 #endif /* ! lint */
 
@@ -39,13 +39,17 @@ extern void free();
 
 /* Local function declarations */
 
-char *atext();
+char *atext(), *avalue();
 void afree();
-ATOM *add_atom(), *insert_atom();
-static char *get_atom(), *set_char();
-static char *set_str(), *set_ws();
-static char *avalue(), *escape_chars();
+ATOM *aappend(), *ainsert(), *aadd();
+ATOM *aprev(), *acut();
+static char *get_atom(), *get_catom();
+static char *get_ctatom(), *get_matom();
+static char *get_tatom(), *get_watom();
+static char *set_char(), *set_str();
+static char *set_ws(), *escape_chars();
 static char *unescape_chars();
+static char *unfold_atom();
 static ATOM *atomise();
 
 /****************************************************************************/
@@ -60,7 +64,7 @@ char *text;
 {
 	/* Break down some text into its component parts */
 
-	return(atomise(text, FALSE));
+	return(atomise(text, get_atom));
 }
 /****************************************************************************/
 ATOM *ctokenise(text)
@@ -68,17 +72,48 @@ char *text;
 {
 	/* Break down a content type into its component parts */
 
-	return(atomise(text, TRUE));
+	return(atomise(text, get_catom));
 }
 /****************************************************************************/
-static ATOM *atomise(text, mime)
+ATOM *cttokenise(text)
 char *text;
-int mime;
+{
+	/* Break down the text of a comment into its component parts */
+
+	return(atomise(text, get_ctatom));
+}
+/****************************************************************************/
+ATOM *mtokenise(text)
+char *text;
+{
+	/* Break down a mailcap entry into its component parts */
+
+	return(atomise(text, get_matom));
+}
+/****************************************************************************/
+ATOM *ttokenise(text)
+char *text;
+{
+	/* Break down a mime.types entry into its component parts */
+
+	return(atomise(text, get_tatom));
+}
+/****************************************************************************/
+ATOM *wtokenise(text)
+char *text;
+{
+	/* Break down some text into its component words */
+
+	return(atomise(text, get_watom));
+}
+/****************************************************************************/
+static ATOM *atomise(text, get_func)
+char *text, *(*get_func)();
 {
 	/* Break down some text into its component parts */
 
-	char *start, *end;
-	int type, tmp;
+	char *start, *end, *value;
+	int type;
 	ATOM *alist = NULL;
 	ATOM *last = NULL;
 
@@ -103,33 +138,33 @@ int mime;
 	while (*start != '\0') {
 		/* Get the next atomic item from the text */
 
-		if ((end = get_atom(start, mime, &type)) == NULL) {
+		if ((end = get_func(start, &type)) == NULL) {
 			afree(alist);
 			return(NULL);
 		}
 
-		/* Store the character after the atom */
+		/* Extract the atom from the text */
 
-		tmp = *end;
-		*end = '\0';
+		value = xmalloc(end - start + 1);
+		(void) strncpy(value, start, end - start);
+		value[end - start] = '\0';
 
 		/* Add the atom to the list */
 
-		alist = insert_atom(alist, last, start, type);
+		alist = aadd(alist, last, value, type);
 		last = (last != NULL) ? last->next : alist;
+		
+		/* Free space and advance start */
 
-		/* Restore the original character and advance start */
-
-		*end = tmp;
+		free(value);
 		start = end;
 	}
 
 	return(alist);
 }
 /****************************************************************************/
-static char *get_atom(start, mime, type)
+static char *get_atom(start, type)
 char *start;
-int mime;
 int *type;
 {
 	/* Find the end of the current atom and set its type */
@@ -148,16 +183,69 @@ int *type;
 		return(set_ws(start, type));
 	case '(':				/* Comment */
 		return(set_str(start, SC_COMMENT, TC_COMMENT,
-			       AT_COMMENT, type));
+			       TRUE, AT_COMMENT, type));
 	case '"':				/* Quoted string */
-		return(set_str(start, SC_QSTRING, TC_QSTRING,
-			       AT_QSTRING, type));
+		return(set_str(start, SC_QSTRING, TC_QSTRING, 
+			       TRUE, AT_QSTRING, type));
 	case '[':				/* Domain literal */
 		return(set_str(start, SC_DLITERAL, TC_DLITERAL,
-			       AT_DLITERAL, type));
+			       TRUE, AT_DLITERAL, type));
 	case '.':
-		return((!mime) ? set_char(start, AT_DOT, type) :
-		       set_str(start, SC_ATOM, TC_ATOM, AT_ATOM, type));
+		return(set_char(start, AT_DOT, type));
+	case '@':
+		return(set_char(start, AT_AT, type));
+	case '%':
+		return(set_char(start, AT_PERCENT, type));
+	case ',':
+		return(set_char(start, AT_COMMA, type));
+	case '<':
+		return(set_char(start, AT_LANGLEB, type));
+	case '>':
+		return(set_char(start, AT_RANGLEB, type));
+	case ':':
+		return(set_char(start, AT_COLON, type));
+	case ';':
+		return(set_char(start, AT_SEMI, type));
+	case ')':
+	case ']':
+	case '\\':
+		a_errno = AERR_ADDRESS;
+		a_errtext = xstrdup(start);
+		return(NULL);
+	default:
+		return(set_str(start, SC_ATOM, TC_ATOM,
+			       FALSE, AT_ATOM, type));
+	}
+	/*NOTREACHED*/
+}
+/****************************************************************************/
+static char *get_catom(start, type)
+char *start;
+int *type;
+{
+	/* Find the end of the current content-type atom and set its type */
+
+	/* We don't know what type this atom is yet */
+
+	*type = AT_WS;
+
+	/* Generate the atom */
+
+	switch (*start) {
+	case ' ':				/* Whitespace */
+	case '\t':
+	case '\n':
+	case '\r':
+		return(set_ws(start, type));
+	case '(':				/* Comment */
+		return(set_str(start, SC_COMMENT, TC_COMMENT,
+			       TRUE, AT_COMMENT, type));
+	case '"':				/* Quoted string */
+		return(set_str(start, SC_QSTRING, TC_QSTRING,
+			       TRUE, AT_QSTRING, type));
+	case '[':				/* Domain literal */
+		return(set_str(start, SC_DLITERAL, TC_DLITERAL,
+			       TRUE, AT_DLITERAL, type));
 	case '@':
 		return(set_char(start, AT_AT, type));
 	case '%':
@@ -179,71 +267,129 @@ int *type;
 		a_errtext = xstrdup(start);
 		return(NULL);
 	case '/':
-		return((mime) ? set_char(start, AT_SLASH, type) :
-			set_str(start, SC_ATOM, TC_ATOM, AT_ATOM, type));
+		return(set_char(start, AT_SLASH, type));
 	case '?':
-		return((mime) ? set_char(start, AT_QUERY, type) :
-			set_str(start, SC_ATOM, TC_ATOM, AT_ATOM, type));
+		return(set_char(start, AT_QUERY, type));
 	case '=':
-		return((mime) ? set_char(start, AT_EQUALS, type) :
-			set_str(start, SC_ATOM, TC_ATOM, AT_ATOM, type));
+		return(set_char(start, AT_EQUALS, type));
 	default:
-		return(set_str(start, SC_ATOM, (mime) ? TC_MATOM :
-			       TC_ATOM, AT_ATOM, type));
+		return(set_str(start, SC_ATOM, TC_CATOM,
+			       FALSE, AT_ATOM, type));
 	}
 	/*NOTREACHED*/
 }
 /****************************************************************************/
-ATOM *add_atom(alist, atom, type)
-ATOM *alist;
-char *atom;
-int type;
+static char *get_ctatom(start, type)
+char *start;
+int *type;
 {
-	/* Add an atom to the list */
+	/* Find the end of the current comment atom and set its type */
 
-	ATOM *node;
+	/* We don't know what type this atom is yet */
 
-	if (alist == NULL) {
-		/* Initialise the new node */
-	
-		node = (ATOM *) xmalloc(sizeof(ATOM));
-		node->text = xstrdup(atom);
-		node->type = type;
-		node->next = NULL;
-		return(node);
-	} else {
-		alist->next = add_atom(alist->next, atom, type);
-		return(alist);
+	*type = AT_WS;
+
+	/* Generate the atom */
+
+	switch (*start) {
+	case ' ':				/* Whitespace */
+	case '\t':
+	case '\n':
+	case '\r':
+		return(set_ws(start, type));
+	case '(':				/* Comment */
+		return(set_str(start, SC_COMMENT, TC_COMMENT,
+			       TRUE, AT_COMMENT, type));
+	default:
+		return(set_str(start, SC_ATOM, TC_CTATOM,
+			       TRUE, AT_ATOM, type));
 	}
 	/*NOTREACHED*/
 }
 /****************************************************************************/
-ATOM *insert_atom(alist, where, atom, type)
-ATOM *alist, *where;
-char *atom;
-int type;
+static char *get_matom(start, type)
+char *start;
+int *type;
 {
-	/* Append an atom after where in alist */
+	/* Find the end of the current mailcap atom and set its type */
 
-	ATOM *node;
+	/* We don't know what type this atom is yet */
 
-	/* Create and fill the node for the new atom */
+	*type = AT_WS;
 
-	node = (ATOM *) xmalloc(sizeof(ATOM));
-	node->text = xstrdup(atom);
-	node->type = type;
-	node->next = NULL;
+	/* Generate the atom */
 
-	/* Append the node after where if possible */
-
-	if (alist != NULL && where != NULL) {
-		node->next = where->next;
-		where->next = node;
+	switch (*start) {
+	case ' ':				/* Whitespace */
+	case '\t':
+	case '\n':
+	case '\r':
+		return(set_ws(start, type));
+	case '#':
+		return(set_str(start, SC_MCOMMENT, TC_MCOMMENT,
+			       TRUE, AT_COMMENT, type));
+	case '=':
+		return(set_char(start, AT_EQUALS, type));
+	case ';':
+		return(set_char(start, AT_SEMI, type));
+	default:
+		return(set_str(start, SC_ATOM, TC_MATOM,
+			       TRUE, AT_ATOM, type));
 	}
+	/*NOTREACHED*/
+}
+/****************************************************************************/
+static char *get_tatom(start, type)
+char *start;
+int *type;
+{
+	/* Find the end of the current mime.types atom and set its type */
 
-	/* And return the updated atom list */
+	/* We don't know what type this atom is yet */
 
-	return((alist != NULL) ? alist : node);
+	*type = AT_WS;
+
+	/* Generate the atom */
+
+	switch (*start) {
+	case ' ':				/* Whitespace */
+	case '\t':
+	case '\n':
+	case '\r':
+		return(set_ws(start, type));
+	case '#':
+		return(set_str(start, SC_TCOMMENT, TC_TCOMMENT,
+			       TRUE, AT_COMMENT, type));
+	default:
+		return(set_str(start, SC_ATOM, TC_TATOM,
+			       TRUE, AT_ATOM, type));
+	}
+	/*NOTREACHED*/
+}
+/****************************************************************************/
+static char *get_watom(start, type)
+char *start;
+int *type;
+{
+	/* Find the end of the current word atom and set its type */
+
+	/* We don't know what type this atom is yet */
+
+	*type = AT_WS;
+
+	/* Generate the atom */
+
+	switch (*start) {
+	case ' ':				/* Whitespace */
+	case '\t':
+	case '\n':
+	case '\r':
+		return(set_ws(start, type));
+	default:
+		return(set_str(start, SC_WORD, TC_WORD,
+			       FALSE, AT_ATOM, type));
+	}
+	/*NOTREACHED*/
 }
 /****************************************************************************/
 static char *set_char(start, atype, type)
@@ -259,9 +405,9 @@ int atype, *type;
 	return(start + 1);
 }
 /****************************************************************************/
-static char *set_str(atom, start, term, atype, type)
+static char *set_str(atom, start, term, quoting, atype, type)
 char *atom, *start, *term;
-int atype, *type;
+int quoting, atype, *type;
 {
 	/* Handle multi-character atoms */
 
@@ -277,7 +423,7 @@ int atype, *type;
 	for (p = atom; *p != '\0'; p++) {
 		/* Handle escaped characters */
 
-		if (atype != AT_ATOM && *p == '\\' && *(p + 1) != '\0') {
+		if (quoting && *p == '\\' && *(p + 1) != '\0') {
 			/* Quoted special character; skip */
 
 			p++;
@@ -336,217 +482,203 @@ int *type;
 	return(p);
 }
 /****************************************************************************/
-ATOM *find_token(token)
-ATOM *token;
-{
-	/*
-	 * Return the first token that isn't whitespace or comment
-	 * in the list passed in, or NULL if none found.
-	 */
-
-	ATOM *a;
-
-	/* Loop over all the whitespace or comment tokens */
-
-	for (a = token; a != NULL && IS_WS(a); a = a->next) {
-		/* NULL LOOP */
-	}
-
-	return(a);
-}
-/****************************************************************************/
-ATOM *asearch(alist, type)
+ATOM *aappend(alist, text, type)
 ATOM *alist;
+char *text;
 int type;
 {
-	/* Return the first token of type 'type' in alist */
+	/* Append an atom to the list, by inserting at the end */
 
-	ATOM *a;
-
-	/* Loop until we find a token of the right type */
-
-	for (a = alist; a != NULL && a->type != type; a = a->next) {
-		/* NULL LOOP */
-	}
-
-	return(a);
+	return(ainsert(alist, NULL, text, type));
 }
 /****************************************************************************/
-ATOM *adiscard(alist, token)
-ATOM *alist, *token;
+ATOM *ainsert(alist, where, text, type)
+ATOM *alist, *where;
+char *text;
+int type;
 {
-	/* Discard token from the list and return the start of the list */
+	/* Insert an atom before where in alist */
 
-	ATOM *a;
+	ATOM *node, *prev;
 
-	/* Check the token isn't at the head of the list */
+	/* Create and fill the node for the new atom */
 
-	if (alist == token) {
-		alist = alist->next;
-	} else {
-		/* Search until we find token */
+	node = (ATOM *) xmalloc(sizeof(ATOM));
+	node->text = xstrdup(text);
+	node->type = type;
+	node->next = NULL;
 
-		for (a = alist; a != NULL; a = a->next) {
-			/* And strip it out of the list */
+	/* Find the atom before where in the list */
 
-			if (a->next == token) {
-				a->next = token->next;
-				break;
-			}
-		}
+	if ((prev = aprev(alist, where)) == NULL) {
+		/* No previous atom, prepend the node */
+
+		node->next = alist;
+		return(node);
 	}
 
-	/* Now free the discarded token */
+	/* Insert the node at the position we found */
 
-	free(token->text);
-	free(token);
+	node->next = prev->next;
+	prev->next = node;
+
+	/* And return the updated atom list */
 
 	return(alist);
 }
 /****************************************************************************/
-ATOM *acut(alist, end)
-ATOM *alist, *end;
+ATOM *aadd(alist, where, text, type)
+ATOM *alist, *where;
+char *text;
+int type;
 {
-	/*
-	 * Strip atoms start up to (but not including) end out of
-	 * a list and return the start token of those cut out.
-	 */
+	/* Append an atom after where in alist */
 
-	ATOM *a;
+	ATOM *node;
 
-	/* This is a no-op if end is null */
+	/* Create and fill the node for the new atom */
 
-	if (end != NULL) {
-		/* Check there is something to cut */
+	node = (ATOM *) xmalloc(sizeof(ATOM));
+	node->text = xstrdup(text);
+	node->type = type;
+	node->next = NULL;
 
-		if (alist == end) {
-			return(NULL);
-		}
+	/* Append the node after where if possible */
 
-		/* Search for the end token */
-
-		for (a = alist; a != NULL; a = a->next) {
-			/* And cut the list before the end token */
-
-			if (a->next == end) {
-				a->next = NULL;
-				break;
-			}
-		}
+	if (alist != NULL && where != NULL) {
+		node->next = where->next;
+		where->next = node;
 	}
 
-	/* Now return the modified list */
+	/* And return the updated atom list */
+
+	return((alist != NULL) ? alist : node);
+}
+/****************************************************************************/
+ATOM *amerge(alist, where, mlist)
+ATOM *alist, *where, *mlist;
+{
+	/* Insert mlist into alist before where, and return the new list */
+
+	ATOM *last, *prev;
+
+	/* First find the last atom in mlist */
+
+	if ((last = aprev(mlist, NULL)) == NULL) {
+		/* Nothing to insert, just return */
+
+		return(alist);
+	}
+
+	/* Find the atom before where in the list */
+
+	if ((prev = aprev(alist, where)) == NULL) {
+		/* No previous atom, prepend the list */
+
+		last->next = alist;
+		return(mlist);
+	}
+
+	/* Now merge the lists together */
+
+	last->next = prev->next;
+	prev->next = mlist;
+
+	/* And return the updated atom list */
 
 	return(alist);
 }
 /****************************************************************************/
-ATOM *acopy(alist)
-ATOM *alist;
-{
-	/* Return an allocated copy of alist */
-
-	ATOM *a, *new_list = NULL;
-
-	/* Simply add each token to the new list */
-
-	for (a = alist; a != NULL; a = a->next) {
-		new_list = add_atom(new_list, a->text, a->type);
-	}
-
-	return(new_list);
-}
-/****************************************************************************/
-ATOM *acat(alist, next)
-ATOM *alist, *next;
-{
-	/* Concatenate next to alist and return the new list */
-
-	ATOM *a;
-
-	/* Return next if the list is empty */
-
-	if (alist == NULL) {
-		return(next);
-	}
-
-	/* Now search for the end of the list */
-
-	for (a = alist; a != NULL; a = a->next) {
-		/* And append next to the list */
-
-		if (a->next == NULL) {
-			a->next = next;
-			break;
-		}
-	}
-
-	/* Return the new list */
-
-	return(alist);
-}
-/****************************************************************************/
-ATOM *asplit(alist, token, pos)
-ATOM *alist, *token;
+ATOM *asplit(alist, atom, pos)
+ATOM *alist, *atom;
 char *pos;
 {
 	/*
-	 * Split token into two atoms at the point in the text
+	 * Split an atom list into two at some point within a single
+	 * atom.  The atom may be split into two atoms, one at the end
+	 * of the old list, and one at the start of the new.
 	 * specified by pos, discarding the character at pos.
 	 * return the newly created atom as the head of a
 	 * new list, separate from the old one.
 	 */
 
-	ATOM *a, *new_list = NULL;
+	char *text;
+	ATOM *prev, *new_list = NULL;
 
-	/* Check for pos lying at the start of the token */
+	/* Check for pos lying at the start of the atom */
 
-	if (pos == token->text) {
-		/* Pos is at start; token starts the new list */
+	if (pos == atom->text) {
+		/* Pos is at start; atom starts the new list */
 
-		for (a = alist; a != NULL; a = a->next) {
-			if (a->next == token) {
-				a->next = NULL;
-				return(token);
-			}
+		if ((prev = aprev(alist, atom)) == NULL) {
+			/* Can't split at the head of the list */
+
+			return(NULL);
 		}
 
-		/* Can't split with token at the head of the list */
+		/* Split the list before the atom */
 
-		return(NULL);
-	} else if (*(pos + 1) == '\0') {
-		/* Pos is at end; token ends the old list */
+		prev->next = NULL;
 
-		*pos = '\0';
-		new_list = token->next;
-	} else {
-		/* Need to split the atom */
+		/* Update the atom's text */
 
-		*pos++ = '\0';
-		new_list = add_atom(NULL, pos, token->type);
-		new_list->next = token->next;
+		text = xstrdup(atom->text + 1);
+		free(atom->text);
+		atom->text = text;
+
+		/* And return the atom */
+
+		return(atom);
 	}
 
+	/* Check for pos lying at the end of the atom */
+
+	if (*(pos + 1) == '\0') {
+		/* Pos is at end; atom ends the old list */
+
+		/* Update the atom's text */
+		*pos = '\0';
+
+		/* And split the list after atom */
+
+		new_list = atom->next;
+		atom->next = NULL;
+
+		/* Now return the new atom list */
+
+		return(new_list);
+	}
+
+	/* We need to split the atom; update the old atom's text */
+
+	*pos++ = '\0';
+
+	/* Now create the new atom list */
+
+	new_list = aappend(NULL, pos, atom->type);
+	new_list->next = atom->next;
+
 	/* Terminate the original list and return the new one */
 
-	token->next = NULL;
+	atom->next = NULL;
 	return(new_list);
 }
 /****************************************************************************/
 ATOM *acomment(alist)
 ATOM *alist;
 {
-	/* Return an atom comtaining alist as a comment */
+	/* Return an atom containing alist as a comment */
 
 	char *buf, *qbuf, *ebuf;
 	ATOM *new_list;
 
 	/* Allocate a new string and set it to the comment text */
-  
+
 	buf = atext(NULL, alist, AC_UNQUOTE);
 	qbuf = vstrcat("(", buf, ")", NULL);
-  
+
 	/* Check the text is a valid comment */
-  
+
 	if ((new_list = tokenise(qbuf)) == NULL
 	    || new_list->next != NULL) {
 		/* We need to fix parens in the comment */
@@ -555,10 +687,10 @@ ATOM *alist;
 		afree(new_list);
 		ebuf = escape_chars(buf, EC_COMMENT);
 		qbuf = vstrcat("(", ebuf, ")", NULL);
-		new_list = add_atom(NULL, qbuf, AT_QSTRING);
+		new_list = aappend(NULL, qbuf, AT_QSTRING);
 	}
-  
-  	/* Clean up and return the new list */
+
+	/* Clean up and return the new list */
 
 	free(buf);
 	free(qbuf);
@@ -579,18 +711,204 @@ ATOM *alist;
 	ebuf = escape_chars(buf, EC_QSTRING);
 	qbuf = vstrcat("\"", ebuf, "\"", NULL);
 
-	/* Make an atom list containing the string and a space */
+	/* Make an atom list containing the string */
 
-	new_list = add_atom(NULL, qbuf, AT_QSTRING);
+	new_list = aappend(NULL, qbuf, AT_QSTRING);
 
-  	/* Clean up and return the new list */
-  
-  	free(buf);
+	/* Clean up and return the new list */
+
+	free(buf);
 	free(qbuf);
 	return(new_list);
 }
 /****************************************************************************/
-static char *avalue(atom)
+ATOM *acopy(alist)
+ATOM *alist;
+{
+	/* Return an allocated copy of alist */
+
+	ATOM *atom, *new_list = NULL;
+
+	/* Simply append each token to the new list */
+
+	for (atom = alist; atom != NULL; atom = atom->next) {
+		new_list = aappend(new_list, atom->text, atom->type);
+	}
+
+	/* And return the new atom list */
+
+	return(new_list);
+}
+/****************************************************************************/
+ATOM *adiscard(alist, atom)
+ATOM *alist, *atom;
+{
+	/* Discard atom from alist and return the updated alist */
+
+	ATOM *prev;
+
+	/* Remove the atom from the list */
+
+	if ((prev = aprev(alist, atom)) != NULL) {
+		/* Update prev to point after atom */
+
+		prev->next = atom->next;
+	} else if (atom == alist) {
+		/* Update alist to point after atom */
+
+		alist = alist->next;
+	}
+
+	/* Now free the discarded atom */
+
+	free(atom->text);
+	free(atom);
+
+	/* And return the updated atom list */
+
+	return(alist);
+}
+/****************************************************************************/
+ATOM *acut(alist, start, end)
+ATOM *alist, *start, *end;
+{
+	/*
+	 * Strip atoms from start up to (but not including) end
+	 * out of alist. Returns alist as modified by the cut.
+	 */
+
+	ATOM *prev;
+
+	/* First find the atom before start */
+
+	if ((prev = aprev(alist, start)) != NULL) {
+		/* Cut from start up to end out of the list */
+
+		prev->next = end;
+	} else if (alist == start) {
+		/* Cutting at the start of the list */
+
+		alist = end;
+	}
+
+	/* Find the atom before end */
+
+	if (end != NULL && (prev = aprev(start, end)) != NULL) {
+		/* Update the end of the cut atoms */
+
+		prev->next = NULL;
+	}
+
+	/* Now return the updated atom list */
+
+	return(alist);
+}
+/****************************************************************************/
+ATOM *adelete(alist, start, end)
+ATOM *alist, *start, *end;
+{
+	/*
+	 * Delete atoms from start up to (but not including) end
+	 * from alist. Returns alist as modified by the cut.
+	 */
+
+	/* Cut the atoms and free any cut section */
+
+	if ((alist = acut(alist, start, end)) != start) {
+		afree(start);
+	}
+
+	/* Return the updated atom list */
+
+	return(alist);
+}
+/****************************************************************************/
+ATOM *aprev(alist, atom)
+ATOM *alist, *atom;
+{
+	/* Return the atom before atom in alist, or NULL if none */
+
+	ATOM *a;
+
+	/* Loop over the list looking for atom */
+
+	for (a = alist; a != NULL && a != atom; a = a->next) {
+		/* Is this the atom's previous atom? */
+
+		if (a->next == atom) {
+			return(a);
+		}
+	}
+
+	/* No previous atom found */
+
+	return(NULL);
+}
+/****************************************************************************/
+ATOM *atoken(alist)
+ATOM *alist;
+{
+	/*
+	 * Return the first atom in alist that is a "token"; ie that
+	 * isn't whitespace or a comment.  Returns NULL if none found.
+	 */
+
+	ATOM *atom;
+
+	/* Loop over all the whitespace or comment tokens */
+
+	for (atom = alist; atom != NULL && IS_WS(atom); atom = atom->next) {
+		/* NULL LOOP */
+	}
+
+	/* Return the atom we found */
+
+	return(atom);
+}
+/****************************************************************************/
+ATOM *afind(alist)
+ATOM *alist;
+{
+	/*
+	 * Return the first atom in alist that isn't whitespace.
+	 * Returns NULL if none found.
+	 */
+
+	ATOM *atom;
+
+	/* Loop over all the whitespace atoms */
+
+	for (atom = alist; atom != NULL && atom->type == AT_WS;
+	     atom = atom->next) {
+		/* NULL LOOP */
+	}
+
+	/* Return the atom we found */
+
+	return(atom);
+}
+/****************************************************************************/
+ATOM *asearch(alist, type)
+ATOM *alist;
+int type;
+{
+	/* Return the first atom of type 'type' in alist */
+
+	ATOM *atom;
+
+	/* Loop until we find an atom of the right type */
+
+	for (atom = alist; atom != NULL && atom->type != type;
+	     atom = atom->next) {
+		/* NULL LOOP */
+	}
+
+	/* Return the atom we found */
+
+	return(atom);
+}
+/****************************************************************************/
+char *avalue(atom)
 ATOM *atom;
 {
 	/*
@@ -643,52 +961,75 @@ int canon;
 {
 	/*
 	 * Return an allocated string containing the concatenation of the
-	 * string held in buf (if non-null) and the atom list supplied.
-	 * Canon specifies the canonicalsation required: none, trim
-	 * leading and trailing white space, trim white space or full.
+	 * string held in buf (if non-null) and the atoms in alist.
+	 * Canon specifies the canonicalisation required: none, trim
+	 * leading and trailing white space, trim white space and unfold,
+	 * canonicalise white space, remove the parens around comments,
+	 * unquote quoted strings, unquote quoted strings and quote double
+	 * quotes, or fullly canonical RFC 822 form.
 	 */
 
 	size_t len = 0;
-	ATOM *a, *start, *end;
-
-	/* Return buf unchanged if alist is NULL */
-
-	if (alist == NULL) {
-		return(buf);
-	}
+	ATOM *atom, *start, *end;
 
 	/* Set the start and end positions in the list */
 
 	start = end = NULL;
-	for (a = alist; canon != AC_NONE && a != NULL; a = a->next) {
-		if (a->type != AT_WS) {
-			start = (start == NULL) ? a : start;
-			end = a->next;
+	for (atom = alist; canon != AC_NONE && atom != NULL;
+	     atom = atom->next) {
+		/* Is this white space we might trim? */
+
+		if (atom->type != AT_WS) {
+			/* Update the start and end atoms */
+
+			start = (start == NULL) ? atom : start;
+			end = atom->next;
 		}
 	}
-	start = (start == NULL) ? alist : start;
+
+	/* Default the start atom if required */
+
+	start = (start == NULL && canon == AC_NONE) ? alist : start;
+
+	/* If we don't have any atoms, there's no text either */
+
+	if (start == NULL) {
+		/* Just return a null string */
+
+		return((buf == NULL) ? xstrdup("") : buf);
+	}
 
 	/* Find out how long the text will be */
 
-	for (a = start; a != end; a = a->next) {
-		if (a->type == AT_QSTRING && canon == AC_UNQUOTE ||
-		    a->type == AT_COMMENT && canon == AC_UNCOMMENT) {
-			/* We will strip the start and end of this token */
+	for (atom = start; atom != end; atom = atom->next) {
+		if (atom->type == AT_QSTRING && canon == AC_UNQUOTE ||
+		    atom->type == AT_COMMENT && canon == AC_UNCOMMENT) {
+			/* We will strip the start and end of this atom */
 
-			len += strlen(a->text) - 2;
-		} else if (a->type == AT_WS && canon == AC_UNQUOTE) {
+			len += strlen(atom->text) - 2;
+		} else if ((atom->type == AT_QSTRING
+			    || atom->type == AT_COMMENT)
+			   && (canon == AC_WS || canon == AC_UNFOLD
+			       || canon == AC_FULL)) {
+			/* We'll add the unfolded text of the atom */
+
+			len += strlen(unfold_atom(atom));
+		} else if (atom->type == AT_WS &&
+			   (canon == AC_WS || canon == AC_UNQUOTE
+			    || canon == AC_UNFOLD &&
+			    strchr(atom->text, '\n') != NULL)) {
 			/* We'll add a single whitespace character */
 
 			len++;
-		} else if (a->type == AT_SEMI && canon == AC_FULL) {
+		} else if (atom->type == AT_SEMI && canon == AC_FULL) {
 			/* We'll be adding a space after the semi */
 
-			len += strlen(a->text) + 1;
-		} else if (canon != AC_FULL || a->type != AT_WS
-			   && a->type != AT_COMMENT) {
-			/* Add the length of this token */
+			len += strlen(atom->text) + 1;
+		} else if (canon != AC_FULL || atom->type != AT_WS
+			   && atom->type != AT_COMMENT) {
+			/* Add the length of this atom */
 
-			len += strlen(a->text);
+			len += strlen(atom->text);
 		}
 	}
 
@@ -703,30 +1044,40 @@ int canon;
 
 	/* Concatenate the list to the string */
 
-	for (a = start; a != end; a = a->next) {
-		if (a->type == AT_QSTRING && canon == AC_UNQUOTE ||
-		    a->type == AT_COMMENT && canon == AC_UNCOMMENT) {
-			/* Add the token with no delimiters */
+	for (atom = start; atom != end; atom = atom->next) {
+		if (atom->type == AT_QSTRING && canon == AC_UNQUOTE ||
+		    atom->type == AT_COMMENT && canon == AC_UNCOMMENT) {
+			/* Add the value of the atom */
 
-			(void) strcat(buf, avalue(a));
-		} else if (a->type == AT_WS && canon == AC_UNQUOTE) {
+			(void) strcat(buf, avalue(atom));
+		} else if ((atom->type == AT_QSTRING
+			    || atom->type == AT_COMMENT)
+			   && (canon == AC_WS || canon == AC_UNFOLD
+			       || canon == AC_FULL)) {
+			/* We'll add the unfolded text of the atom */
+
+			(void) strcat(buf, unfold_atom(atom));
+		} else if (atom->type == AT_WS &&
+			   (canon == AC_WS || canon == AC_UNQUOTE
+			    || canon == AC_UNFOLD &&
+			    strchr(atom->text, '\n') != NULL)) {
 			/* Add a single (canonical) space */
 
 			(void) strcat(buf, " ");
-		} else if (a->type == AT_SEMI && canon == AC_FULL) {
-			/* Add the semi foillowed by a space */
+		} else if (atom->type == AT_SEMI && canon == AC_FULL) {
+			/* Add the semi followed by a space */
 
-			(void) strcat(buf, a->text);
+			(void) strcat(buf, atom->text);
 			(void) strcat(buf, " ");
-		} else if (canon != AC_FULL || a->type != AT_WS
-			   && a->type != AT_COMMENT) {
-			/* Add the token to the buffer */
+		} else if (canon != AC_FULL || atom->type != AT_WS
+			   && atom->type != AT_COMMENT) {
+			/* Add the atom to the buffer */
 
-			(void) strcat(buf, a->text);
+			(void) strcat(buf, atom->text);
 		}
 	}
 
-	/* Return the string */
+	/* Return the string buffer */
 
 	return(buf);
 }
@@ -756,19 +1107,20 @@ char *a_strerror()
 		AERR_BASE_TEXT, AERR_BASE_TEXT, AERR_BASE_TEXT,
 		AERR_BASE_TEXT, AERR_BASE_TEXT, CERR_BASE_TEXT,
 		CERR_BASE_TEXT, CERR_BASE_TEXT, EERR_BASE_TEXT,
-		EERR_BASE_TEXT, RERR_BASE_TEXT, RERR_BASE_TEXT,
-		RERR_BASE_TEXT
+		EERR_BASE_TEXT, DERR_BASE_TEXT, DERR_BASE_TEXT,
+		DERR_BASE_TEXT, SERR_BASE_TEXT, RERR_BASE_TEXT,
+		RERR_BASE_TEXT, RERR_BASE_TEXT
 	};
 
 	/* The list of error messages */
 
 	static char *aerr_text[] = {
 		ATEXT_NONE, ATEXT_NULL, ATEXT_ADDRESS, ATEXT_BRACKET, 
-		ATEXT_ROUTE, ATEXT_LOCAL, ATEXT_DOMAIN, ATEXT_UUCP,
-		ATEXT_QSTRING, ATEXT_DLITERAL, ATEXT_COMMENT,
-		CTEXT_NULL, CTEXT_TYPE, CTEXT_PARAM, ETEXT_NULL,
-		ETEXT_ENCODING, RTEXT_TOKEN, RTEXT_BRACKET,
-		RTEXT_REFERENCE
+		ATEXT_ROUTE, ATEXT_LOCAL, ATEXT_DOMAIN, ATEXT_QSTRING,
+		ATEXT_DLITERAL, ATEXT_COMMENT, ATEXT_GROUP, CTEXT_NULL,
+		CTEXT_TYPE, CTEXT_PARAM, ETEXT_NULL, ETEXT_ENCODING,
+		DTEXT_NULL, DTEXT_DISP, DTEXT_PARAM, STEXT_CHARSET,
+		RTEXT_TOKEN, RTEXT_BRACKET, RTEXT_REFERENCE
 	};
 
 	/* A static buffer for the error message */
@@ -900,3 +1252,31 @@ char *text, *echars;
 	return(buf);
 }
 /****************************************************************************/
+static char *unfold_atom(atom)
+ATOM *atom;
+{
+	static char *buf = NULL;
+	ATOM *alist;
+
+	/* Free any old return buffer */
+
+	if (buf != NULL) {
+		free(buf);
+		buf = NULL;
+	}
+
+	/* Atomise the atom's text and check status */
+
+	if ((alist = atomise(atom->text, get_watom)) != NULL) {
+		/* Extract the unfolded text of the atom */
+
+		buf = atext(NULL, alist, AC_UNFOLD);
+		afree(alist);
+	}
+
+	/* Now return the unfolded text */
+
+	return(buf);
+}
+/****************************************************************************/
+
