@@ -1,5 +1,5 @@
 /* Buffer.c - Buffer handling for af.
-   Copyright (C) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1990 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,7 +30,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: buffer.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: buffer.c,v 2.3 2002/08/21 23:54:48 malc Exp $";
 #endif /* ! lint */
 
 /****************************************************************************/
@@ -46,17 +46,23 @@ extern long atol(), filesize();
 extern void free(), msg(), msgl(), cmsg();
 extern void typeout(), show_new_buffer();
 extern void set_sys_tags(), toggle_mode();
+extern void free_text();
 extern MESSAGE *get_messages(), *null_msg();
 extern MESSAGE *find_start(), *find_prev();
 extern MESSAGE *find_next();
 extern CLIST *add_clist();
 
+#ifdef READ_VIA_POP3
+extern void kill_pop3();
+#endif /* READ_VIA_POP3 */
+
 /* Local function declarations */
 
+char *unique_name();
 int sync_buffer(), new_messages();
 unsigned count_messages();
 unsigned count_bad_messages();
-void free_messages(), free_mtext();
+void free_messages();
 void rm_buffer();
 static char *buf_name();
 static int read_buffer();
@@ -84,38 +90,38 @@ unsigned modes;
 	 * The buffer is inserted before the current buffer.
 	 */
 
-	char *newnam;
-	MAILBUF *newbuf;
-	
-	/* Set the buffer name and generate a buffer */
+	char *name;
+	MAILBUF *buf;
 
-	newnam = (bufnam != NULL) ? xstrdup(bufnam) : buf_name(list, filnam);
-	newbuf = new_buffer(newnam, filnam, pending, modes);
-	free(newnam);
+	/* Get the buffer name and generate a buffer */
+
+	name = (bufnam != NULL) ? xstrdup(bufnam) : buf_name(list, filnam);
+	buf = new_buffer(name, filnam, pending, modes);
+	free(name);
 
 	/* Read any file associated with the buffer */
 
-	if (filnam != NULL && !read_buffer(newbuf)) {
-		rm_buffer(newbuf);
+	if (filnam != NULL && !read_buffer(buf)) {
+		rm_buffer(buf);
 		return(list);
 	} else if (filnam == NULL) {
-		newbuf->messages = newbuf->point = null_msg(NULL);
+		buf->messages = buf->point = null_msg(NULL);
 	}
 
 	/* Add the buffer to the list */
 
 	if (list == NULL) {
-		newbuf->prev = newbuf->next = newbuf;
+		buf->prev = buf->next = buf;
 	} else {
-		list->prev->next = newbuf;
-		newbuf->prev = list->prev;
-		list->prev = newbuf;
-		newbuf->next = list;
+		list->prev->next = buf;
+		buf->prev = list->prev;
+		list->prev = buf;
+		buf->next = list;
 	}
 
 	/* Return the new buffer */
 
-	return(newbuf);
+	return(buf);
 }
 /****************************************************************************/
 int buf_pending(buf, filnam)
@@ -187,6 +193,14 @@ MAILBUF *buf;
 	buf->next->prev = buf->prev;
 	buf->prev->next = buf->next;
 
+#ifdef READ_VIA_POP3
+	/* Close any related POP3 connection */
+
+	if (buf->file != NULL && (buf->modes & M_POP3)) {
+		kill_pop3(buf->file);
+	}
+#endif /* READ_VIA_POP3 */	
+
 	/* And free the space it uses */
 
 	free(buf->name);
@@ -288,15 +302,13 @@ MESSAGE *list;
 	m->next = next;
 	next->prev = m;
 
-	/* Set the mark at the first new message */
+	/* That's all folks */
 
-	win->mark = list;
 	return;
 }
 /****************************************************************************/
-int sync_buffer(buf, silent)
+int sync_buffer(buf)
 MAILBUF *buf;
-int silent;
 {
 	/* Make a buffer identical to its file */
 
@@ -434,7 +446,7 @@ MAILBUF *buf;
 	buf->peofpos = (pstatus == WM_OK) ? 0L : buf->peofpos;
 
 	/* Resynchronise or reread the buffer if required */
-  
+
 	if ((status == WM_REREAD || pstatus == WM_REREAD)
 	    && reread_buffer(buf)) {
 		/* Confirm the reread */
@@ -442,7 +454,7 @@ MAILBUF *buf;
 		msgl("Reread ", utos(count_messages(buf->messages, FALSE)),
 		     " messages to resynchronise with mailbox", NULL);
 	} else if ((status == WM_RESYNC || pstatus == WM_RESYNC)
-		   && sync_buffer(buf, FALSE)) {
+		   && sync_buffer(buf)) {
 		/* Confirm the resync */
 
 		msgl("(New mail has arrived in ",
@@ -483,51 +495,57 @@ MAILBUF *list;
 	return(nbuffers);
 }
 /****************************************************************************/
-static char *buf_name(list, folder)
+char *unique_name(list, name)
 MAILBUF *list;
-char *folder;
+char *name;
 {
-	/* Return a unique buffer name based on folder */
+	/* Return a unique buffer name based on name */
 
-	char *p, *basenam, *bufnam = NULL;
-	unsigned buf_ext = 1;
+	char *bufnam = NULL;
+	int ext = 1;
 
-	/*
-	 * Set the buffer name.  This defaults to the last component
-	 * of the pathname used to open the file.
-	 */
-
-	if ((p = strrchr(folder, DIRSEP)) != NULL) {
-		basenam = p + 1;
-	} else {
-		basenam = folder;
-	}
-
-	/* If the list is NULL then that's it */
+	/* If the list is NULL then name will do */
 
 	if (list == NULL) {
-		return(xstrdup(basenam));
+		return(xstrdup(name));
 	}
 
-	/* Loop until the buffer name is unique */
+	/* Loop over the buffers looking for a name */
 
 	do {
 		/* Select the name to try */
 
 		if (bufnam != NULL) {
 			free(bufnam);
-			bufnam = vstrcat(basenam, "<",
-				 utos(buf_ext++), ">", NULL);
+			bufnam = vstrcat(name, "<", utos(ext++), ">", NULL);
 		} else {
-			bufnam = xstrdup(basenam);
+			bufnam = xstrdup(name);
 		}
 	} while (find_buffer(list, bufnam) != NULL);
-					
+
 	/* Return the unique buffer name */
 
 	return(bufnam);
 }
 /****************************************************************************/
+static char *buf_name(list, folder)
+MAILBUF *list;
+char *folder;
+{
+	/* Return a unique buffer name based on folder */
+
+	char *slash, *name;
+
+	/* Get the last component of the folder name */
+
+	name = ((slash = strrchr(folder, DIRSEP)) != NULL)
+		? slash + 1 : folder;
+
+	/* Return a unique name based on the base name */
+
+	return(unique_name(list, name));
+}
+/****************************************************************************/
 static MAILBUF *new_buffer(bufnam, filnam, pending, modes)
 char *bufnam, *filnam, *pending;
 unsigned modes;
@@ -771,7 +789,7 @@ MESSAGE *message, *original;
 	/* Return TRUE if two messages are identical */
 
 	int headers = TRUE;
-	MSG_TEXT *t, *u;
+	TEXTLINE *t, *u;
 
 	/* If we have an ID for both messages then just compare them */
 
@@ -796,10 +814,11 @@ MESSAGE *message, *original;
 		/* Skip From and status lines in the message */
 
 		while (headers &&
-		       (!strncmp(t->line, MFROM, strlen(MFROM)) ||
-			!strncasecmp(t->line, STATUS, strlen(STATUS)) ||
-			!strncasecmp(t->line, X_STATUS, strlen(X_STATUS)) ||
-			!strncasecmp(t->line, AFSTATUS, strlen(AFSTATUS)))) {
+		       (!strncmp(t->line, MFROM, strlen(MFROM))
+			|| !strncasecmp(t->line, STATUS, strlen(STATUS))
+			|| !strncasecmp(t->line, X_STATUS, strlen(X_STATUS))
+			|| !strncasecmp(t->line, X_AFSTATUS,
+					strlen(X_AFSTATUS)))) {
 			/* Skip the From or status line */
 
 			t = t->next;
@@ -808,10 +827,11 @@ MESSAGE *message, *original;
 		/* Skip From and status lines in the original */
 
 		while (headers &&
-		       (!strncmp(u->line, MFROM, strlen(MFROM)) ||
-			!strncasecmp(u->line, STATUS, strlen(STATUS)) ||
-			!strncasecmp(u->line, X_STATUS, strlen(X_STATUS)) ||
-			!strncasecmp(u->line, AFSTATUS, strlen(AFSTATUS)))) {
+		       (!strncmp(u->line, MFROM, strlen(MFROM))
+			|| !strncasecmp(u->line, STATUS, strlen(STATUS))
+			|| !strncasecmp(u->line, X_STATUS, strlen(X_STATUS))
+			|| !strncasecmp(u->line, X_AFSTATUS,
+					strlen(X_AFSTATUS)))) {
 			/* Skip the From or status line */
 
 			u = u->next;
@@ -874,6 +894,15 @@ MESSAGE *list;
 		if (list->encoding != NULL) {
 			free(list->encoding);
 		}
+		if (list->charset != NULL) {
+			free(list->charset);
+		}
+		if (list->description != NULL) {
+			free(list->description);
+		}
+		if (list->filename != NULL) {
+			free(list->filename);
+		}
 		if (list->date != NULL) {
 			free(list->date);
 		}
@@ -891,25 +920,12 @@ MESSAGE *list;
 		if (list->user_tags != NULL) {
 			free(list->user_tags);
 		}
-		free_mtext(list->text);
+		free_text(list->text);
 		free(list);
 	}
 	return;
 }
 /****************************************************************************/
-void free_mtext(text)
-MSG_TEXT *text;
-{
-	/* Free the text of a message */
-
-	if (text != NULL) {
-		free_mtext(text->next);
-		free(text->line);
-		free(text);
-	}
-	return;
-}
-/****************************************************************************/
 unsigned count_messages(list, all)
 MESSAGE *list;
 int all;
