@@ -1,5 +1,5 @@
 /* Commands.c - Command handling for af.
-   Copyright (C) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1990 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -45,7 +45,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: commands.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: commands.c,v 2.2 2002/08/21 23:54:48 malc Exp $";
 static char *CmdId = COMMANDID;
 #endif /* ! lint */
 
@@ -65,7 +65,7 @@ extern void kill_macros(), list_macros();
 extern void unget_key(), typeout(), table();
 extern void msg(), msgl(), emsg(), emsgl();
 extern void cmsg(), clearmsg(), hide_cursor();
-extern void redisplay();
+extern void alldisplay();
 extern KEYMAP *find_keymap(), *mode_keymap();
 extern BINDING *find_binding();
 extern KEYSEQ *new_seq(), *add_seq();
@@ -250,6 +250,20 @@ FORM *exec_type_key()
 	return(exec_key(NULL, NULL, NULL, NULL));
 }
 /****************************************************************************/
+FORM *exec_comp_key()
+{
+	/* Handle compose mode key-to-command parsing */
+
+	/* Initialise key sequence echoing */
+
+	echodelay = get_vval(V_ECHO);
+	echoing = FALSE;
+
+	/* Now scan the compose keymaps for the next keypress */
+
+	return(exec_key(NULL, NULL, NULL, NULL));
+}
+/****************************************************************************/
 /*ARGSUSED*/
 FORM *exec_command(seq, arg, forms)
 KEYSEQ *seq;
@@ -375,8 +389,7 @@ FORM *forms;
 	/* And return quit or error */
 
 	return((seq != NULL && seq->len < 2 &&
-		(cmodes(0) & (M_MBUF | M_TYPEOUT)))
-	       ? NULL : c_errored());
+		(cmodes(0) & (M_TYPEOUT | M_MBUF))) ? NULL : c_errored());
 }
 /****************************************************************************/
 static FORM *exec_key(seq, arg, map, gmap)
@@ -960,7 +973,6 @@ int st;
 	/* Resynchronize all buffers on alarm signal */
 
 	MAILBUF *buf = cwin->buf;
-	WINDOW *win;
 
 	/* Ignore alarm signals while we resync */
 
@@ -969,19 +981,10 @@ int st;
 	/* Loop through all existing buffers */
 
 	do {
-		if (buf->file != NULL && sync_buffer(buf, TRUE)) {
-			/* Update all windows displaying the buffer */
-
-			win = cwin;
-			do {
-				if (win->buf == buf) {
-					redisplay(win);
-				}
-				win = win->next;
-			} while (win != cwin);
-
-			/* Announce the arraival of new mail */
+		if (buf->file != NULL && sync_buffer(buf)) {
+			/* Update the display and announce the mail */
 
+			alldisplay(buf);
 			msgl("(New mail has arrived in ",
 			     buf->name, ")", NULL);
 		}
@@ -990,7 +993,7 @@ int st;
 
 	/* Update the display */
 
-	vtupdate(FALSE);
+	(void) vtupdate(FALSE);
 	hide_cursor();
 
 	/* Reset the SIGALRM handler and timer */
@@ -1015,7 +1018,7 @@ int st;
 	/* Force the screen display to be up to date */
 
 	vtredraw();
-	vtupdate(TRUE);
+	(void) vtupdate(TRUE);
 	msg("(Window size changed)");
 
 	/* Reset the SIGWINCH handler */
