@@ -1,5 +1,5 @@
 /* Date.c - Date reading and display for af.
-   Copyright (C) 1992, 1993, 1994, 1995, 1996 Malc Arnold.
+   Copyright (C) 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,7 +29,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: date.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: date.c,v 2.2 1997/09/07 15:41:14 malc Exp $";
 #endif /* ! lint */
 
 /****************************************************************************/
@@ -41,7 +41,7 @@ extern pid_t getpid();
 extern void afree();
 extern time_t time();
 extern struct tm *localtime(), *gmtime();
-extern ATOM *tokenise(), *find_token();
+extern ATOM *tokenise(), *atoken();
 
 #ifdef HAVE_TZSET
 extern void tzset();
@@ -51,7 +51,7 @@ extern void tzset();
 
 static int day_of_week(), month(), atomval();
 static long tz_offset();
-static DATEZONE *sysdate(), *posixdate(), *dateval();
+static DATEZONE *sysdate(), *rfc822date(), *dateval();
 static ATOM *get_zone();
 
 /****************************************************************************/
@@ -64,11 +64,11 @@ char *date;
 	 * epoch (1 Jan 70), and the timezone offset in seconds.
 	 * A return of NULL indicates an error.
 	 *
-	 * The date may be in POSIX or system form, and may contain
+	 * The date may be in rfc822 or system form, and may contain
 	 * comments or arbitrary white space.
 	 */
 
-	int is_posix = FALSE;
+	int is_rfc822 = FALSE;
 	DATEZONE *date_val = NULL;
 	ATOM *alist, *start;
 
@@ -80,7 +80,7 @@ char *date;
 
 	/* Find the first token in the list */
 
-	if ((start = find_token(alist)) == NULL) {
+	if ((start = atoken(alist)) == NULL) {
 		afree(alist);
 		return(NULL);
 	}
@@ -90,34 +90,34 @@ char *date;
 	if (day_of_week(start) >= 0) {
 		/* Skip the day of week, checking for errors */
 
-		if ((start = find_token(start->next)) == NULL) {
+		if ((start = atoken(start->next)) == NULL) {
 			afree(alist);
 			return(NULL);
 		}
 
-		/* Skip the comma which follows in POSIX dates */
+		/* Skip the comma which follows in rfc822 dates */
 
 		if (start->type == AT_COMMA) {
 			/* Skip the comma, checking for errors */
 
-			if ((start = find_token(start->next)) == NULL) {
+			if ((start = atoken(start->next)) == NULL) {
 				afree(alist);
 				return(NULL);
 			}
 
-			/* This must be a POSIX date */
+			/* This must be a rfc822 date */
 
-			is_posix = TRUE;
+			is_rfc822 = TRUE;
 		}
 	}
 
-	/* Check if the date is POSIX or system format */
+	/* Check if the date is rfc822 or system format */
 
-	is_posix = (is_posix || month(start) < 0);
+	is_rfc822 = (is_rfc822 || month(start) < 0);
 
 	/* Parse the date according to its type */
 
-	date_val = (is_posix) ? posixdate(start) : sysdate(start);
+	date_val = (is_rfc822) ? rfc822date(start) : sysdate(start);
 
 	/* Free the atom list and return the date value */
 
@@ -211,15 +211,15 @@ ATOM *token;
 	return(val);
 }
 /****************************************************************************/
-static DATEZONE *posixdate(start)
+static DATEZONE *rfc822date(start)
 ATOM *start;
 {
 	/*
-	 * Read a POSIX format date.  The general format is
+	 * Read a rfc822 format date.  The general format is
 	 * dd Month yy hh:mm[:ss] zone
 	 */
 
-	int day, mo, year, hour, min, sec;
+	int day, mo, year, hour, min, sec = 0;
 	long zone;
 	ATOM *a = start;
 
@@ -231,49 +231,47 @@ ATOM *start;
 
 	/* Get the month */
 
-	if ((a = find_token(a->next)) == NULL || (mo = month(a)) < 0) {
+	if ((a = atoken(a->next)) == NULL || (mo = month(a)) < 0) {
 		return(NULL);
 	}
 
 	/* Get the year */
 
-	if ((a = find_token(a->next)) == NULL || (year = atomval(a)) < 0) {
+	if ((a = atoken(a->next)) == NULL || (year = atomval(a)) < 0) {
 		return(NULL);
 	}
 
 	/* Get the hour */
 
-	if ((a = find_token(a->next)) == NULL || (hour = atomval(a)) < 0) {
+	if ((a = atoken(a->next)) == NULL || (hour = atomval(a)) < 0) {
 		return(NULL);
 	}
 
 	/* Should be a colon after the hour */
 
-	if ((a = find_token(a->next)) == NULL || a->type != AT_COLON) {
+	if ((a = atoken(a->next)) == NULL || a->type != AT_COLON) {
 		return(NULL);
 	}
 
 	/* Get the minute */
 
-	if ((a = find_token(a->next)) == NULL || (min = atomval(a)) < 0) {
+	if ((a = atoken(a->next)) == NULL || (min = atomval(a)) < 0) {
 		return(NULL);
 	}
 
 	/* Get the next token */
 
-	if ((a = find_token(a->next)) != NULL && a->type == AT_COLON) {
+	if ((a = atoken(a->next)) != NULL && a->type == AT_COLON) {
 		/* Get the second */
 
-		if ((a = find_token(a->next)) == NULL
+		if ((a = atoken(a->next)) == NULL
 		    || (sec = atomval(a)) < 0) {
 			return(NULL);
 		}
 
 		/* Find any atom containing the time zone */
 
-		a = find_token(a->next);
-	} else {
-		sec = 0;
+		a = atoken(a->next);
 	}
 
 	/* Set the time zone */
@@ -293,7 +291,7 @@ ATOM *start;
 	 * Month dd hh:mm:ss zone yyyy
 	 */
 
-	int day, mo, year, hour, min, sec;
+	int day, mo, year, hour, min, sec = 0;
 	long zone;
 	ATOM *a = start;
 
@@ -305,43 +303,41 @@ ATOM *start;
 
 	/* Get the day of the month */
 
-	if ((a = find_token(a->next)) == NULL || (day = atomval(a)) < 1) {
+	if ((a = atoken(a->next)) == NULL || (day = atomval(a)) < 1) {
 		return(NULL);
 	}
 
 	/* Get the hour */
 
-	if ((a = find_token(a->next)) == NULL || (hour = atomval(a)) < 0) {
+	if ((a = atoken(a->next)) == NULL || (hour = atomval(a)) < 0) {
 		return(NULL);
 	}
 
 	/* Should be a colon after the hour */
 
-	if ((a = find_token(a->next)) == NULL || a->type != AT_COLON) {
+	if ((a = atoken(a->next)) == NULL || a->type != AT_COLON) {
 		return(NULL);
 	}
 
 	/* Get the minute */
 
-	if ((a = find_token(a->next)) == NULL || (min = atomval(a)) < 0) {
+	if ((a = atoken(a->next)) == NULL || (min = atomval(a)) < 0) {
 		return(NULL);
 	}
 
 	/* Get the next token */
 
-	if ((a = find_token(a->next)) != NULL && a->type == AT_COLON) {
+	if ((a = atoken(a->next)) != NULL && a->type == AT_COLON) {
 		/* Get the second */
 
-		if ((a = find_token(a->next)) == NULL
+		if ((a = atoken(a->next)) == NULL
 		    || (sec = atomval(a)) < 0) {
 			return(NULL);
 		}
 
 		/* Find any atom containing the time zone */
 
-		a = find_token(a->next);
-	} else {
-		sec = 0;
+		a = atoken(a->next);
 	}
 
 	/* Set the time zone */
@@ -397,7 +393,7 @@ long *zone;
 
 		/* Can't have additional tokens if in this form */
 
-		return(find_token(a->next));
+		return(atoken(a->next));
 	}
 
 	/* Try looking up the time zone in the internal table */
@@ -414,9 +410,9 @@ long *zone;
 	 * kludge used in some systems, eg WET DST not WDT.
 	 */
 
-	if ((a = find_token(a->next)) != NULL && IS_DST(a)) {
+	if ((a = atoken(a->next)) != NULL && IS_DST(a)) {
 		SET_DST(*zone);
-		return(find_token(a->next));
+		return(atoken(a->next));
 	}
 
 	/* There wasn't an extra atom used to specify DST */
@@ -436,21 +432,21 @@ long zone;
 
 	/* Handle years specified as two digits */
 
-	if (year < 100) {
-		/* Update the year */
+	if (year < 100 && (year = year + BASE_YEAR) < EPOCH_YEAR) {
+		/* Handle a date after the year 1999 */
 
-		year += BASE_YEAR;
-
-		/* And make sure it's valid */
-
-		if (year < EPOCH_YEAR) {
-			year += 100;
-		}
+		year += 100;
 	}
 
-	/* Check the year */
+	/* Check the date is a real date */
+
+	if (year < EPOCH_YEAR || mo > NO_MONTHS
+	    || ((!IS_LEAP_YEAR(year) || !IS_LEAP_MONTH(mo))
+		&& day > mo_days[mo]) || day > mo_days[mo] + 1
+	    || hour >= HOURS_PER_DAY || min >= MINUTES_PER_HOUR
+	    || sec >= SECONDS_PER_MINUTE) {
+		/* This isn't a sensible date */
 
-	if (year < EPOCH_YEAR) {
 		return(NULL);
 	}
 
@@ -463,12 +459,6 @@ long zone;
 		}
 	}
 
-	/* Check the month */
-
-	if (mo > NO_MONTHS) {
-		return(NULL);
-	}
-
 	/* Add seconds from start of year to start of month */
 
 	for (i = 0; i < mo; i++) {
@@ -478,24 +468,10 @@ long zone;
 		}
 	}
 
-	/* Check the day of the month */
-
-	if ((!IS_LEAP_YEAR(year) || !IS_LEAP_MONTH(mo))
-	    && day > mo_days[mo] || day > mo_days[mo] + 1) {
-		return(NULL);
-	}
-
 	/* Add seconds from start of month to start of day */
 
 	dval += SECONDS_PER_DAY * (day - 1);
 
-	/* Check the hour, minute and second */
-
-	if (hour >= HOURS_PER_DAY || min >= MINUTES_PER_HOUR
-	    || sec >= SECONDS_PER_MINUTE) {
-		return(NULL);
-	}
-
 	/* Add seconds from start of day to specified time */
 
 	dval += SECONDS_PER_HOUR * (hour);
@@ -521,7 +497,7 @@ DATEZONE *date;
 int local;
 {
 	/*
-	 * Return a static string containing the POSIX date which
+	 * Return a static string containing the rfc822 date which
 	 * is given by date.  The date is shown in local time if
 	 * local is TRUE, or in the original time zone otherwise.
 	 */
@@ -532,7 +508,8 @@ int local;
 
 	long offset;
 	time_t dval;
-	int tzhours, tzmins, tzsign;
+	int tzhours, tzmins;
+	int tzsign;
 	struct tm *tval;
 
 	/* Check the date is valid */
@@ -570,8 +547,9 @@ int local;
 
 	(void) sprintf(dstr, "%s, %d %s %d %02d:%02d %c%02d%02d",
 		       weekdays[tval->tm_wday], tval->tm_mday,
-		       months[tval->tm_mon], tval->tm_year % 100,
-		       tval->tm_hour, tval->tm_min, tzsign, tzhours, tzmins);
+		       months[tval->tm_mon], tval->tm_year + BASE_YEAR,
+		       tval->tm_hour, tval->tm_min,
+		       tzsign, tzhours, tzmins);
 	return(dstr);
 }
 /****************************************************************************/
@@ -634,8 +612,8 @@ int local, with_time;
 			       months[tval->tm_mon]);
 	} else if (with_time) {
 		(void) sprintf(dstr, "%02d %s  %4d", tval->tm_mday,
-			       months[tval->tm_mon], tval->tm_year
-			       + BASE_YEAR);
+			       months[tval->tm_mon],
+			       tval->tm_year + BASE_YEAR);
 	} else {
 		(void) sprintf(dstr, "%s %02d", months[tval->tm_mon],
 			       tval->tm_year % 100);
