@@ -1,5 +1,5 @@
 /* Display.c - High-level screen handling functions for af.
-   Copyright (C) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1990 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,13 +24,14 @@
 #include "functions.h"
 #include "variable.h"
 #include "mode.h"
+#include "typeout.h"
 #include STRING_HDR
 
 /****************************************************************************/
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: display.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: display.c,v 2.3 2002/08/21 23:54:48 malc Exp $";
 #endif /* ! lint */
 
 /****************************************************************************/
@@ -40,7 +41,7 @@ extern char *xmalloc(), *xrealloc();
 extern char *xstrdup(), *get_vtext();
 extern int loading(), executing();
 extern int keys_pending(), get_vval();
-extern int tlines(), tcols();
+extern int tlines(), tcols(), active();
 extern unsigned cmodes();
 extern void free(), unformat(), tbeep();
 extern void emsgl(), tmove(), tclrline();
@@ -395,7 +396,10 @@ WINDOW *win;
 void alldisplay(buf)
 MAILBUF *buf;
 {
-	/* Update all windows displaying buf */
+	/*
+	 * Update all the active windows, or just the windows
+	 * displaying buf, if buf is non-null.
+	 */
 
 	WINDOW *w = cwin;
 
@@ -404,7 +408,7 @@ MAILBUF *buf;
 	do {
 		/* Update if required */
 
-		if (w->buf == buf) {
+		if (buf == NULL || w->buf == buf) {
 			redisplay(w);
 		}
 
@@ -531,17 +535,17 @@ int vtrow;
 }
 /****************************************************************************/
 void type_display(first)
-MSG_TEXT *first;
+TYPE_LINE *first;
 {
 	/* Display a screen of typeout lines */
 
 	int line = 0;
-	MSG_TEXT *t;
+	TYPE_LINE *l;
 
 	/* Show the lines of the typeout */
 
-	for (t = first; t != NULL && line < vt_lines - 1; t = t->next) {
-		vtnewline(line++, xstrdup(t->line));
+	for (l = first; l != NULL && line < vt_lines - 1; l = l->next) {
+		vtnewline(line++, xstrdup(l->line));
 	}
 
 	/* Clear any remaining lines */
@@ -567,7 +571,9 @@ int vtrow;
 
 	/* Set up the line format and allocate space for the line */
 
-	format = get_vtext(V_HDRLINE);
+	format = get_vtext((active(win->buf, M_BODY_PARTS)) ? V_BODYLINE :
+			   (active(win->buf, M_COMPOSE)) ? V_COMPLINE :
+			   V_HDRLINE);
 	line = xmalloc(win->cols);
 
 	/* Now expand the format string into line */
