@@ -1,5 +1,5 @@
 /* File_cmd.c - File-handling commands for af.
-   Copyright (C) 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1991 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,6 +17,7 @@
 
 
 #include <stdio.h>
+#include <errno.h>
 #include "af.h"
 #include "keyseq.h"
 #include "functions.h"
@@ -34,7 +35,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: file_cmd.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: file_cmd.c,v 2.3 2002/09/08 21:26:22 malc Exp $";
 #endif /* ! lint */
 
 /****************************************************************************/
@@ -171,7 +172,7 @@ FORM *forms;
 
 		/* Check for a quit or the save failing */
 
-		if (!user_quit && status && !do_save(cwin, cwin->buf)) {
+		if (!user_quit && status && !do_save(cwin->buf)) {
 			return(c_errored());
 		}
 	}
@@ -357,9 +358,10 @@ FORM *forms;
 	free(m->next);
 	m->next = NULL;
 
-	/* Do the insert and update the display */
+	/* Do the insert, set the mark, and update the display */
 
 	insert(cwin, new_msgs);
+	cwin->mark = new_msgs;
 	cwin->buf->mod = TRUE;
 	alldisplay(cwin->buf);
 
@@ -391,7 +393,7 @@ FORM *forms;
 
 	msgl("Resynchronising ", cwin->buf->file, "...", NULL);
 
-	if (sync_buffer(cwin->buf, FALSE)) {
+	if (sync_buffer(cwin->buf)) {
 		/* Update the display and go to the first new message */
 
 		alldisplay(cwin->buf);
@@ -457,7 +459,7 @@ FORM *forms;
 
 	/* Save the file and return status */
 
-	return((do_save(cwin, cwin->buf)) ? c_t() : c_errored());
+	return((do_save(cwin->buf)) ? c_t() : c_errored());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -525,7 +527,7 @@ FORM *forms;
 
 	/* Write the file and check status */
 
-	if (!do_save(cwin, cwin->buf)) {
+	if (!do_save(cwin->buf)) {
 		/* Restore the old file and pending */
 
 		free(cwin->buf->file);
@@ -590,7 +592,7 @@ FORM *forms;
 
 			/* Check for a quit or the save failing */
 
-			if (user_quit || status && !do_save(cwin, buf)) {
+			if (user_quit || status && !do_save(buf)) {
 				return(c_errored());
 			}
 		}
@@ -631,7 +633,7 @@ FORM *forms;
 
 			needed_save = TRUE;
 
-			if (!do_save(cwin, buf)) {
+			if (!do_save(buf)) {
 				return(c_errored());
 			}
 		}
@@ -668,7 +670,7 @@ FORM *forms;
 		return(c_errored());
 	}
 
-	/* Expand + and ~ in the file name */
+	/* Expand escape chars in the file name */
 
 	filnam = expand(filnam);
 
@@ -826,35 +828,34 @@ char *bufnam, *filnam;
 	return(TRUE);
 }
 /****************************************************************************/
-static int do_save(win, buf)
-WINDOW *win;
+static int do_save(buf)
 MAILBUF *buf;
 {
 	/* Handle saving a buffer to disk */
 
 	int status, msg_no = 1;
 	int killed = FALSE;
-	MESSAGE *msg, *next;
+	MESSAGE *message, *next;
 
 #ifdef READ_VIA_POP3
 	MESSAGE *m;
 #endif /* READ_VIA_POP3 */
 
-       	/* Write the buffer, checking for failure */
+	/* Write the buffer, checking for failure */
 
 	status = write_buffer(buf);
 
 	/* Kill any deleted messages and update positions */
 
-	msg = buf->messages;
-	while (status && msg != NULL && msg->text != NULL) {
+	message = buf->messages;
+	while (status && message != NULL && message->text != NULL) {
 		/* Get the next message in the list */
 
-		next = msg->next;
+		next = message->next;
 
 		/* Kill or renumber this message as required */
 
-		if (msg->deleted) {
+		if (message->deleted) {
 			/* Do we need to set up a new kill buffer? */
 
 			if (!killed) {
@@ -868,13 +869,13 @@ MAILBUF *buf;
 			     && POP3_MBOX(buf->file); m = m->next) {
 				/* Renumber this message */
 
-				m->pos -= (m->pos > msg->pos) ? 1 : 0;
+				m->pos -= (m->pos > message->pos) ? 1 : 0;
 			}
 #endif /* READ_VIA_POP3 */
 
 			/* Now kill the message */
 
-			kill_message(buf, msg, FALSE);
+			kill_message(buf, message, FALSE);
 #ifdef READ_VIA_POP3
 		} else if (!POP3_MBOX(buf->file)) {
 #else /* ! READ_VIA_POP3 */
@@ -882,12 +883,12 @@ MAILBUF *buf;
 #endif /* ! READ_VIA_POP3 */
 			/* Renumber this message */
 
-			msg->pos = msg_no++;
+			message->pos = msg_no++;
 		}
 
 		/* And move on to the next message */
 
-		msg = next;
+		message = next;
 	}
 
 	/* Update the screen and return status */
