@@ -1,5 +1,5 @@
 /* Files.c - File handling for af.
-   Copyright (C) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1990 - 2003 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 #include "complete.h"
 #include "io.h"
 #include "misc.h"
+#include "mime.h"
 #include STRING_HDR
 #include DIR_HDR
 
@@ -41,18 +42,21 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: files.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: files.c,v 2.6 2003/10/27 23:16:20 malc Exp $";
 #endif /* ! lint */
 
 /****************************************************************************/
 /* Global function declarations */
 
 extern char *xmalloc(), *xrealloc(), *xstrdup(), *vstrcat();
-extern char *getenv(), *strerror(), *strudate(), *get_user();
-extern char *get_home(), *get_pwd(), *get_mailbox();
-extern char *get_vtext(), *get_line(), *utos();
+extern char *getenv(), *strerror(), *strudate(), *utos();
+extern char *get_user(), *get_home(), *get_pwd(), *get_vtext();
+extern char *get_mailbox(), *typeonly(), *decode_header_line();
+extern char *get_line();
 extern int readlink(), strncasecmp(), get_vval(), listed();
-extern int is_header(), is_blank(), close_folder();
+extern int is_header(), is_blank(), close_folder(), touch_folder();
+extern int viewable_ctype(), viewable_charset();
+extern int write_decoded_text();
 extern void free(), msgl(), emsgl(), cmsg(), abort_folder();
 extern FILE *open_folder();
 extern DATEZONE *date_now();
@@ -67,13 +71,10 @@ extern long size_pop3();
 /* Local function declarations */
 
 char *expand();
-int mmdf_form(), nonempty();
-int write_text();
+int write_text(), mmdf_form(), nonempty();
 long filesize();
 static char *dir_tail();
-static int afstatus(), aftags();
-static int at_digest_end();
-static MSG_TEXT *next_digest_entry();
+static int mime_warning(), afstatus(), aftags();
 
 #ifdef MTA_CONTENT_LENGTH
 static int conlength();
@@ -84,11 +85,6 @@ static int conlength();
 
 extern int errno;
 
-/****************************************************************************/
-/* The next line of a mail digest to be parsed */
-
-static MSG_TEXT *digest_text = NULL;
-
 /****************************************************************************/
 int write_messages(list, folder, eofpos)
 MESSAGE *list;
@@ -101,7 +97,7 @@ long eofpos;
 	 * WM_FAILED on failure, or WM_RESYNC if the file has grown.
 	 */
 
-	int mmdf, update, status;
+	int fmt, update, status;
 	unsigned msg_no = 0;
 	MESSAGE *m;
 	FILE *fp;
@@ -126,9 +122,9 @@ long eofpos;
 		return(WM_RESYNC);
 	}
 
-	/* Check if the file is in MMDF format */
+	/* Set the output format for the messages */
 
-	mmdf = mmdf_form(folder);
+	fmt = (mmdf_form(folder)) ? WF_MBOX | WF_MMDF : WF_MBOX;
 
 	/* Open the file for writing */
 
@@ -159,7 +155,7 @@ long eofpos;
 
 		/* Write the message */
 
-		if (status = write_text(fp, m, NULL, NULL, HS_MBOX, mmdf)) {
+		if (status = write_text(fp, m, fmt, NULL)) {
 			emsgl("Error writing ", folder, ": ",
 			      strerror(status), NULL);
 			abort_folder(folder, fp);
@@ -173,6 +169,10 @@ long eofpos;
 		return(WM_FAILED);
 	}
 
+	/* Update timestamps on the folder (for mail watchers) */
+
+	(void) touch_folder(folder);
+
 	/* Confirm the write and return success */
 
 	if (list != NULL) {
@@ -182,25 +182,31 @@ long eofpos;
 	return(WM_OK);
 }
 /****************************************************************************/
-int write_text(fp, message, prefix, hdrlist, skip_hdrs, mmdf)
+int write_text(fp, message, fmt, hdrlist)
 FILE *fp;
 MESSAGE *message;
-char *prefix, *hdrlist;
-int skip_hdrs, mmdf;
+int fmt;
+char *hdrlist;
 {
 	/*
 	 * Write the text of the message to the file indicated
-	 * by fp with the specified prefix.  Skip header lines
-	 * as given by skip_hdrs.
+	 * by fp with the specified prefix.  Fmt controls the
+	 * format of the output in various ways.
 	 */
 
-	int status = 0;
+	static char *mmdf_delim = MMDF_DELIM;
+	static char *mbox_delim = MFROM;
+	static char *delim_prefix = DELIM_PFX;
+	char *enc, *decoded;
+	char *delim = NULL;
 	int found = FALSE;
-	MSG_TEXT *t;
+	int status = 0;
+	int flags;
+	TEXTLINE *t;
 
 	/* Write any header required to the file */
 
-	if (mmdf && skip_hdrs == HS_MBOX && fputs(MMDF_DELIM, fp) == EOF) {
+	if ((fmt & WF_MMDF) && fputs(MMDF_DELIM, fp) == EOF) {
 		return(errno);
 	}
 
@@ -210,21 +216,28 @@ int skip_hdrs, mmdf;
 	     !is_blank(t->line); t = t->next) {
 		/* Write this header if we need to */
 
-		if (skip_hdrs != HS_ALL &&
-		    (skip_hdrs != HS_FROM ||
-		     strncmp(t->line, MFROM, strlen(MFROM))) &&
-		    (skip_hdrs != HS_COPY || listed(t->line, hdrlist)) &&
-		    (skip_hdrs != HS_SHOW || !listed(t->line, hdrlist)) &&
+		if (WF_FMT(fmt) != WF_BODY
+		    && (WF_FMT(fmt) != WF_MESG ||
+			strncmp(t->line, MFROM, strlen(MFROM)))
+		    && (WF_FMT(fmt) != WF_SOME || !listed(t->line, hdrlist))
 #ifdef MTA_CONTENT_LENGTH
-		    strncasecmp(t->line, CONTENT_LENGTH,
-				strlen(CONTENT_LENGTH)) &&
-#endif */ MTA_CONTENT_LENGTH */
-    		    strncasecmp(t->line, AFSTATUS, strlen(AFSTATUS)) &&
-		    strncasecmp(t->line, AFTAGS, strlen(AFTAGS))) {
-			/* Write the header and set found */
+		    && strncasecmp(t->line, CONTENT_LENGTH,
+				   strlen(CONTENT_LENGTH))
+#endif /* MTA_CONTENT_LENGTH */
+    		    && strncasecmp(t->line, X_AFSTATUS, strlen(X_AFSTATUS))
+		    && strncasecmp(t->line, X_AFTAGS, strlen(X_AFTAGS))) {
+			/* Determine the flags to decode the header with */
 
-			if (prefix != NULL && fputs(prefix, fp) == EOF
-			    || fputs(t->line, fp) == EOF) {
+			flags = (fmt & WF_SHOW) ? WR_SHOW | WR_FOLD : WR_NONE;
+
+			/* Decode or copy the header as required */
+
+			decoded = (!(fmt & WF_DECODE)) ? t->line
+				: decode_header_line(t->line, flags);
+
+			/* Write the (decoded) header and set found */
+
+			if (fputs(decoded, fp) == EOF) {
 				/* Error writing the header */
 
 				return(errno);
@@ -235,7 +248,7 @@ int skip_hdrs, mmdf;
 
 	/* If we're writing all headers then add status and tags */
 
-	if (skip_hdrs == HS_MBOX || skip_hdrs == HS_NONE) {
+	if (WF_FMT(fmt) == WF_MBOX || WF_FMT(fmt) == WF_TEXT) {
 #ifdef MTA_CONTENT_LENGTH
 		/* We also need to write the Content-Length header */
 
@@ -243,7 +256,6 @@ int skip_hdrs, mmdf;
 			return(status);
 		}
 #endif /* MTA_CONTENT_LENGTH */
-
 		if (status = afstatus(fp, message)) {
 			return(status);
 		}
@@ -252,42 +264,47 @@ int skip_hdrs, mmdf;
 		}
 	}
 
-	/* Skip the blank line if no headers found */
-
-	t = (!found && t != NULL) ? t->next : t;
-
-	/* Now copy the message body */
-
-	while (t != NULL && t->line != NULL) {
-#ifndef MTA_CONTENT_LENGTH
-	/* Quote the line if it begins with a message delimiter */
-
-	if (skip_hdrs == HS_MBOX && prefix == NULL &&
-	    (mmdf && !strncmp(t->line, MMDF_DELIM, strlen(MMDF_DELIM))
-	     || !mmdf && !strncmp(t->line, MFROM, strlen(MFROM)))
-	    && fputs(DELIM_PFX, fp) == EOF) {
-		/* Error writing the prefix */
+	/* Write the blank line if headers were output or warnings will be */
 
+	if (found && t != NULL && (!(fmt & WF_NOBODY) || WF_WARN(fmt))
+	    && fputs(t->line, fp) == EOF) {
 		return(errno);
 	}
+	t = (t != NULL) ? t->next : t;
+
+	/* Handle warning the users about MIME problems */
+
+	if (status = mime_warning(fp, message, fmt)) {
+		/* Error writing the warning(s) */
+
+		return(status);
+	}
+
+	/* Check if and how we want to decode the message */
+
+	enc = (fmt & WF_DECODE) ? message->encoding : NULL;
+
+#ifndef MTA_CONTENT_LENGTH
+	/* Set up the delimiter for the message body */
+
+	delim = (fmt & WF_MMDF) ? mmdf_delim
+		: (WF_FMT(fmt) == WF_MBOX) ? mbox_delim : NULL;
 #endif /* ! MTA_CONTENT_LENGTH */
 
-		/* Copy each line with any prefix */
+	/* Now write the decoded text of the message */
 
-		if (prefix != NULL && fputs(prefix, fp) == EOF
-		    || fputs(t->line, fp) == EOF) {
-			/* Error writing text */
-
-			return(errno);
-		}
-		t = t->next;
+	if (status = (fmt & WF_NOBODY) ? 0
+	    : write_decoded_text(t, fp, enc, delim,
+				 delim_prefix, message->textual)) {
+		return(status);
 	}
 
 	/* Append any footer required to the text */
 
-	if (mmdf && skip_hdrs == HS_MBOX) {
+	if (fmt & WF_MMDF) {
 		status = (fputs(MMDF_DELIM, fp) == EOF) ? errno : 0;
-	} else if (skip_hdrs != HS_COPY) {
+	} else if (!(fmt & WF_NOBLANK) &&
+		   (WF_FMT(fmt) != WF_BODY || message->textual)) {
 		status = (putc('\n', fp) == EOF) ? errno : 0;
 	}
 
@@ -296,6 +313,93 @@ int skip_hdrs, mmdf;
 	return(status);
 }
 /****************************************************************************/
+int copy_file_contents(ifp, ofp)
+FILE *ifp, *ofp;
+{
+	/* Copy the contents of ifp into ofp and return status */
+
+	char buf[BUFSIZ];
+	size_t nchars;
+
+	/* Now copy the data */
+
+	while (nchars = fread(buf, sizeof(char), BUFSIZ, ifp)) {
+		/* Write the line to the file */
+
+		if (fwrite(buf, sizeof(char), nchars, ofp) < nchars) {
+			/* Error writing the file */
+
+			return(errno);
+		}
+	}
+
+	/* Now return success */
+
+	return(0);
+}
+/****************************************************************************/
+int mime_warning(fp, message, fmt)
+FILE *fp;
+MESSAGE *message;
+int fmt;
+{
+	/* Handle warning the user about MIME messages */
+	
+	/* Warn the user about unknown multipart subtypes */
+
+	if (fmt & WF_WARN_MPART &&
+	    (fputs("(** This ", fp) == EOF
+	     || fputs((fmt & WF_BODY_PART) ? "body_part"
+		      : "message", fp) == EOF
+	     || fputs(" is marked as being of content-type ", fp) == EOF
+	     || fputs(typeonly(message->contype), fp) == EOF
+	     || fputs(" **)\n", fp) == EOF)) {
+		/* Error writing the warning */
+
+		return(errno);
+	}
+
+	/* Warn the user about text they can't view properly */
+
+	if ((fmt & WF_WARN_TEXT) &&
+	    (fputs("(** This ", fp) == EOF
+	     || fputs((fmt & WF_BODY_PART) ? "body_part"
+		      : "message", fp) == EOF
+	     || fputs(" is marked as being of content-type ", fp) == EOF
+	     || fputs(typeonly(message->contype), fp) == EOF
+	     || fputs(" **)\n", fp) == EOF)) {
+		/* Error writing the warning */
+
+		return(errno);
+	}
+
+	/* Warn about non-viewable character sets */
+
+	if ((fmt & WF_WARN_CSET) &&
+	    (fputs("(** This ", fp) == EOF
+	     || fputs((fmt & WF_BODY_PART) ? "body_part"
+		      : "message", fp) == EOF
+	     || fputs(" is in the ", fp) == EOF		    
+	     || fputs(message->charset, fp) == EOF
+	     || fputs(" character set **)\n", fp) == EOF)) {
+		/* Error writing the warning */
+
+		return(errno);
+	}
+
+	/* And separate out the warning(s) */
+
+	if (WF_WARN(fmt) && fputs("\n", fp) == EOF) {
+		/* Error writing the separator */
+
+		return(errno);
+	}
+
+	/* Successfully wrote the warnings */
+
+	return(0);
+}
+/****************************************************************************/
 #ifdef MTA_CONTENT_LENGTH
 static int conlength(fp, length)
 FILE *fp;
@@ -316,7 +420,7 @@ MESSAGE *message;
 {
 	/* Write a status header for message to fp */
 
-	if (fputs(AFSTATUS, fp) == EOF) {
+	if (fputs(X_AFSTATUS, fp) == EOF) {
 		return(errno);
 	}
 	if (putc(' ', fp) == EOF) {
@@ -354,7 +458,7 @@ MESSAGE *message;
 	/* Write a tags header for message to fp */	
 
 	if (message->user_tags != NULL) {
-		if (fputs(AFTAGS, fp) == EOF) {
+		if (fputs(X_AFTAGS, fp) == EOF) {
 			return(errno);
 		}
 		if (putc(' ', fp) == EOF) {
@@ -371,145 +475,6 @@ MESSAGE *message;
 	return(0);
 }
 /****************************************************************************/
-int open_digest(message)
-MESSAGE *message;
-{
-	/* Open message as a mail digest to be parsed for entries */
-
-	/* Initialise the current digest line */
-
-	digest_text = message->text;
-
-	/* Skip the digest's preamble */
-
-	while (digest_text != NULL &&
-	       next_digest_entry(digest_text) == NULL) {
-		digest_text = digest_text->next;
-	}
-
-	/* Return whether this is a valid digest */
-
-	return(digest_text != NULL);
-}
-/****************************************************************************/
-/*ARGSUSED*/
-char *get_digest(fp, fold)
-FILE *fp;
-int fold;
-{
-	/*
-	 * Process the lines of the message pointed to by the
-	 * static variable digest_text, expanding digest entries
-	 * into individual messages and returning each line of
-	 * the message as an allocated string, one line per call.
-	 * If fold is true then process folded lines (newlines
-	 * followed by white space are not line separators).
-	 */
-
-	char *line;
-	MSG_TEXT *next_entry;
-
-	/* Have we reached the end of an entry in the digest? */
-
-	if ((next_entry = next_digest_entry(digest_text)) != NULL) {
-		/* Move to the start of the next entry */
-
-		digest_text = next_entry;
-
-		/* And return a delimiter */
-
-		return(vstrcat(MFROM, get_user(), "@", get_vtext(V_DOMAIN),
-			       " ", strudate(date_now()), "\n", NULL));
-	}
-
-	/* Or the end of the digest itself? */
-
-	if (digest_text == NULL || at_digest_end(digest_text)) {
-		return(NULL);
-	}
-
-	/* Copy the current line of the digest */
-
-	line = xstrdup(digest_text->line);
-	digest_text = digest_text->next;
-
-	/* Handle folds in the current line */
-
-	while (fold && is_header(line) && digest_text != NULL
-	       && (digest_text->line[0] == ' ' ||
-		   digest_text->line[0] == '\t')) {
-		/* Add the next line to the current one */
-
-		line = xrealloc(line, strlen(line) +
-				strlen(digest_text->line));
-		(void) strcat(line, digest_text->line);
-
-		/* And check if further lines are folded */
-
-		digest_text = digest_text->next;
-	}
-			
-	/* Return the expanded line of the digest */
-
-	return(line);
-}
-/****************************************************************************/
-static MSG_TEXT *next_digest_entry(text)
-MSG_TEXT *text;
-{
-	/*
-	 * If text points to the end of an entry in a mail
-	 * digest then return the first line in the next
-	 * entry, otherwise return NULL.
-	 */
-
-	MSG_TEXT *t;
-
-	/* The current line must be a digest delimiter */
-
-	if (text == NULL || strncasecmp(text->line, DIGEST_DELIM,
-					strlen(DIGEST_DELIM))) {
-		return(NULL);
-	}
-	t = text->next;
-
-	/* Skip any trailing blank lines */
-
-	while (t != NULL && is_blank(t->line)) {
-		t = t->next;
-	}
-
-	/* If we now have a header, return it */
-
-	return((t != NULL && is_header(t->line)) ? t : NULL);
-}
-/****************************************************************************/
-static int at_digest_end(text)
-MSG_TEXT *text;
-{
-	/* Return TRUE if text points to the end of a digest */
-
-	MSG_TEXT *t;
-
-	/* The current line must be a digest delimiter */
-
-	if (strncasecmp(text->line, DIGEST_DELIM, strlen(DIGEST_DELIM))) {
-		return(FALSE);
-	}
-
-	/* Check for more digest entries before end-of-message */
-
-	for (t = text->next; t != NULL; t = t->next) {
-		if (next_digest_entry(t) != NULL) {
-			return(FALSE);
-		}
-	}
-
-	/* We have reached the end of the digest */
-
-	return(TRUE);
-}
-/****************************************************************************/
 long filesize(folder)
 char *folder;
 {
@@ -564,7 +529,7 @@ char *folder;
 	if (nonempty(folder) && (fp = fopen(folder, "r")) != NULL) {
 		/* Opened the file, try to get a line */
 
-		if ((line = get_line(fp, FALSE)) != NULL) {
+		if ((line = get_line(fp)) != NULL) {
 			/* Check the file type and return status */
 #ifdef MTA_MMDF_FORMAT
 			mmdf = (strncasecmp(line, MFROM, strlen(MFROM)));
@@ -670,7 +635,7 @@ char *filnam;
 	char *dirs, *dir, *newpath;
 
 #ifdef HAVE_READLINK
-	char link[MAXPATHLEN + 1];
+	char linkto[MAXPATHLEN + 1];
 	char *newdirs;
 	int len, no_links = 0;
 #endif /* HAVE_READLINK */
@@ -728,10 +693,10 @@ char *filnam;
 #ifdef HAVE_READLINK
 			/* Check whether the file is a link */
 
-			if ((len = readlink(path, link, MAXPATHLEN)) > 0) {
+			if ((len = readlink(path, linkto, MAXPATHLEN)) > 0) {
 				/* Make sure the link is null-terminated */
 
-				link[len] = '\0';
+				linkto[len] = '\0';
 
 				/* Have we reached the maximum no of links? */
 
@@ -755,14 +720,14 @@ char *filnam;
 
 				/* Remove the obsoleted section of path */
 
-				slash = (link[0] != '/') ?
+				slash = (linkto[0] != '/') ?
 					strrchr(path, '/') : path;
 				*slash = '\0';
 				path = xrealloc(path, slash - path + 1);
 
 				/* And add the link expansion to dirs */
 
-				newdirs = vstrcat(link, "/", next, NULL);
+				newdirs = vstrcat(linkto, "/", next, NULL);
 				free(dirs);
 				next = dirs = newdirs;
 			}
@@ -781,6 +746,51 @@ char *filnam;
 	return(path);
 }
 /****************************************************************************/
+char *safe_filename(filnam)
+char *filnam;
+{
+	/* Return a safe file name derived from filnam */
+
+	char *slash, *name;
+	char *safe_name, *p, *q;
+
+	/* Strip any leading path from the name */
+
+	slash = strrchr(filnam, '/');
+	name = (slash != NULL) ? slash + 1 : filnam;
+
+	/* Allocate the space for the name */
+
+	q = safe_name = xmalloc(strlen(name) + 1);
+
+	/* Now copy any safe characters across */
+
+	for (p = name; *p != '\0'; p++) {
+		/* Alphanumerics, dash, underscore and . are safe */
+
+		if (isascii(*p) && isalnum(*p) || *p == '-' || *p == '_'
+		    || q != safe_name && *p == '.') {
+			/* Copy this character across */
+
+			*q++ = *p;
+		}
+	}
+	*q = '\0';
+
+	/* Now resize the name, or free it if null */
+
+	if (q > safe_name) {
+		safe_name = xrealloc(safe_name, strlen(safe_name) + 1);
+	} else {
+		free(safe_name);
+		safe_name = NULL;
+	}
+
+	/* And return the safe name */
+
+	return(safe_name);
+}
+/****************************************************************************/
 char *pathfind(filnam, path, suffix)
 char *filnam, *path, *suffix;
 {
