@@ -1,5 +1,5 @@
 /* Format.c - Format string handling for af.
-   Copyright (C) 1992, 1993, 1994, 1995, 1996 Malc Arnold.
+   Copyright (C) 1992 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -32,19 +32,21 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: format.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: format.c,v 2.2 2002/08/21 23:54:48 malc Exp $";
 #endif /* ! lint */
 
 /****************************************************************************/
 /* Global function declarations */
 
 extern char *xmalloc(), *xrealloc(), *get_vtext();
-extern char *strkey(), *modelist(), *utos();
+extern char *strkey(), *modelist(), *decode_text();
 extern char *strqdate(), *typeonly();
+extern char *strcanon(), *utos();
 extern int get_vval(), active();
 extern int to_reading(), tlines();
 extern unsigned cmodes(), position();
 extern unsigned to_lines(), to_position();
+extern unsigned text_chars(), text_lines();
 extern void free(), emsgl(), display();
 extern MESSAGE *find_start();
 
@@ -172,6 +174,7 @@ int columns, modeline;
 			f = get_prefix(f + 1, &width);
 			conv = find_conv(*f);
 			contents = get_contents(conv, win, message, printall);
+			contents = strcanon(contents, SK_FIELD);
 
 			/* Set up the field and padding widths */
 
@@ -217,7 +220,7 @@ int columns, modeline;
 			}
 			start = end;
 
-			/* Add any padding to the line */
+		/* Add any padding to the line */
 
 			len += setfill(line + len, padding, modeline);
 			padding = 0;
@@ -294,6 +297,7 @@ int fold;
 
 			contents = get_contents(conv, NULL, message,
 						(message != NULL));
+			contents = strcanon(contents, SK_FIELD);
 
 			/* Check we can hold the field */
 
@@ -326,7 +330,6 @@ int fold;
 
 	/* Now fold the line if we have to */
 
-
 	for (f = line; fold > 0 && *f != '\0'; f++) {
 		/* Increment the width since last fold */
 
@@ -552,17 +555,7 @@ MESSAGE *message;
 {
 	/* Return the number of characters in message */
 
-	unsigned no_chars = 0;
-	MSG_TEXT *t;
-
-	/* Count the characters in the message */
-
-	if (message != NULL) {
-		for (t = message->text; t != NULL; t = t->next) {
-			no_chars += strlen(t->line) + 1;
-		}
-	}
-	return(utos(no_chars));
+	return(utos(text_chars(message->text)));
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -572,17 +565,7 @@ MESSAGE *message;
 {
 	/* Return the number of lines in the message */
 
-	unsigned no_lines = 0;
-	MSG_TEXT *t;
-
-	/* Count the lines in the message */
-
-	if (message != NULL) {
-		for (t = message->text; t != NULL; t = t->next) {
-			no_lines++;
-		}
-	}
-	return(utos(no_lines));
+	return(utos(text_lines(message->text)));
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -681,7 +664,8 @@ MESSAGE *message;
 
 	/* Return the sender if there is one */
 
-	return((message != NULL) ? message->from : nothing);
+	return((message == NULL) ? nothing :
+	       decode_text(message->from, TRUE));
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -709,7 +693,23 @@ MESSAGE *message;
 
 	/* Return the subject if there is one */
 
-	return((message != NULL) ? message->subject : nothing);
+	return((message == NULL) ? nothing :
+	       decode_text(message->subject, TRUE));
+}
+/****************************************************************************/
+/*ARGSUSED*/
+static char *m_description(win, message)
+WINDOW *win;
+MESSAGE *message;
+{
+	/* Return the content-description of the message */
+
+	static char *nothing = "";
+
+	/* Return the description if there is one */
+
+	return((message == NULL || message->description == NULL)
+	       ? nothing : decode_text(message->description, TRUE));
 }
 /****************************************************************************/
 /*ARGSUSED*/
