@@ -1,5 +1,5 @@
 /* Functions.c - Function handling for af.
-   Copyright (C) 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1995 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,9 +17,9 @@
 
 
 #include <stdio.h>
-#include <ctype.h>
 #include <regex.h>
 #include "af.h"
+#include "atom.h"
 #include "keyseq.h"
 #include "functions.h"
 #include "commands.h"
@@ -38,7 +38,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: functions.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: functions.c,v 2.3 2002/08/21 23:54:48 malc Exp $";
 static char *FuncId = FUNCTIONID;
 #endif /* ! lint */
 
@@ -47,12 +47,14 @@ static char *FuncId = FUNCTIONID;
 
 extern char *xmalloc(), *xstrdup(), *vstrcat();
 extern char *strseq(), *getenv(), *syscmd();
+extern char *atext();
 extern int fset_var(), set_key(), named_macro();
 extern int finish_startup(), startup_display();
 extern int strcasecmp(), strncasecmp();
 extern unsigned cmodes();
-extern void free(), free_seq(), msg();
-extern void emsg(), emsgl(), typeout();
+extern void free(), free_seq(), afree();
+extern void msg(), emsg(), emsgl(), typeout();
+extern ATOM *wtokenise();
 extern KEYSEQ *make_seq();
 extern COMMAND *find_command();
 extern KEYMAP *find_keymap();
@@ -309,19 +311,19 @@ FORM *args;
 	return(newargs);
 }
 /****************************************************************************/
-FORM *add_form(list, form)
-FORM *list, *form;
+FORM *add_form(forms, form)
+FORM *forms, *form;
 {
 	/* Append the form to the list */
 
-	if (list == NULL) {
+	if (forms == NULL) {
 		return(form);
 	}
 
 	/* Append the form later in the list */
 
-	list->next = add_form(list->next, form);
-	return(list);
+	forms->next = add_form(forms->next, form);
+	return(forms);
 }
 /****************************************************************************/
 static FORM *copy_list(forms)
@@ -420,7 +422,7 @@ FORM *forms;
 /****************************************************************************/
 FORM *atom(seq, type)
 KEYSEQ *seq;
-short type;
+unsigned type;
 {
 	/* Make an atom containing the key sequence */
 
@@ -440,7 +442,7 @@ short type;
 /****************************************************************************/
 FORM *atomstr(text, type)
 char *text;
-int type;
+unsigned type;
 {
 	/* Make an atom containing a verbatim copy of text */
 
@@ -682,7 +684,8 @@ int headings;
 	for (func = functions; !user_quit && func->name != NULL; func++) {
 		/* Print the function if it matches the regex */
 
-		if (rexpr == NULL || !regexec(rexpr, func->name, 0, NULL)) {
+		if (rexpr == NULL
+		    || !regexec(rexpr, func->name, 0, NULL, 0)) {
 			/* Print a header if required */
 
 			if (headings) {
@@ -822,7 +825,7 @@ FORM *forms;
 FORM *f_progn(forms)
 FORM *forms;
 {
-	/* Set a variable to the value specified in forms */
+	/* Evaluate a list of forms and return the value of the last */
 
 	FORM *f, *last = NULL;
 
@@ -987,7 +990,7 @@ FORM *forms;
 	/* Run a shell command and return it's output */
 
 	char *output, *canon;
-	char *start, *end, *p, *q;
+	ATOM *alist;
 	FORM *value;
 
 	/* Get the output from the command */
@@ -999,47 +1002,15 @@ FORM *forms;
 		return(c_nil());
 	}
 
-	/* Skip leading spaces in the output */
-
-	for (start = output; isspace(*start); start++) {
-		/* NULL LOOP */
-	}
-
-	/* And trailing spaces and the newline */
-
-	end = start + strlen(start);
-	while (end > start && isascii(*(end - 1))
-	       && isspace(*(end - 1))) {
-		end--;
-	}
-
-	/* Allocate space for the canonical output and fill it */
-
-	q = canon = xmalloc(end - start + 1);
-	p = start;
-
-	while (p < end) {
-		if (isspace(*p)) {
-			/* Simply include a canonical space */
-
-			*q++ = ' ';
-			while (isspace(*(p + 1))) {
-				p++;
-			}
-		} else {
-			/* Just copy the character */
-
-			*q++ = *p++;
-		}
-	}
-	*q = '\0';
-
-	/* Set the function value as a string */
+	/* Canonicalise the output of the command */
 
+	alist = wtokenise(output);
+	canon = atext(NULL, alist, AC_WS);
 	value = atomstr(canon, FT_STRING);
 
 	/* Clean up and return the value */
 
+	afree(alist);
 	free(output);
 	free(canon);
 	return(value);
@@ -1066,13 +1037,6 @@ FORM *c_nil()
 	return(list(NULL));
 }
 /****************************************************************************/
-FORM *c_errored()
-{
-	/* Return a error-marking form */
-
-	return(atom(NULL, FT_ERRORED));
-}
-/****************************************************************************/
 FORM *c_version()
 {
 	/*
@@ -1084,3 +1048,10 @@ FORM *c_version()
 	return(atomstr(VERSION, FT_STRING));
 }
 /****************************************************************************/
+FORM *c_errored()
+{
+	/* Return a error-marking form */
+
+	return(atom(NULL, FT_ERRORED));
+}
+/****************************************************************************/
