@@ -1,5 +1,5 @@
 /* Io.c - Terminal input and output functions for af.
-   Copyright (C) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1990 - 2003 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,24 +19,32 @@
 #include <stdio.h>
 #include <ctype.h>
 #include <fcntl.h>
-#include <varargs.h>
+#include <errno.h>
 #include "af.h"
 #include "keyseq.h"
 #include "functions.h"
 #include "variable.h"
 #include "complete.h"
 #include "io.h"
+#include "misc.h"
 #include STRING_HDR
+#include TIME_HDR
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif /* HAVE_UNISTD_H */
 
+#ifdef __STDC__
+#include <stdarg.h>
+#else /* ! __STDC__ */
+#include <varargs.h>
+#endif /* ! __STDC__ */
+
 /****************************************************************************/
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: io.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: io.c,v 2.5 2003/11/27 01:45:57 malc Exp $";
 static char *IoId = IOID;
 #endif /* ! lint */
 
@@ -49,6 +57,8 @@ extern char *readline();
 extern int strcasecmp(), get_vval();
 extern int mklower(), tlines();
 extern int tcols(), vtupdate();
+extern unsigned sleep(), save_atimer();
+extern unsigned restore_atimer();
 extern void free(), typeout(), tbeep();
 extern void tclear(), tclrline();
 extern void tmove(), treturn(), panic();
@@ -264,11 +274,11 @@ int quit;
 {
 	/* Get a yes or no response and return TRUE if it's yes */
 
-	int qchar, c;
+	int qchar, key, folded;
 
 	/* Get the user's quit character */
 
-	qchar = mklower(get_vval(V_QUITCHAR));
+	qchar = get_vval(V_QUITCHAR);
 
 	/* Force the screen to be up to date */
 
@@ -282,31 +292,33 @@ int quit;
 
 	/* Get a key from the user */
 
-	c = get_key();
+	key = get_key();
+	folded = mklower(key);
 
 	/* Check we have a valid key */
 
-	while (c != qchar && mklower(c) != 'y' && mklower(c) != 'n') {
+	while (key != qchar && folded != CONFIRM_YES
+	       && folded != CONFIRM_NO) {
 		/* Invalid key; beep and try again */
 
 		tbeep();
-		c = get_key();
+		key = get_key();
+		folded = mklower(key);
 	}
 
-	/* This may set the user quit flag */
+	/* Set and handle the user quit flag */
 
-	user_quit = (c == qchar);
-
-	/* Output a message if quit set */
-
-	if (user_quit || quit && mklower(c) != 'y') {
+	if ((user_quit = (key == qchar))
+	    || quit && folded != CONFIRM_YES) {
 		tbeep();
 		msg("Quit");
 	} else {
 		clearmsg();
 	}
 
-	return(mklower(c) == 'y');
+	/* And return whether we confirmed */
+
+	return(folded == CONFIRM_YES);
 }
 /****************************************************************************/
 int long_confirm(prompt, quit)
@@ -332,19 +344,19 @@ int quit;
 
 	do {
 		s = get_str(NULL, full_prompt);
-	} while (!user_quit && (s == NULL || strcasecmp(s, "yes")
-				&& strcasecmp(s, "no")));
+	} while (!user_quit && (s == NULL || strcasecmp(s, LCONFIRM_YES)
+				&& strcasecmp(s, LCONFIRM_NO)));
 
 	/* Have we responded yes? */
 
-	ok = (!user_quit && !strcasecmp(s, "yes"));
+	ok = (!user_quit && !strcasecmp(s, LCONFIRM_YES));
 
 	/* Output a message if quit set */
 
-	if (user_quit || quit && !ok) {
+	if (!user_quit && quit && !ok) {
 		tbeep();
 		msg("Quit");
-	} else {
+	} else if (!user_quit) {
 		clearmsg();
 	}
 
@@ -354,6 +366,62 @@ int quit;
 	return(ok);
 }
 /****************************************************************************/
+int select_key(prompt, selections, fold_case)
+char *prompt, *selections;
+int fold_case;
+{
+	/*
+	 * Select a key from a list of selections.  The null
+	 * character is never a valid selection.  Returns the
+	 * key selected, or EOF if the user quit.  If case_fold
+	 * is true then fold input keys to lower case before
+	 * checking them.
+	 */
+
+	int qchar, key, folded;
+
+	/* Get the user's quit character */
+
+	qchar = get_vval(V_QUITCHAR);
+
+	/* Output the prompt and options */
+
+	msg(prompt);
+
+	/* Get a key from the user */
+
+	key = get_key();
+	folded = (fold_case) ? mklower(key) : key;
+
+	/* Check we have a valid key */
+
+	while (key != qchar &&
+	       (key == '\0' || strchr(selections, folded) == NULL)) {
+		/* Invalid key; beep and try again */
+
+		tbeep();
+		key = get_key();
+		folded = (fold_case) ? mklower(key) : key;
+	}
+
+	/* This may set the user quit flag */
+
+	user_quit = (key == qchar);
+
+	/* Output a message if quit set */
+
+	if (user_quit) {
+		tbeep();
+		msg("Quit");
+	} else {
+		clearmsg();
+	}
+
+	/* Now return the selected key */
+
+	return((key == qchar) ? EOF : folded);
+}
+/****************************************************************************/
 char *get_str(form, prompt)
 FORM *form;
 char *prompt;
@@ -525,8 +593,12 @@ char *text;
 }
 /****************************************************************************/
 /*VARARGS*/
+#ifdef __STDC__
+void msgl(char *arg, ...)
+#else /* ! __STDC__ */
 void msgl(va_alist)
 va_dcl
+#endif /* ! __STDC__ */
 {
 	/*
 	 * Display a list of strings for the user.
@@ -538,15 +610,23 @@ va_dcl
 	 * need, and vsprintf is not universally available.
 	 */
 	
-	char *arg;
 	va_list arglist;
 
+#ifndef __STDC__
+	char *arg;
+#endif /* ! __STDC__ */
+
 	/* We don't show messages in scripts */
 
 	if (!scripting) {
 		/* Initialise the varargs handling and message length */
 
+#ifdef __STDC__
+		va_start(arglist, arg);
+#else /* ! __STDC__ */
 		va_start(arglist);
+		arg = va_arg(arglist, char *);
+#endif /* ! __STDC__ */
 		msg_len = 0;
 
 		/* Set up for a new message */
@@ -555,8 +635,11 @@ va_dcl
 
 		/* Now loop through the arguments, printing them */
 
-		while ((arg = va_arg(arglist, char *)) != NULL) {
+		while (arg != NULL) {
+			/* Print this argument and move to the next */
+
 			show_msg(arg);
+			arg = va_arg(arglist, char *);
 		}
 
 		/* End the message and clean up */
@@ -568,8 +651,12 @@ va_dcl
 }
 /****************************************************************************/
 /*VARARGS*/
+#ifdef __STDC__
+void emsgl(char *arg, ...)
+#else /* ! __STDC__ */
 void emsgl(va_alist)
 va_dcl
+#endif /* ! __STDC__ */
 {
 	/*
 	 * Beep and display a list of strings for the user.
@@ -581,12 +668,20 @@ va_dcl
 	 * need, and vsprintf is not universally available.
 	 */
 	
-	char *arg;
 	va_list arglist;
 
+#ifndef __STDC__
+	char *arg;
+#endif /* ! __STDC__ */
+
 	/* Initialise the varargs handling */
 
-	va_start(arglist);
+#ifdef __STDC__
+		va_start(arglist, arg);
+#else /* ! __STDC__ */
+		va_start(arglist);
+		arg = va_arg(arglist, char *);
+#endif /* ! __STDC__ */
 
 	/* Beep and set up for a new message */
 
@@ -595,12 +690,18 @@ va_dcl
 
 	/* Now loop through the arguments, printing them */
 	
-	while ((arg = va_arg(arglist, char *)) != NULL) {
+	while (arg != NULL) {
+		/* Print this argument to echo area or typeout */
+
 		if (type_errors) {
 			typeout(arg);
 		} else {
 			show_msg(arg);
 		}
+
+		/* And move to the next argument */
+
+		arg = va_arg(arglist, char *);
 	}
 
 	/* Flush the output or end the line and clean up */
@@ -614,6 +715,25 @@ va_dcl
 	return;
 }
 /****************************************************************************/
+void show_error(text)
+char *text;
+{
+	/* Show the text of the message for a short time */
+
+	ATIMER tbuf;
+
+	/* First print the error message */
+
+	emsg(text);
+
+	/* Handle alarm here since some sleeps won't */
+
+	(void) save_atimer(&tbuf);
+	(void) sleep(ECHO_DELAY);
+	(void) restore_atimer(&tbuf);
+	return;
+}
+/****************************************************************************/
 void clearmsg()
 {
 	/* Clear any previously displayed message */
@@ -656,7 +776,7 @@ void show_lossage()
 			/* A line break here would look good */
 
 			typeout("\n");
-			col = 0;
+			col = strlen(keystr) + 1;
 		}
 
 		/* Type-out the character and a space */
@@ -667,7 +787,7 @@ void show_lossage()
 
 		/* And increment the pointer */
 
-		p = (p >= lossage + LOSSAGE_SIZE) ? lossage : p + 1;
+		p = (p >= lossage + LOSSAGE_SIZE - 1) ? lossage : p + 1;
 	}
 
 	/* End the typeout and return */
@@ -771,10 +891,10 @@ unsigned delay;
 	/* Add the character to the lossage */
 
 	*lossage_end = key;
-	lossage_end = (lossage_end >= lossage + LOSSAGE_SIZE)
+	lossage_end = (lossage_end >= lossage + LOSSAGE_SIZE - 1)
 		? lossage : lossage_end + 1;
 	lossage_start = (lossage_end > lossage_start ||
-			 lossage_end == lossage + LOSSAGE_SIZE)
+			 lossage_end == lossage + LOSSAGE_SIZE - 1)
 		? lossage : lossage_end + 1;
 
 	/* Postprocess the key if required and return it */
