@@ -1,5 +1,5 @@
 /* Keymaps.c - Keymap handling for af.
-   Copyright (C) 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1992 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -34,7 +34,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: keymaps.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: keymaps.c,v 2.2 2002/08/21 23:54:48 malc Exp $";
 static char *KeymapId = KEYMAPID;
 #endif /* ! lint */
 
@@ -64,9 +64,9 @@ static char *cmd_bindings(), *strrange();
 static int same_binding(), rebind(), last_of_range();
 static int def_binding(), write_binding();
 static void mode_bindings();
-static KEYMAP *add_keymap();
+static KEYMAP *add_keymap(), *prefix_map();
 static BINDING *make_binding(), *add_binding();
-static BINDING *del_binding();
+static BINDING *del_binding(), *direct_binding();
 static KEYMAP_DEFAULTS *def_map();
 
 /****************************************************************************/
@@ -289,29 +289,47 @@ int key;
 {
 	/* Return the binding of key in map */
 
+	BINDING *binding;
+	KEYMAP *pmap;
+
+	/* First try to find a local binding for the key */
+
+	if ((binding = direct_binding(map, key)) != NULL) {
+		/* Return this binding */
+
+		return(binding);
+	}
+
+	/* Try converting metacharacters to the prefixed character */
+
+	if ((pmap = prefix_map(map, key)) != NULL) {
+		/* Return the key's binding in the prefix keymap */
+
+		return(direct_binding(pmap, toascii(key)));
+	}
+
+	/* No binding found for this key */
+
+	return(NULL);
+}
+/****************************************************************************/
+static BINDING *direct_binding(map, key)
+KEYMAP *map;
+int key;
+{
+	/* Return key's binding in keymap map */
+
 	BINDING *case_binding = NULL, *b;
 
-	/* Convert metacharacters to the meta prefix character */
+	/* Can't match metachars unless handle-meta-bindings is set */
 
-	if (!isascii(key)) {
-		/* Find the meta-prefix's binding in the map */
-
-		b = find_binding(map, get_vval(V_METACHAR));
-
-		/* And return the key's binding in that map */
-
-		if (b != NULL && IS_KEYMAP(b)) {
-			/* Recurse to find the binding */
-
-			return(find_binding(b->object.map, toascii(key)));
-		}
-
-		/* Invalid prefix key used; return error */
+	if (!isascii(key) && !get_vval(V_META_BIND)) {
+		/* Can't be a local binding */
 
 		return(NULL);
 	}
 
-	/* Search the bindings for the key */
+	/* Loop through the keymap looking for a binding */
 
 	for (b = map->bindings; b != NULL; b = b->next) {
 		/* Check for a (possibly case-folded) binding */
@@ -329,6 +347,34 @@ int key;
 	return((IS_UNBOUND(case_binding)) ? NULL : case_binding);
 }
 /****************************************************************************/
+static KEYMAP *prefix_map(map, key)
+KEYMAP *map;
+int key;
+{
+	/* Return the keymap the prefix for key is bound to in map */
+
+	BINDING *binding;
+
+	/* The only prefix we have applies only to metachars */
+
+	if (isascii(key)) {
+		return(NULL);
+	}
+
+	/* Check if the meta prefix character is bound to a keymap */
+
+	if ((binding = direct_binding(map, get_vval(V_METACHAR))) != NULL
+	    && binding->type == B_KEYMAP) {
+		/* Return the prefix keymap */
+
+		return(binding->object.map);
+	}
+
+	/* The prefix key isn't bound to a keymap */
+
+	return(NULL);
+}
+/****************************************************************************/
 BINDING *dereference(map, mode, keys)
 KEYMAP *map;
 unsigned mode;
@@ -458,9 +504,9 @@ char *bindname;
 {
 	/* Set a key binding for keys in the keymap */
 
-	int key = 0, asckey = 0;
-	int metakey = EOF;
-	KEYMAP *map;
+	int keycount = 0, key = 0;
+	int bind_meta;
+	KEYMAP *map, *pmap;
 	BINDING *binding;
 
 	/* Find the base keymap for the keys */
@@ -470,32 +516,43 @@ char *bindname;
 		return(FALSE);
 	}
 
+	/* Check if we can bind metachars directly */
+
+	bind_meta = get_vval(V_META_BIND);
+
 	/* Now dereference the keys to find the map */
 
-	while (key < keys->len) {
-		/* Set up the key to dereference */
+	while (map != NULL && keycount < keys->len) {
+		/* Get and dereference the next key */
 
-		asckey = (metakey != EOF) ? metakey : keys->keys[key++];
-		metakey = (!isascii(asckey)) ? toascii(asckey) : EOF;
-		asckey = (!isascii(asckey)) ? get_vval(V_METACHAR) : asckey;
+		key = keys->keys[keycount++];
+		binding = direct_binding(map, key);
 
-		/* And dereference the key */
+		/* Try the meta prefix if no direct binding found */
 
-		if ((binding = find_binding(map, asckey)) == NULL
-		    || !IS_KEYMAP(binding)) {
-			break;
+		if (binding == NULL && (keycount < keys->len || !bind_meta)
+		    && (pmap = prefix_map(map, key)) != NULL) {
+			/* Update the keymap and dereference the key */
+
+			map = pmap;
+			key = toascii(key);
+			binding = direct_binding(map, key);
 		}
 
-		/* Update the keymap if required */
+		/* And update the active keymap if required */
 
-		if (key < keys->len && metakey == EOF) {
-			map = binding->object.map;
+		if (keycount < keys->len) {
+			map = (binding != NULL && binding->type == B_KEYMAP)
+				? binding->object.map : NULL;
 		}
 	}
 
-	/* Check for an error in the key sequence */
+	/* Check for an error in the sequence */
+
+	if (!(keys->len) || map == NULL || keycount < keys->len
+	    || !bind_meta && !isascii(key)) {
+		/* Error in the key sequence */
 
-	if (!(keys->len) || key < keys->len || metakey != EOF) {
 		emsgl("Key sequence ", strseq(keys, SK_KEYSEQ),
 		      " uses invalid prefix characters", NULL);
 		return(FALSE);
@@ -503,7 +560,7 @@ char *bindname;
 
 	/* Rebind the key and return status */
 
-	return(rebind(map, asckey, bindname));
+	return(rebind(map, key, bindname));
 }
 /****************************************************************************/
 static int rebind(map, key, bindname)
@@ -562,12 +619,14 @@ char *prompt;
 {
 	/* Get and return a keymap prefix string */
 
-	int key, metakey = EOF;
+	int key, lkey, gkey;
 	int initialise = TRUE;
 	KEYSEQ *seq = NULL;
 	KEYMAP *lmap = map;
 	KEYMAP *gmap = NULL;
-	BINDING *local, *global;
+	KEYMAP *pmap;
+	BINDING *local = NULL;
+	BINDING *global = NULL;
 
 	/* Output the prompt */
 
@@ -576,22 +635,15 @@ char *prompt;
 	/* Loop until the key is complete */
 
 	do {
-		/* Get a key if required */
+		/* Get a key from the user */
 
-		key = (metakey != EOF) ? metakey : get_key();
+		lkey = gkey = key = get_key();
 
 		/* Add and show the key */
 
-		if (metakey == EOF) {
-			seq = add_seq(seq, key);
-			cmsg(strkey(key, SK_KEYSEQ));
-			cmsg(" ");
-		}
-
-		/* Set up the key to dereference */
-
-		metakey = (!isascii(key)) ? toascii(key) : EOF;
-		key = (!isascii(key)) ? get_vval(V_METACHAR) : key;
+		seq = add_seq(seq, key);
+		cmsg(strkey(key, SK_KEYSEQ));
+		cmsg(" ");
 
 		/* If no map is supplied then use the defaults */
 
@@ -601,14 +653,33 @@ char *prompt;
 			initialise = FALSE;
 		}
 
-		/* Dereference the key in each map */
+		/* Check for a direct local dereference of the key */
 
-		local = find_binding(lmap, key);
-		global = (gmap != NULL) ? find_binding(gmap, key) : NULL;
+		if ((local = direct_binding(lmap, lkey)) == NULL
+		    && (pmap = prefix_map(lmap, lkey)) != NULL) {
+			/* Try the prefix binding and update the map */
+
+			local = direct_binding(lmap, toascii(lkey));
+			lmap = (get_vval(V_META_BIND)) ? lmap : pmap;
+			lkey = (get_vval(V_META_BIND)) ? lkey : toascii(lkey);
+		}
+
+		/* Check for a direct global dereference of the key */
+
+		if (gmap != NULL &&
+		    (global = direct_binding(gmap, key)) == NULL
+		    && (pmap = prefix_map(gmap, key)) != NULL) {
+			/* Try the prefix binding and update the map */
+
+			global = direct_binding(gmap, toascii(lkey));
+			gmap = (get_vval(V_META_BIND)) ? gmap : pmap;
+			gkey = (get_vval(V_META_BIND)) ? gkey : toascii(gkey);
+		}
 
 		/* Decide which binding to use */
 
-		local = (local == NULL || local->key != key &&
+		key = (IS_UNBOUND(local)) ? lkey : gkey;
+ 		local = (local == NULL || local->key != key &&
 			 global != NULL && IS_ACTIVE(global, mode)
 			 && global->key == key) ? global : local;
 
@@ -622,7 +693,7 @@ char *prompt;
 
 	/* Check for an error in the key sequence */
 
-	if (metakey != EOF) {
+	if (!isascii(key) && !get_vval(V_META_BIND)) {
 		emsgl("Key sequence ", strseq(seq, SK_KEYSEQ),
 		      " uses invalid prefix characters", NULL);
 		return(NULL);
@@ -635,7 +706,7 @@ char *prompt;
 /****************************************************************************/
 void list_keymaps()
 {
-	/* List all availabe keymaps to typeout */
+	/* List all available keymaps to typeout */
 
 	KEYMAP *map;
 
@@ -659,17 +730,22 @@ void list_bindings()
 
 	/* List the bindings in the mail keymaps */
 
-	typeout("\nBindings local to mode mail:\n");
+	typeout("\nBindings local to mode Mail:\n");
 	mode_bindings(find_keymap(MAIL_MAP), NULL, NULL);
 
+	/* List the bindings in the compose  keymaps */
+
+	typeout("\nBindings local to mode Compose:\n");
+	mode_bindings(find_keymap(COMPOSE_MAP), NULL, NULL);
+
 	/* List the bindings in the typeout keymaps */
 
-	typeout("\nBindings local to mode typeout:\n");
+	typeout("\nBindings local to mode Typeout:\n");
 	mode_bindings(find_keymap(TYPEOUT_MAP), NULL, NULL);
 
 	/* List the bindings in the minibuffer keymaps */
 
-	typeout("\nBindings local to mode minibuffer:\n");
+	typeout("\nBindings local to mode Minibuffer:\n");
 	mode_bindings(find_keymap(MBUF_MAP), NULL, NULL);
 
 	return;
@@ -782,6 +858,8 @@ char *cmdname;
 			   &binding, MI_GLOBAL, NULL, NULL);
 	buf = cmd_bindings(buf, find_keymap(MAIL_MAP),
 			   &binding, MI_MAIL, NULL, NULL);
+	buf = cmd_bindings(buf, find_keymap(COMPOSE_MAP),
+			   &binding, MI_COMPOSE, NULL, NULL);
 	buf = cmd_bindings(buf, find_keymap(TYPEOUT_MAP),
 			   &binding, MI_TYPEOUT, NULL, NULL);
 	buf = cmd_bindings(buf, find_keymap(MBUF_MAP),
