@@ -1,5 +1,5 @@
 /* Mail_cmd.c - Message sending and handling commands for af.
-   Copyright (C) 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1991 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
 
 #include <stdio.h>
 #include <ctype.h>
+#include <errno.h>
 #include "af.h"
 #include "keyseq.h"
 #include "functions.h"
@@ -33,44 +34,40 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: mail_cmd.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: mail_cmd.c,v 2.5 2002/09/08 21:26:22 malc Exp $";
 #endif /* ! lint */
 
 /****************************************************************************/
 /* Global function declarations */
 
-extern char *xmalloc(), *xstrdup(), *vstrcat(), *tempnam();
-extern char *getenv(), *strerror(), *expand(), *formtext();
-extern char *strdate(), *some_addresses(), *local_part();
-extern char *get_vtext(), *get_str(), *get_cstr();
-extern char *get_dstr(), *get_dcstr(), *utos();
-extern int unlink(), chk_msg(), active(), get_vval();
-extern int send_reply(), send_forward(), send_bounce();
-extern int is_blank(), mmdf_form(), write_text();
-extern int non_ascii(), listed(), confirm(); 
-extern int edit_file(), tagset(), open_digest();
-extern int close_pipe(), vtupdate();
-extern unsigned count_messages();
-extern void free(), free_messages(), free_texpr();
-extern void set_sys_tags(), send_mail(), insert();
-extern void msg(), msgl(), emsg(), emsgl(), cmsg();
-extern void typeout(), iso_typeout(), alldisplay();
-extern void show_buffer();
-extern FILE *open_pipe();
-extern MAILBUF *find_buffer();
-extern MESSAGE *get_messages(), *get_one_message();
+extern char *xmalloc(), *xstrdup(), *vstrcat(), *getenv();
+extern char *expand(), *formtext(), *strdate(), *get_vtext();
+extern char *subtract_addresses(), *local_part(), *get_str();
+extern char *get_cstr(), *get_dstr(), *get_dcstr(), *utos();
+extern char *unique_name();
+extern int chk_msg(), send_mail(), send_file(), send_reply();
+extern int send_forward(), send_bounce(), send_attached();
+extern int page_messages(), msg_status_changed(), msg_touched();
+extern int confirm(), get_vval(), tagset(), edit_message();
+extern int save_selected(), print_selected(), pipe_selected();
+extern unsigned count_messages(), text_lines();
+extern void free(), free_texpr(), cmsg(), msgl(), emsg(), emsgl();
+extern void show_buffer(), alldisplay(), redisplay();
+extern WINDOW *add_window(), *del_window();
+extern MAILBUF *add_buffer();
+extern MESSAGE *expand_message();
 extern REGION *get_region();
 extern TAG_EXPR *tagexpr();
 extern ARGUMENT *form_or_arg();
-extern CLIST *fn_complete(), *buf_complete();
+extern CLIST *fn_complete();
 
 /* Local function declarations */
 
 static char *sname();
-static int skip(), msg_touched(), write_mime();
 static int do_reply(), do_page(), do_save();
-static int do_print(), do_pipe();
-static void show_typeout(), show_pager();
+static int do_print(), do_pipe(), do_attach();
+static int do_open_body_parts();
+static int output_format();
 
 /****************************************************************************/
 /* Import the system error number */
@@ -94,14 +91,58 @@ FORM *forms;
 
 	char *to;
 
+	/* Convert any second form to an argument */
+
+	arg = (forms != NULL) ? form_or_arg(forms->next, arg) : arg;
+
 	/* Get the destination of the message */
 
 	to = (forms != NULL) ? formtext(forms) : NULL;
 
 	/* And send the mail */
 
-	send_mail(to, NULL);
-	return(c_t());
+	return((send_mail(to, NULL, NULL, NULL, NULL, arg != NULL))
+	       ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *mail_file(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* Send a file to a specified user or users */
+
+	char *filnam, *to;
+	int status;
+
+	/* Convert any third form to an argument */
+
+	if (forms != NULL) {
+		arg = form_or_arg(forms->next->next, arg);
+	}
+
+	/* Get the name of the file to send */
+
+	if ((filnam = get_cstr(forms, "Send file: ", fn_complete,
+			       C_STRICT)) == NULL) {
+		return(c_errored());
+	}
+	forms = (forms != NULL) ? forms->next : NULL;
+
+	/* Expand escape chars in the file name */
+
+	filnam = expand(filnam);
+
+	/* Get the destination of the mail */
+
+	to = (forms != NULL) ? formtext(forms) : NULL;
+
+	/* Send the mail and clean up */
+
+	status = send_file(filnam, to, arg != NULL);
+	free(filnam);
+	return((status) ? c_t() : c_errored());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -112,7 +153,7 @@ FORM *forms;
 {
 	/* Reply to the current message */
 
-	return(do_reply(FALSE) ? c_t() : c_errored());
+	return((do_reply(FALSE)) ? c_t() : c_errored());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -123,7 +164,7 @@ FORM *forms;
 {
 	/* Group reply to the current message */
 
-	return(do_reply(TRUE) ? c_t() : c_errored());
+	return((do_reply(TRUE)) ? c_t() : c_errored());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -135,6 +176,7 @@ FORM *forms;
 	/* Forward a mail message to a specified user or users */
 
 	char *to;
+	int status;
 
 	/* Check there is a message to forward */
 
@@ -148,12 +190,14 @@ FORM *forms;
 
 	/* Forward the mail and update status */
 
-	if (send_forward(to, cwin->point) &&
-	    msg_touched(cwin->point, ST_FORWARDED)) {
+	if ((status = send_forward(to, cwin->point))
+	    && msg_touched(cwin, cwin->point, ST_FORWARDED, FALSE)) {
 		alldisplay(cwin->buf);
 	}
 
-	return(c_t());
+	/* Now return status */
+
+	return((status) ? c_t() : c_errored());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -165,6 +209,7 @@ FORM *forms;
 	/* Bounce a mail message to a specified user or users */
 
 	char *to;
+	int status;
 
 	/* Check there is a message to bounce */
 
@@ -178,12 +223,201 @@ FORM *forms;
 
 	/* Bounce the mail and update status */
 
-	if (send_bounce(to, cwin->point) &&
-	    msg_touched(cwin->point, ST_FORWARDED)) {
+	if ((status = send_bounce(to, cwin->point))
+	    && msg_touched(cwin, cwin->point, ST_FORWARDED, FALSE)) {
 		alldisplay(cwin->buf);
 	}
 
-	return(c_t());
+	/* Now return status */
+
+	return((status) ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *attach_msg(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* Attach a mail message to an outgoing mail */
+
+	char *to, *subject;
+	int status;
+	MESSAGE **messages;
+
+	/* Check there is a message to attach */
+
+	if (!chk_msg(cwin, TRUE)) {
+		return(c_errored());
+	}
+
+	/* Convert any second form to an argument */
+
+	if (forms != NULL) {
+		arg = form_or_arg(forms->next, arg);
+	}
+
+	/* Build the array of messages to attach */
+
+	messages = (MESSAGE **) xmalloc(2 * sizeof(MESSAGE *));
+	*messages = cwin->point;
+	*(messages + 1) = NULL;
+
+	/* Get the destination and subject of the message */
+
+	to = (forms != NULL) ? formtext(forms) : NULL;
+	subject = cwin->point->subject;
+
+	/* Resend the mail and update status */
+
+	status = do_attach(to, subject, messages, arg != NULL);
+
+	/* Clean up and return status */
+
+	free(messages);
+	return((status) ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *attach_region(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* Send the messages in the region as an attachment or digest */
+
+	char *to;
+	int no_msgs, status;
+	MESSAGE **messages, **p, *m;
+	REGION *region;
+
+	/* Get the region */
+
+	if ((region = get_region(cwin)) == NULL) {
+		return(c_errored());
+	}
+
+	/* Convert any second form to an argument */
+
+	if (forms != NULL) {
+		arg = form_or_arg(forms->next, arg);
+	}
+
+	/* Count the number of messages we're attaching */
+
+	no_msgs = 0;
+	for (m = region->start; m->text != NULL && m != region->end;
+	     m = m->next) {
+		/* Count the message if it's visible */
+
+		if (m->visible) {
+			no_msgs++;
+		}
+	}
+
+	/* Build the array of messages to attach */
+
+	messages = (MESSAGE **) xmalloc((no_msgs + 1) * sizeof(MESSAGE *));
+	p = messages;
+	for (m = region->start; m->text != NULL && m != region->end;
+	     m = m->next) {
+		/* Add the message if it's visible */
+
+		if (m->visible) {
+			*p++ = m;
+		}
+	}
+
+	/* Terminate the array with a null pointer */
+
+	*p = NULL;
+
+	/* Get the destination of the message */
+
+	to = (forms != NULL) ? formtext(forms) : NULL;
+
+	/* Send the mail and update status */
+
+	status = do_attach(to, NULL, messages, arg != NULL);
+
+	/* Clean up and return status */
+
+	free(messages);
+	return((status) ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *attach_tagset(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* Send the messages in a tagset as an attachment or digest */
+
+	char *tags, *to;
+	int no_msgs, status;
+	MESSAGE **messages, **p, *m;
+	TAG_EXPR *texpr;
+
+	/* Get the tags defining the tagset to make a digest from */
+
+	if ((tags = get_dstr(forms, "Digestify messages tagged: ",
+			     DEFAULT_TAG)) == NULL) {
+		return(c_errored());
+	}
+	forms = (forms != NULL) ? forms->next : NULL;
+
+	/* Convert the tags to a tag expression */
+
+	if ((texpr = tagexpr(cwin, tags)) == NULL) {
+		return(c_errored());
+	}
+
+	/* Convert any third (now second) form to an argument */
+
+	if (forms != NULL) {
+		arg = form_or_arg(forms->next, arg);
+	}
+
+	/* Count the number of messages we're attaching */
+
+	no_msgs = 0;
+	for (m = cwin->buf->messages; m->text != NULL; m = m->next) {
+		/* Save the message if it's selected and visible */
+
+		if (m->visible && tagset(m, texpr)) {
+			no_msgs++;
+		}
+	}
+
+	/* Build the array of messages to attach */
+
+	messages = (MESSAGE **) xmalloc((no_msgs + 1) * sizeof(MESSAGE *));
+	p = messages;
+	for (m = cwin->buf->messages; m->text != NULL; m = m->next) {
+		/* Add the message if it's selected and visible */
+
+		if (m->visible && tagset(m, texpr)) {
+			*p++ = m;
+		}
+	}
+
+	/* Terminate the array with a null pointer */
+
+	*p = NULL;
+
+	/* Get the destination of the message */
+
+	to = (forms != NULL) ? formtext(forms) : NULL;
+
+	/* Send the mail and update status */
+
+	status = do_attach(to, NULL, messages, arg != NULL);
+
+	/* Clean up and return status */
+
+	free(messages);
+	return((status) ? c_t() : c_errored());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -194,7 +428,16 @@ FORM *forms;
 {
 	/* "Open" a message - ie. display it to typeout or to a pager */
 
-	return((do_page(get_vtext(V_PAGER), forms, arg, TRUE))
+	/* Check there is a message to open */
+
+	if (!chk_msg(cwin, TRUE)) {
+		return(c_errored());
+	}
+
+	/* Open the message and return status */
+
+	return((do_page(cwin->point, cwin->point->next,
+			get_vtext(V_PAGER), forms, arg, PG_MESSAGE))
 	       ? c_t() : c_errored());
 }
 /****************************************************************************/
@@ -206,8 +449,96 @@ FORM *forms;
 {
 	/* Send a message to the user's preferred pager */
 
-	return((do_page(getenv(PAGER), forms, arg, FALSE)) ?
-	       c_t() : c_errored());
+	/* Check there is a message to page */
+
+	if (!chk_msg(cwin, TRUE)) {
+		return(c_errored());
+	}
+
+	/* Open the message and return status */
+
+	return((do_page(cwin->point, cwin->point->next,
+			getenv(PAGER), forms, arg, PG_MESSAGE))
+	       ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *open_raw_msg(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* View the raw text of a message via typeout or a pager */
+
+	/* Check there is a message to open */
+
+	if (!chk_msg(cwin, TRUE)) {
+		return(c_errored());
+	}
+
+	/* Open the raw message and return status */
+
+	return((do_page(cwin->point, cwin->point->next,
+			get_vtext(V_PAGER), forms, arg, PG_RAW))
+	       ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *open_body_parts(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* View the body parts of a message as messages in a new buffer */
+
+	return((do_open_body_parts()) ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *owin_open_body_parts(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* View the body parts of a message in the other window */
+
+	int newlines = 0;
+	WINDOW *oldwin = cwin;
+
+	/* Make the new window if required */
+
+	if ((cwin = cwin->next) == oldwin) {
+		/* How many lines will the window have? */
+
+		newlines = (cwin->bottom - cwin->top + 1) / 2;
+
+		/* Now try to create the other window */
+
+		if ((cwin = add_window(cwin, newlines)) == NULL) {
+			return(c_errored());
+		}
+
+		/* Display the current buffer in the new window */
+
+		show_buffer(cwin, oldwin->buf);
+	}
+
+	/* Display the body parts in the new window */
+
+	if (!do_open_body_parts()) {
+		/* Delete any new window and fail */
+
+		if (newlines > 0) {
+			(void) del_window(cwin);
+		}
+		cwin = oldwin;
+		return(c_errored());
+	}
+
+	/* Update the old window and return success */
+
+	redisplay(oldwin);
+	return(c_t());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -459,62 +790,6 @@ FORM *forms;
 }
 /****************************************************************************/
 /*ARGSUSED*/
-FORM *explode(seq, arg, forms)
-KEYSEQ *seq;
-ARGUMENT *arg;
-FORM *forms;
-{
-	/* Explode a mail digest into many messages */
-
-	MESSAGE *messages = NULL, *m;
-
-	/* Check there is a current message */
-
-	if (!chk_msg(cwin, TRUE)) {
-		return(c_errored());
-	}
-
-	/* Let the user know what we're doing */
-
-	msg("Exploding digest...");
-
-	/* Set up the digest for parsing and parse it */
-
-	if (open_digest(cwin->point)) {
-		messages = get_messages(NULL, NULL, 0L);
-	}
-
-	/* Check the message was a digest */
-	
-	if (messages == NULL || count_messages(messages, TRUE) < 2) {
-		emsg("Message is not a valid digest");
-		free_messages(messages);
-		return(c_errored());
-	}
-
-	/* Remove the null message from the list */
-
-	for (m = messages; m->next->text != NULL; m = m->next) {
-		/* NULL LOOP */
-	}
-	free(m->next);
-	m->next = NULL;
-
-	/* Insert the messages after the digest */
-
-	cwin->point = cwin->point->next;
-	insert(cwin, messages);
-	cwin->buf->mod = TRUE;
-
-	/* And update the display */
-
-	alldisplay(cwin->buf);
-	cmsg(" Done");
-
-	return(c_t());
-}
-/****************************************************************************/
-/*ARGSUSED*/
 FORM *edit_msg(seq, arg, forms)
 KEYSEQ *seq;
 ARGUMENT *arg;
@@ -522,12 +797,6 @@ FORM *forms;
 {
 	/* Store the mail in a buffer, and then call an editor on it */
 
-	char *tfile;
-	int status;
-	MESSAGE *newpoint, *oldpoint;
-	WINDOW *win;
-	FILE *fp;
-
 	/* Check there is a message to edit */
 
 	if (!chk_msg(cwin, TRUE)) {
@@ -538,125 +807,10 @@ FORM *forms;
 
 	arg = form_or_arg(forms, arg);
 
-	/* Get a temporary file */
+	/* Now edit the message and return status */
 
-	if ((tfile = tempnam(TFILEDIR, TFILEPFX)) == NULL) {
-		emsg("Can't create temporary file: Can't make unique name");
-		return(c_errored());
-	}
-
-	if ((fp = fopen(tfile, "w")) == NULL) {
-		emsgl("Can't open temporary file: ", strerror(errno), NULL);
-		return(c_errored());
-	}
-
-	/* We need to write the file in local mail format */
-
-#ifdef MTA_MMDF_FORMAT
-	if (status = write_text(fp, cwin->point, NULL, NULL,
-		(arg != NULL) ? HS_ALL : HS_NONE, TRUE)) {
-#else /* ! MTA_MMDF_FORMAT */
-	if (status = write_text(fp, cwin->point, NULL, NULL,
-		(arg != NULL) ? HS_ALL : HS_NONE, FALSE)) {
-#endif /* ! MTA_MMDF_FORMAT */
-		/* Error writing the temporary file */
-
-		emsgl("Error writing temporary file: ",
-		      strerror(status), NULL);
-		(void) fclose(fp);
-		(void) unlink(tfile);
-		(void) free(tfile);
-		return(c_errored());
-	}
-	(void) fclose(fp);
-
-	/* Assume that we're going to fail */
-
-	status = FALSE;
-
-	/* Edit the file and read back the message */
-
-	if (edit_file(tfile)) {
-		/* Let the user know what we're doing */
-
-		(void) vtupdate(TRUE);
-		msgl("Reading ", tfile, "...", NULL);
-
-		/* Now read the file and check status */
-
-		if ((newpoint = get_one_message(cwin->point, tfile,
-						arg != NULL)) == NULL) {
-			/* Error; clean up and return failure */
-
-			(void) unlink(tfile);
-			(void) free(tfile);
-			return(c_errored());
-		}
-
-		/* Succeeded, update the buffer's pointers */
-
-		if (cwin->buf->messages == cwin->point) {
-			cwin->buf->messages = newpoint;
-		}
-
-		/* Update the new message's pointers */
-
-		if ((newpoint->prev = cwin->point->prev) != NULL) {
-			newpoint->prev->next = newpoint;
-		}
-		newpoint->next = cwin->point->next;
-		newpoint->next->prev = newpoint;
-
-		/* Remove the old message from the list */
-
-		oldpoint = cwin->point;
-		oldpoint->next = oldpoint->prev = NULL;
-
-		/* Now update all the windows' pointers */
-
-		win = cwin;
-		do {
-			/* Update the window's first, point and mark */
-
-			if (win->first == oldpoint) {
-				win->first = newpoint;
-			}
-			if (win->point == oldpoint) {
-				win->point = newpoint;
-			}
-			if (win->mark == oldpoint) {
-				win->mark = newpoint;
-			}
-
-			/* Move on to the next window */
-
-			win = win->next;
-		} while (win != cwin);
-
-		/* Now free the old point */
-
-		free_messages(oldpoint);
-
-		/* Mark the buffer as modified */
-
-		cwin->buf->mod = TRUE;
-
-		/* Update status, message, and the display */
-
-		cmsg(" Done");
-		(void) msg_touched(cwin->point, ST_READ);
-		alldisplay(cwin->buf);
-		status = TRUE;
-	}
-
-	/* Remove the temporary file */
-
-	(void) unlink(tfile);
-	(void) free(tfile);
-
-	/* And return status */
-
-	return((status) ? c_t() : c_errored());
+	return((edit_message(cwin, cwin->point, arg != NULL, TRUE))
+	       ? c_t() : c_errored());
 }
 /****************************************************************************/
 /*ARGSUSED*/
@@ -668,8 +822,6 @@ FORM *forms;
 	/* Give miscellaneous info about the current message */
 
 	char *dstr;
-	unsigned lines = 0;
-	MSG_TEXT *line;
 
 	/* Check there is a current message */
 
@@ -677,19 +829,14 @@ FORM *forms;
 		return(c_errored());
 	}
 
-	/* Count the number of lines in the message */
-
-	for (line = cwin->point->text; line != NULL; line = line->next) {
-		lines++;
-	}
-
 	/* Get the string representing the date the message was sent */
 
 	dstr = strdate(cwin->point->date, get_vval(V_SHOW_LTIME));
 
 	/* Output the basic message info */
 
-	msgl("Sent ", dstr, ", ", utos(lines), " lines", NULL);
+	msgl("Sent ", dstr, ", ", utos(text_lines(cwin->point->text)),
+	     " lines", NULL);
 
 	/* Now append the message status */
 
@@ -745,7 +892,7 @@ int group;
 
 	/* Strip any ignored addresses in the list */
 
-	if ((to = some_addresses(to, ignored)) == NULL) {
+	if ((to = subtract_addresses(to, ignored)) == NULL) {
 		emsg("Can't reply: All addresses ignored");
 		return(FALSE);
 	}
@@ -753,7 +900,7 @@ int group;
 	/* Set up any carbon-copies */
 
 	cc = (group && get_vval(V_KEEP_CC)) ?
-		some_addresses(cwin->point->cc,	ignored) : NULL;
+		subtract_addresses(cwin->point->cc, ignored) : NULL;
 
 	/* The user may want a warning if we've already replied */
 
@@ -767,7 +914,7 @@ int group;
 	/* Do the actual replying and update status */
 
 	if (send_reply(to, cc, cwin->point) &&
-	    msg_touched(cwin->point, ST_REPLIED)) {
+	    msg_touched(cwin, cwin->point, ST_REPLIED, FALSE)) {
 		alldisplay(cwin->buf);
 	}
 	free(to);
@@ -777,63 +924,51 @@ int group;
 	return(TRUE);
 }
 /****************************************************************************/
-static int do_page(pager, forms, arg, use_mime)
+static int do_page(first, last, pager, forms, arg, how)
+MESSAGE *first, *last;
 char *pager;
 FORM *forms;
 ARGUMENT *arg;
-int use_mime;
+int how;
 {
 	/* Handle displaying messages */
 
-	char *defcmd = DEFPAGER;
+	static char *defcmd = DEFPAGER;
 	char *cmd, *hdrlist;
-	int skip_hdrs;
-
-	/* Check there is a message to page */
-
-	if (!chk_msg(cwin, TRUE)) {
-		return(FALSE);
-	}
+	int fmt;
 
 	/* Convert any form into an argument */
 
 	arg = form_or_arg(forms, arg);
 
-	/* Which headers are we displaying? */
+	/* Set or default the command to use as required */
 
-	skip_hdrs = skip(arg);
-
-	/* Get the command to use for paging */
-
-	if (cwin->point->nontext && skip_hdrs == HS_SHOW &&
-	    use_mime && (cmd = get_vtext(V_MIMEPAGER)) != NULL) {
-		/* We can't skip any headers */
-
-		skip_hdrs = HS_NONE;
-	} else if ((cmd = pager) == NULL && (cmd = getenv(PAGER)) == NULL) {
+	if ((cmd = pager) == NULL && (cmd = getenv(PAGER)) == NULL) {
 		/* No variable or environment, use default */
 
 		cmd = defcmd;
 	}
 
-	/* Which headers are we to skip? */
+	/* Which output format and headers do we need? */
 
-	hdrlist = (skip_hdrs == HS_SHOW) ? get_vtext(V_NOTDISP) : NULL;
+	fmt = (how == PG_RAW) ? WF_TEXT | WF_SHOW | WF_NOBLANK
+		: output_format(arg) | WF_NOBLANK | WF_DECODE | WF_SHOW;
+	hdrlist = (WF_FMT(fmt) == WF_SOME) ? get_vtext(V_NOTDISP) : NULL;
 
-	/* Page the message as required */
+	/* Clear the global "message status changed" flag */
 
-	if (!strcmp(cmd, V_USE_TYPEOUT)) {
-		show_typeout(cwin->point, skip_hdrs, hdrlist);
-	} else {
-		show_pager(cwin->point, cmd, skip_hdrs, hdrlist);
-	}
+	(void) msg_status_changed();
 
-	/* Update the message status */
+	/* Now page the messages until we're finished */
 
-	if (msg_touched(cwin->point, ST_READ)) {
+	(void) page_messages(cwin, first, last, FALSE, FALSE, cmd,
+			     fmt, hdrlist, how, FALSE, 0, 0);
+
+	/* Update the display and return success */
+
+	if (msg_status_changed()) {
 		alldisplay(cwin->buf);
 	}
-
 	return(TRUE);
 }
 /****************************************************************************/
@@ -846,12 +981,8 @@ TAG_EXPR *texpr;
 {
 	/* Handle saving of messages to a file */
 
-	char *mcmd, *prompt, *deflt;
-	char *filnam = NULL;
-	int skip_hdrs, all_mime;
-	int status, mmdf = FALSE;
-	MESSAGE *m;
-	FILE *fp = NULL;
+	char *prompt, *deflt, *filnam = NULL;
+	int fmt;
 
 	/* Convert any second form into an argument */
 
@@ -859,13 +990,9 @@ TAG_EXPR *texpr;
 		arg = form_or_arg(forms->next, arg);
 	}
 
-	/* Which headers are we saving? */
+	/* Which format should we save in? */
 
-	skip_hdrs = (arg != NULL) ? HS_ALL : HS_MBOX;
-
-	/* What command do we use for saving MIME messages */
-
-	mcmd = (skip_hdrs == HS_ALL) ? get_vtext(V_MIMESAVER) : NULL;
+	fmt = (arg != NULL) ? WF_BODY | WF_DECODE | WF_SHOW : WF_MBOX;
 
 	/* Find the first message to be saved */
 
@@ -873,123 +1000,44 @@ TAG_EXPR *texpr;
 		start = start->next;
 	}
 
-	/* Check if all messages to be saved are in MIME format */
+	/* Form the prompt for the file name */
 
-	all_mime = (mcmd != NULL);
-	for (m = start; all_mime && m->text != NULL
-	     && m != end; m = m->next) {
-		if (m->visible && (texpr == NULL || tagset(m, texpr))
-		    && !(m->nontext)) {
-			all_mime = FALSE;
-		}
+	prompt = vstrcat("Save ", sname(stype, fmt), " to file: ", NULL);
+
+	/* Get the file to save the messages to */
+
+	if (WF_FMT(fmt) == WF_BODY && start->filename != NULL) {
+		/* Default the file from the message */
+
+		deflt = xstrdup(start->filename);
+		filnam = get_dcstr(forms, prompt, deflt, fn_complete,
+				   C_PERMISSIVE);
+		free(deflt);
+	} else if (WF_FMT(fmt) == WF_MBOX && strcmp(start->addr, ERRUSER)) {
+		/* Default the file from the first message */
+
+		deflt = vstrcat("+", local_part(start->addr), NULL);
+		filnam = get_dcstr(forms, prompt, deflt, fn_complete,
+				   C_PERMISSIVE);
+		free(deflt);
+	} else {
+		/* No default file in this case */
+
+		filnam = get_cstr(forms, prompt, fn_complete, C_PERMISSIVE);
 	}
+	free(prompt);
 
-	/* Form the prompt for the file name if required */
+	/* Check and expand the file name */
 
-	if (!all_mime) {
-		prompt = vstrcat("Save ", sname(stype, skip_hdrs),
-				 " to file: ", NULL);
-
-		/* Get the file to save the messages to */
-
-		if (skip_hdrs == HS_MBOX && strcmp(start->addr, ERRUSER)) {
-			/* Default the file from the first messages */
-
-			deflt = vstrcat("+", local_part(start->addr), NULL);
-			filnam = get_dcstr(forms, prompt, deflt, fn_complete,
-					   C_PERMISSIVE);
-			free(deflt);
-		} else {
-			/* No default file in this case */
-
-			filnam = get_cstr(forms, prompt, fn_complete,
-					  C_PERMISSIVE);
-		}
-		free(prompt);
-
-		/* Check and expand the file name */
-
-		if (filnam == NULL) {
-			return(FALSE);
-		}
-		filnam = expand(filnam);
-
-		/* Give the user a message */
-
-		msgl("Writing ", sname(stype, skip_hdrs),
-		     " to ", filnam, "...", NULL);
-
-		/* Check the type of the target file */
-
-		mmdf = mmdf_form(filnam);
-
-		/* Open the file we're writing to */
-
-		if ((fp = fopen(filnam, "a")) == NULL) {
-			emsgl("Can't open ", filnam, ": ",
-			      strerror(errno), NULL);
-			free(filnam);
-			return(FALSE);
-		}
+	if (filnam == NULL) {
+		return(FALSE);
 	}
+	filnam = expand(filnam);
 
-	/* Now save the text of the messages */
+	/* Now save the messages as required */
 
-	for (m = start; m->text != NULL && m != end; m = m->next) {
-		/* Is this message one we should save? */
-
-		if (m->visible && (texpr == NULL || tagset(m, texpr))) {
-			/* Print the message and check the status */
-
-			status = (m->nontext && mcmd != NULL)
-				? write_mime(mcmd, "save", m)
-				: write_text(fp, m, NULL, NULL,
-					     skip_hdrs, mmdf);
-
-			/* Check for errors printing the message */
-
-			if (status && !(m->nontext || mcmd == NULL)) {
-				/* Display the error, clean up, and fail */
-
-				emsgl("Error writing ", filnam, ": ",
-				      strerror(status), NULL);
-				(void) fclose(fp);
-				free(filnam);
-				return(FALSE);
-			} else if (status && !all_mime) {
-				/* Clean up and fail */
-
-				(void) fclose(fp);
-				free(filnam);
-				return(FALSE);
-			} else if (status) {
-				/* Just return failure */
-
-				return(FALSE);
-			}
-
-			/* Flag the message as saved */
-
-			(void) msg_touched(m, ST_SAVED);
-		}
-	}
-
-	/* Close the file if we opened it */
-
-	if (!all_mime) {
-		(void) fclose(fp);
-	}
-
-	/* Confirm the write and show any changed tags */
-
-	msgl("Writing ", sname(stype, skip_hdrs), " to ",
-	     filnam, "... Done", NULL);
-	alldisplay(cwin->buf);
-
-	/* Free the file name and return */
-
-	free(filnam);
-	return(TRUE);
+	return(save_selected(cwin, start, end, texpr, filnam,
+			     sname(stype, fmt), fmt));
 }
 /****************************************************************************/
 static int do_print(stype, forms, arg, start, end, texpr)
@@ -1001,18 +1049,16 @@ TAG_EXPR *texpr;
 {
 	/* Handle printing of messages */
 
-	char *cmd, *mcmd, *hdrlist;
-	int skip_hdrs, status;
-	MESSAGE *m;
-	FILE *fp;
+	char *cmd;
+	int fmt;
 
 	/* Convert any form into an argument */
 
 	arg = form_or_arg(forms, arg);
 
-	/* Which headers are we displaying? */
+	/* Which output format do we need */
 
-	skip_hdrs = skip(arg);
+	fmt = output_format(arg) | WF_DECODE | WF_SHOW;
 
 	/* Get the commands to print with */
 
@@ -1020,7 +1066,6 @@ TAG_EXPR *texpr;
 		emsg("Can't print: print-command variable not set");
 		return(FALSE);
 	}
-	mcmd = get_vtext(V_MIMEPRINTER);
 
 	/* Check we really mean to do this */
 
@@ -1028,65 +1073,10 @@ TAG_EXPR *texpr;
 		return(FALSE);
 	}
 
-	/* Which headers are we to print? */
+	/* Now save the messages as required */
 
-	hdrlist = (skip_hdrs == HS_SHOW) ? get_vtext(V_NOTDISP) : NULL;
-
-	/* Print a message for the user */
-
-	msgl("Printing ", sname(stype, skip_hdrs), "...", NULL);
-
-	/* Open a pipe to the print command */
-
-	if ((fp = open_pipe(cmd, "w", TRUE)) == NULL) {
-		(void) close_pipe(fp, TRUE, FALSE);
-		emsgl("Can't start process ", cmd,
-		      ": ", strerror(errno), NULL);
-		return(FALSE);
-	}
-
-	/* Now print the text of the messages */
-
-	for (m = start; m->text != NULL && m != end; m = m->next) {
-		/* Is this message one we should print? */
-
-		if (m->visible && (texpr == NULL || tagset(m, texpr))) {
-			/* Print the message and save the status */
-
-			status = (m->nontext && mcmd != NULL)
-				? write_mime(mcmd, "print", m)
-				: write_text(fp, m, NULL, hdrlist,
-					     skip_hdrs, FALSE);
-
-			/* Check for errors printing the message */
-
-			if (status) {
-				/* Only print an error for non-MIME */
-
-				if (!m->nontext || mcmd == NULL) {
-					emsgl("Error writing pipe: ",
-					      strerror(status), NULL);
-				}
-				(void) close_pipe(fp, TRUE, FALSE);
-				return(FALSE);
-			}
-
-			/* Flag the message as printed */
-
-			(void) msg_touched(m, ST_PRINTED);
-		}
-	}
-
-	/* Close the pipe */
-
-	(void) close_pipe(fp, TRUE, FALSE);
-
-	/* Confirm the print and show changed tags */
-
-	msgl("Printing ", sname(stype, skip_hdrs), "... Done", NULL);
-	alldisplay(cwin->buf);
-
-	return(TRUE);
+	return(print_selected(cwin, start, end, texpr, cmd,
+			      sname(stype, fmt), fmt));
 }
 /****************************************************************************/
 static int do_pipe(stype, forms, arg, start, end, texpr)
@@ -1098,10 +1088,8 @@ TAG_EXPR *texpr;
 {
 	/* Handle piping of messages */
 
-	char *prompt, *cmd, *hdrlist;
-	int skip_hdrs, status;
-	MESSAGE *m;
-	FILE *fp;
+	char *prompt, *cmd;
+	int fmt;
 
 	/* Convert any second form into an argument */
 
@@ -1109,13 +1097,13 @@ TAG_EXPR *texpr;
 		arg = form_or_arg(forms->next, arg);
 	}
 
-	/* Which headers are we displaying? */
+	/* Which output format do we need? */
 
-	skip_hdrs = skip(arg);
+	fmt = output_format(arg) | WF_DECODE | WF_SHOW;
 
 	/* Form the prompt for the command */
 
-	prompt = vstrcat("Pipe ", sname(stype, skip_hdrs),
+	prompt = vstrcat("Pipe ", sname(stype, fmt),
 			 " into command: ", NULL);
 
 	/* Get the command to pipe into */
@@ -1126,163 +1114,103 @@ TAG_EXPR *texpr;
 	}
 	free(prompt);
 
-	/* Which headers are we to print? */
+	/* Now pipe the messages as required */
 
-	hdrlist = (skip_hdrs == HS_SHOW) ? get_vtext(V_NOTDISP) : NULL;
+	return(pipe_selected(cwin, start, end, texpr, cmd,
+			     sname(stype, fmt), fmt));
+}
+/****************************************************************************/
+static int do_open_body_parts()
+{
+	/* Open a buffer containing a message's body-parts */
+	
+	char *name, *unique;
+	MAILBUF *buf;
+	MESSAGE *body_parts;
 
-	/* Open a pipe to the command */
+	/* Check there is a current message */
 
-	if ((fp = open_pipe(cmd, "w", TRUE)) == NULL) {
-		emsgl("Can't start process ", cmd, ": ",
-		      strerror(errno), NULL);
+	if (!chk_msg(cwin, TRUE)) {
 		return(FALSE);
 	}
 
-	/* Actually write the text */
+	/* Get the body parts of the message */
 
-	for (m = start; m->text != NULL && m != end; m = m->next) {
-		/* Is this message one we should pipe? */
+	if ((body_parts = expand_message(cwin, cwin->point)) == NULL) {
+		/* No submessage and not a valid digest either */
 
-		if (m->visible && (texpr == NULL || tagset(m, texpr))) {
-			if (status = write_text(fp, m, NULL, hdrlist,
-						skip_hdrs, FALSE)) {
-				/* Error writing the pipe */
-
-				(void) close_pipe(fp, TRUE, FALSE);
-				emsgl("Error writing pipe: ",
-				      strerror(status), NULL);
-				return(FALSE);
-			}
-			(void) msg_touched(m, ST_READ);
-		}
+		emsg("Message does not have any body parts");
+		return(FALSE);
 	}
 
-	/* Close the pipe and show any changed tags */
+	/* Set up the buffer name and add the buffer */
 
-	(void) close_pipe(fp, TRUE, TRUE);
-	alldisplay(cwin->buf);
+	name = vstrcat("Message from ", cwin->point->from, NULL);
+	unique = unique_name(cwin->buf, name);
+	buf = add_buffer(cwin->buf, unique, NULL, NULL,
+			 M_MAIL | M_BODY_PARTS);
 
+	/* Clean up the names */
+
+	free(name);
+	free(unique);
+
+	/* Set up the buffer's details */
+
+	buf->point = buf->messages = body_parts;
+	buf->no_msgs = count_messages(buf->messages, TRUE);
+	buf->st_mod = TRUE;
+
+	/* Now update the display and return success */
+
+	show_buffer(cwin, buf);
+	redisplay(cwin);
 	return(TRUE);
-}
+}	
 /****************************************************************************/
-static int write_mime(cmd, op, message)
-char *cmd, *op;
-MESSAGE *message;
+static int do_attach(to, subject, messages, all_headers)
+char *to, *subject;
+MESSAGE **messages;
+int all_headers;
 {
-	/* Handle MIME-format messages via a pipe */
+	/* Handle including one or more messages in an outgoing mail */
 
-	int status;
-	FILE *fp;
+	int status, touched = FALSE;
+	MESSAGE **m;
 
-	/* Open a pipe to the command */
+	/* Send the mail and check status */
 
-	if ((fp = open_pipe(cmd, "w", TRUE)) == NULL) {
-		status = errno;
-		emsgl("Can't start process ", cmd, " to ", op,
-		      " message: ", strerror(status), NULL);
-		return(status);
-	}
+	
+	if (status = send_attached(to, subject, messages, all_headers)) {
+		/* Update the status of the attached messages */
 
-	/* Actually write the text */
+		for (m = messages; m != NULL && *m != NULL; m++) {
+			touched = msg_touched(cwin, *m, ST_FORWARDED, FALSE)
+				|| touched;
+		}
 
-	if (status = write_text(fp, message, NULL, NULL, HS_NONE, FALSE)) {
-		(void) close_pipe(fp, TRUE, FALSE);
-		emsgl("Error writing pipe: ", strerror(status), NULL);
-		return(status);
-	}
-	(void) msg_touched(message, ST_READ);
+		/* Update the display if required */
 
-	/* Close the pipe and show any changed tags */
-
-	(void) close_pipe(fp, TRUE, TRUE);
-	alldisplay(cwin->buf);
-
-	return(0);
-}
-/****************************************************************************/
-static void show_typeout(message, skip_hdrs, hdrlist)
-MESSAGE *message;
-int skip_hdrs;
-char *hdrlist;
-{
-	/* Display a message via typeout */
-
-	int found = FALSE;
-	MSG_TEXT *t;
-
-	/* Write the headers of the message to typeout */
-
-	for (t = message->text; !user_quit && t != NULL
-	     && !is_blank(t->line); t = t->next) {
-		/* Display the line if required */
-
-		if (skip_hdrs == HS_NONE || skip_hdrs == HS_SHOW
-		    && !listed(t->line, hdrlist)) {
-			/* We need to typeout this line */
-
-			typeout(t->line);
-			found = TRUE;
+		if (touched) {
+			alldisplay(cwin->buf);
 		}
 	}
 
-	/* Skip the blank line if no headers displayed */
+	/* And return status */
 
-	t = (!found && t != NULL) ? t->next : t;
-
-	/* Handle typing-out 8-bit messages */
-
-	if (non_ascii(message)) {
-		iso_typeout();
-	}
-
-	/* Display the text of the message */
-
-	while (!user_quit && t != NULL && t->line != NULL) {
-		typeout(t->line);
-		t = t->next;
-	}
-
-	/* End the typeout and return */
-
-	typeout(NULL);
-	return;
+	return(status);
 }
 /****************************************************************************/
-static void show_pager(message, cmd, skip_hdrs, hdrlist)
-MESSAGE *message;
-char *cmd, *hdrlist;
-int skip_hdrs;
-{
-	/* Display a message via a pager */
-
-	FILE *fp;
-
-	/* Start up the pager if possible */
-
-	if ((fp = open_pipe(cmd, "w", TRUE)) == NULL) {
-		return;
-	}
-
-	/* Write the text of the message to the pipe */
-
-	(void) write_text(fp, message, NULL, hdrlist, skip_hdrs, FALSE);
-
-	/* Close the pipe and return */
-
-	(void) close_pipe(fp, TRUE, get_vval(V_PAUSE));
-	return;
-}
-/****************************************************************************/
-static int skip(arg)
+static int output_format(arg)
 ARGUMENT *arg;
 {
-	/* Calculate which headers we should skip according to arg */
+	/* Calculate which output format we need according to arg */
 
-	return((arg != NULL) ? (arg->negative) ? HS_ALL : HS_NONE : HS_SHOW);
+	return((arg != NULL) ? (arg->negative) ? WF_BODY : WF_TEXT : WF_SOME);
 }
 /****************************************************************************/
-static char *sname(stype, skip_hdrs)
-int stype, skip_hdrs;
+static char *sname(stype, fmt)
+int stype, fmt;
 {
 	/* Return text describing a selection of messages */
 
@@ -1299,64 +1227,8 @@ int stype, skip_hdrs;
 
 	/* Simply return the correct entry from the arrays */
 
-	return((skip_hdrs == HS_ALL) ? bdy_types[stype] :
-		(skip_hdrs == HS_NONE) ? hdr_types[stype]
-					: msg_types[stype]);
-}
-/****************************************************************************/
-static int msg_touched(message, operation)
-MESSAGE *message;
-int operation;
-{
-	/*
-	 * The message has been touched, update the status and
-	 * tags according to the operation carried out.
-	 * Return TRUE if the system tags were changed.
-	 */
-
-	int modified;
-
-	/* Check for status change */
-
-	modified = (message->new || !message->read);
-
-	/* Update the message's new and read flags */
-	
-	message->new = FALSE;
-	message->read = TRUE;
-
-	/* Handle the operation-specific settings */
-
-	switch (operation) {
-	case ST_REPLIED:
-		modified = (modified || !message->replied);
-		message->replied = TRUE;
-		break;
-	case ST_FORWARDED:
-		modified = (modified || !message->forwarded);
-		message->forwarded = TRUE;
-		break;
-	case ST_SAVED:
-		modified = (modified || !message->saved);
-		message->saved = TRUE;
-		break;
-	case ST_PRINTED:
-		modified = (modified || !message->printed);
-		message->printed = TRUE;
-		break;
-	}
-
-	/* Set the system tags and modification flag if possible */
-
-	if (modified) {
-		if (!active(cwin->buf, M_READONLY)) {
-			cwin->buf->st_mod = TRUE;
-		}
-		set_sys_tags(message);
-	}
-
-	/* Return TRUE if message statuses have changed */
-
-	return(modified);
+	return((WF_FMT(fmt) == WF_BODY) ? bdy_types[stype]
+	       : (WF_FMT(fmt) & WF_TEXT) ? hdr_types[stype]
+	       : msg_types[stype]);
 }
 /****************************************************************************/
