@@ -1,5 +1,5 @@
 /* Misc.c - Miscellaneous utility routines for af.
-   Copyright (C) 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1991 - 2003 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,19 +13,20 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
+   MA 02111-1307  USA */
 
 
 #include <stdio.h>
 #include <ctype.h>
 #include <signal.h>
 #include <pwd.h>
-#include <varargs.h>
 #include <sys/types.h>
 #include <sys/param.h>
 #include "af.h"
 #include "misc.h"
 #include STRING_HDR
+#include TIME_HDR
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -35,6 +36,12 @@
 #include <sys/utsname.h>
 #endif /* HAVE_UNAME */
 
+#ifdef __STDC__
+#include <stdarg.h>
+#else /* ! __STDC__ */
+#include <varargs.h>
+#endif /* ! __STDC__ */
+
 #ifdef HAVE_GETHOSTBYNAME
 #include <netdb.h>
 #endif /* HAVE_GETHOSTBYNAME */
@@ -49,14 +56,14 @@
 /* RCS info. */
 
 #ifndef lint
-static char *RcsId = "$Id: misc.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: misc.c,v 2.5 2003/11/27 01:45:57 malc Exp $";
 static char *MiscId = MISCID;
 #endif /* ! lint */
 
 /****************************************************************************/
 /* Global function declarations */
 
-extern char *malloc(), *realloc(), *strdup();
+extern char *malloc(), *realloc();
 extern char *getlogin(), *getenv(), *getcwd();
 extern int kill(), strncasecmp(), pclose();
 extern int mkupper();
@@ -67,6 +74,10 @@ extern struct passwd *getpwnam(), *getpwuid();
 extern FILE *popen();
 extern RETSIGTYPE (*signal())();
 
+#ifndef strdup
+extern char *strdup();
+#endif /* ! strdup */
+
 #ifdef HAVE_GETHOSTNAME
 extern int gethostname();
 #else /* ! HAVE_GETHOSTNAME */
@@ -78,8 +89,14 @@ extern int uname();
 /* Local function declarations */
 
 char *xmalloc(), *xrealloc(), *xstrdup();
-char *vstrcat(), *get_user(), *get_home();
-int is_header();
+char *get_user(), *get_home();
+int is_textual(), is_header();
+
+#ifdef __STDC__
+char *vstrcat(char *, ...);
+#else /* ! __STDC__ */
+char *vstrcat();
+#endif /* ! __STDC__ */
 
 /****************************************************************************/
 char *xmalloc(siz)
@@ -123,8 +140,12 @@ char *s;
 }
 /****************************************************************************/
 /*VARARGS*/
+#ifdef __STDC__
+char *vstrcat(char *arg, ...)
+#else /* ! __STDC__ */
 char *vstrcat(va_alist)
 va_dcl
+#endif /* ! __STDC__ */
 {
 	/*
 	 * Concatenate a list of strings and return the value as per
@@ -137,12 +158,21 @@ va_dcl
 	 * need, and vsprintf is not universally available.
 	 */
 
-	char *buf, *arg;
+	char *buf;
 	va_list arglist;
 
+#ifndef __STDC__
+	char *arg;
+#endif /* ! __STDC__ */
+
 	/* Initialise the varargs handling */
 
+#ifdef __STDC__
+	va_start(arglist, arg);
+#else /* ! __STDC__ */
 	va_start(arglist);
+	arg = va_arg(arglist, char *);
+#endif /* ! __STDC__ */
 
 	/* Initialise the buffer */
 
@@ -150,9 +180,15 @@ va_dcl
 
 	/* Now loop through the arguments, adding them */
 
-	while ((arg = va_arg(arglist, char *)) != NULL) {
+	while (arg != NULL) {
+		/* Add this argument to the line */
+
 		buf = xrealloc(buf, strlen(buf) + strlen(arg) + 1);
 		(void) strcat(buf, arg);
+
+		/* Move to the next argument */
+
+		arg = va_arg(arglist, char *);
 	}
 
 	/* Clean up the varargs handling and return the buffer */
@@ -446,20 +482,29 @@ unsigned num;
 	return(buf);
 }
 /****************************************************************************/
-char *get_line(fp, fold)
+char *xtos(num)
+unsigned num;
+{
+	/* Return a static buffer containing num as a string */
+
+	static char buf[MAXUSTRLEN + 1];
+
+	/* Simply fill and return the buffer */
+
+	(void) sprintf(buf, "%x", num);
+	return(buf);
+}
+/****************************************************************************/
+char *get_line(fp)
 FILE *fp;
-int fold;
 {
 	/*
 	 * Read an arbitrarily long line from the file pointed
 	 * to by fp into a string, returning an allocated string
 	 * containing the line.
-	 * If fold is true then process folded lines (newlines
-	 * followed by white space are not line separators).
 	 */
 
 	char *line = NULL, *tmp;
-	int c, in_fold = FALSE;
 
 	/* Read physical lines until we have a full logical line */
 
@@ -484,25 +529,7 @@ int fold;
 			(void) strcat(line, tmp);
 			free(tmp);
 		}
-
-		/* Process folded headers if required */
-
-		if (fold && is_header(line)) {
-			/* Check for a fold */
-
-			c = getc(fp);
-
-			/* Push the character back */
-
-			if (c != EOF) {
-				(void) ungetc(c, fp);
-			}
-
-			/* Check for a linear-white-space character */
-
-			in_fold = (c == ' ' || c == '\t');
-		}
-	} while (line[strlen(line) - 1] != '\n' || in_fold);
+	} while (line[strlen(line) - 1] != '\n');
 
 	/* Reallocate the line and return it */
 
@@ -510,6 +537,192 @@ int fold;
 	return(line);
 }
 /****************************************************************************/
+char *get_fline(fp, unfold)
+FILE *fp;
+int unfold;
+{
+	/* Process a logical line, which may be folded */
+
+	char *line, *tmp;
+	int c, in_fold;
+
+	/* First get the physical line */
+
+	if ((line = get_line(fp)) != NULL && unfold && is_header(line)) {
+		/* Loop over any folded header lines */
+
+		do {
+			/* Check for a folded header line */
+
+			if ((c = getc(fp)) != EOF) {
+				/* Push the character back */
+
+				(void) ungetc(c, fp);
+			}
+
+			/* Is this line folded? */
+
+			if (in_fold = (c == ' ' || c == '\t')) {
+				/* Read and append the next physical line */
+
+				tmp = get_line(fp);
+				line = xrealloc(line, strlen(line)
+						+ strlen(tmp) + 1);
+				(void) strcat(line, tmp);
+				free(tmp);
+			}
+		} while (in_fold);
+	}
+
+	/* Now return the logical line */
+
+	return(line);
+}
+/****************************************************************************/
+char *get_cline(fp)
+FILE *fp;
+{
+	/* Process a logical line, which may be continued */
+
+	char *line, *tmp;
+
+	/* First get the physical line */
+
+	if ((line = get_line(fp)) != NULL) {
+		/* Loop over any continuation lines */
+
+		while (strlen(line) > 1 && line[strlen(line) - 2] == '\\'
+		       && (tmp = get_line(fp)) != NULL) {
+			/* Read and append the next logical line */
+
+			line[strlen(line) - 2] = '\0';
+			line = xrealloc(line, strlen(line)
+					+ strlen(tmp) + 1);
+			(void) strcat(line, tmp);
+			free(tmp);
+		}
+	}
+
+	/* Now return the logical line */
+
+	return(line);
+}
+/****************************************************************************/
+char *get_binline(fp, len)
+FILE *fp;
+size_t *len;
+{
+	/*
+	 * Read an arbitrarily long line from the file pointed
+	 * to by fp into a string, returning an allocated string
+	 * containing the line.  The length of the string is
+	 * returned in len as a side-effect.
+	 */
+
+	char *line;
+	int c, siz;
+
+	/* Initialise the line and length */
+
+	siz = BUFSIZ;
+	line = xmalloc(siz);
+	*len = 0;
+
+	/* Read characters until we have a complete line */
+
+	while ((c = getc(fp)) != EOF) {
+		/* Resize the line if required */
+
+		if (*len + 1 >= siz) {
+			/* Add another chunk to the line */
+
+			siz += BUFSIZ;
+			line = xrealloc(line, siz);
+		}
+
+		/* Add the character to the line and update the length */
+
+		if ((line[(*len)++] = c) == '\n') {
+			/* We've read the entire line */
+
+			break;
+		}
+	}
+
+	/* Reallocate or free the line */
+
+	if (*len) {
+		/* Reallocate the line to size */
+
+		line = xrealloc(line, *len + 1);
+		line[*len] = '\0';
+	} else {
+		/* Nothing read; free the line */
+
+		free(line);
+		line = NULL;
+	}
+
+	/* And return it */
+
+	return(line);
+}
+/****************************************************************************/
+char *get_fbinline(fp, len, unfold)
+FILE *fp;
+size_t *len;
+int unfold;
+{
+	/* Process a logical line, which may be binary or folded */
+
+	char *line, *tmp;
+	int c, in_fold;
+	size_t tlen;
+
+	/* First get the physical line */
+
+	if ((line = get_binline(fp, len)) != NULL && unfold
+	    && is_textual(line, *len) && is_header(line)) {
+		/* Loop over any folded header lines */
+
+		do {
+			/* Check for a folded header line */
+
+			if ((c = getc(fp)) != EOF) {
+				/* Push the character back */
+
+				(void) ungetc(c, fp);
+			}
+
+			/* Is this line folded? */
+
+			if (in_fold = (c == ' ' || c == '\t')) {
+				/* Read and append the next physical line */
+
+				tmp = get_binline(fp, &tlen);
+				line = xrealloc(line, *len + tlen + 1);
+				(void) memcpy(line + *len, tmp, tlen);
+				*(line + *len + tlen) = '\0';
+				*len += tlen;
+				free(tmp);
+			}
+		} while (in_fold);
+	}
+
+	/* Now return the logical line */
+
+	return(line);
+}
+/****************************************************************************/
+int is_textual(line, len)
+char *line;
+size_t len;
+{
+	/* Return whether this line is textual - no null characters */
+
+	return(strlen(line) >= len);
+}
+/****************************************************************************/
 int is_header(line)
 char *line;
 {
@@ -536,6 +749,19 @@ char *line;
 	return(FALSE);
 }
 /****************************************************************************/
+int is_mime_header(line)
+char *line;
+{
+	/* Return TRUE is the line is a MIME header */
+
+	return(!strncmp(line, MIME_VERSION, strlen(MIME_VERSION))
+	       || !strncmp(line, CONTENT_DESC, strlen(CONTENT_DESC))
+	       || !strncmp(line, CONTENT_DISP, strlen(CONTENT_DISP))
+	       || !strncmp(line, CONTENT_TYPE, strlen(CONTENT_TYPE))
+	       || !strncmp(line, C_T_ENCODING, strlen(C_T_ENCODING))
+	       || !strncmp(line, CONTENT_ID, strlen(CONTENT_ID)));
+}
+/****************************************************************************/
 int is_fromline(line)
 char *line;
 {
