@@ -1,5 +1,5 @@
 /* Netmail.c - POP3 and SMTP mail handlers
-   Copyright (C) 1995, 1996, 1997 Malc Arnold, Ian Dickinson.
+   Copyright (C) 1995 - 2003 Malc Arnold, Ian Dickinson.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
+   MA 02111-1307  USA */
 
 
 #include <stdio.h>
@@ -27,22 +28,28 @@
 #include <netdb.h>
 #include <ctype.h>
 #include <fcntl.h>
-#include <varargs.h>
 #include "af.h"
-#include "sendmail.h"
-#include "netmail.h"
 #include "keyseq.h"
 #include "functions.h"
 #include "variable.h"
 #include "io.h"
+#include "sendmail.h"
+#include "netmail.h"
 #include "misc.h"
 #include STRING_HDR
+#include TIME_HDR
+
+#ifdef __STDC__
+#include <stdarg.h>
+#else /* ! __STDC__ */
+#include <varargs.h>
+#endif /* ! __STDC__ */
 
 /****************************************************************************/
 /* RCS info. */
 
 #ifndef lint
-static char *RcsId = "$Id: netmail.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: netmail.c,v 2.6 2003/11/27 01:45:57 malc Exp $";
 static char *NetmailId = NETMAILID;
 #endif /* ! lint */
 
@@ -57,10 +64,10 @@ static char *NetmailId = NETMAILID;
 /* Global function declarations */
 
 extern char *xmalloc(), *xrealloc(), *xstrdup();
-extern char *vstrcat(), *strerror(), *get_user();
-extern char *get_host(), *get_vtext(), *get_hdr();
-extern char *get_line(), *get_pwstr(), *utos();
-extern int atoi();
+extern char *vstrcat(), *strerror(), *get_host();
+extern char *get_vtext(), *get_line(), *get_hdr();
+extern char *get_pwstr(), *utos();
+extern int atoi(), is_blank(), is_header();
 extern time_t time();
 extern unsigned alarm();
 extern unsigned save_atimer();
@@ -71,15 +78,18 @@ extern void clearmsg(), typeout();
 extern RETSIGTYPE (*signal())();
 
 #ifdef MTA_IS_SMTP
-static int nethdrs(), netfile();
 #ifdef NO_MTA_CC
 extern int strcasecmp();
 #endif /* NO_MTA_CC */
+
+static int nethdrs(), netfile();
 #endif/* MTA_IS_SMTP */
 
 #ifdef READ_VIA_POP3
+extern char *get_user(), *get_addr();
 extern char *strudate();
 extern DATEZONE *date_now();
+
 FILE *open_pop3();
 int close_pop3();
 static char *netslurp();
@@ -94,12 +104,17 @@ static POPFOLDER *get_pfolder();
 /* Local function declarations */
 
 static char *netread(), *tget_line();
-static int netsend(), netclose();
-static int reopen_pop3();
+static int netclose(), reopen_pop3();
 static int get_password();
 static CONNECTION *netopen();
 static RETSIGTYPE time_out();
 
+#ifdef __STDC__
+static int netsend(CONNECTION *, char *, ...);
+#else /* ! __STDC__ */
+static int netsend();
+#endif /* ! __STDC__ */
+
 /****************************************************************************/
 /* Import the global and host error flags */
 
@@ -254,7 +269,7 @@ HEADER *hdrs;
 	for (h = hdrs; h != NULL; h = h->next) {
 		/* Send any header which is defined */
 
-		if (h->hdr_text != NULL) {
+		if (h->text != NULL) {
 #ifdef NO_MTA_CC
 			/* Handle the Bcc header when sending */
 
@@ -277,10 +292,10 @@ HEADER *hdrs;
 				continue;
 			}
 #endif /* NO_MTA_CC */
+
 			/* Send the text and check the return */
 
-			if (!netsend(conn, h->hdr_name, " ",
-				     h->hdr_text, NULL)) {
+			if (!netsend(conn, h->name, " ", h->text, NULL)) {
 				/* Error sending the text */
 
 				return(FALSE);
@@ -460,17 +475,12 @@ int update;
 {
 	/* Shut down and disconnect the current POP3 folder */
 
-	if (update || pop3_folder->error) {
-		/* Close the folder to update or clear any errors */
+	if (update && netsend(pop3_folder->conn, "QUIT", NULL)
+	    && netread(pop3_folder->conn, POP3_OK) != NULL
+	    || pop3_folder->error) {
+		/* Close the actual connection */
 
-		if ((update || netsend(pop3_folder->conn, "RSET", NULL)
-		     && netread(pop3_folder->conn, POP3_OK) != NULL)
-		    && netsend(pop3_folder->conn, "QUIT", NULL)
-		    && netread(pop3_folder->conn, POP3_OK) != NULL) {
-			/* Close the actual connection */
-
-			(void) netclose(pop3_folder->conn);
-		}
+		(void) netclose(pop3_folder->conn);
 
 		/* And update the folder's status */
 
@@ -483,18 +493,20 @@ int update;
 	return(pop3_folder->error);
 }
 /****************************************************************************/
-void kill_pop3()
+void kill_pop3(folder)
+char *folder;
 {
-	/* Kill off all outstanding POP3 connections */
+	/* Kill off a POP3 connection, or all connections if folder is NULL */
 
 	POPFOLDER *pfolder;
 
 	/* Loop through the cache entries closing connections */
 
 	for (pfolder = pop3_cache; pfolder != NULL; pfolder = pfolder->next) {
-		/* Is this folder still connected to a server? */
+		/* Do we need to kill off this connection? */
 
-		if (pfolder->conn != NULL) {
+		if ((folder == NULL || !strcmp(pfolder->folder, folder))
+		    && pfolder->conn != NULL) {
 			/* Reset and close the connection */
 			
 			if (reopen_pop3(pfolder, pfolder->offset) &&
@@ -532,7 +544,7 @@ int hdrs;
 
 	if (pop3_folder->msg > pop3_folder->offset) {
 		if ((line = netslurp(pop3_folder->conn, hdrs)) != NULL) {
-			/* We got a line, just return it */
+			/* Just return the line */
 
 			return(line);
 		}
@@ -553,18 +565,19 @@ int hdrs;
 		    netread(pop3_folder->conn, POP3_OK) == NULL) {
 			/* Error sending the command */
 
-			pop3_folder->conn = NULL;
+			emsgl("Error reading ", pop3_folder->conn->server,
+			      ": ", strerror(net_error), NULL);
 			pop3_folder->error = net_error;
 			return(NULL);
 		}
 
 		/* Return a separator for the message */
 
-		return(vstrcat(MFROM, get_user(), "@", get_vtext(V_DOMAIN),
-			       " ", strudate(date_now()), "\n", NULL));
+		return(vstrcat(MFROM, get_addr(), " ",
+			       strudate(date_now()), "\n", NULL));
 	}
 
-	/* End of input; set synchonised flag and return */
+	/* End of input; set synchronised flag and return */
 
 	pop3_folder->synced = TRUE;
 	return(NULL);
@@ -745,8 +758,7 @@ long offset;
 
 	if (fprintf(folder->conn->wsock, "STAT\r\n") != EOF
 	    && fflush(folder->conn->wsock) != EOF
-	    && (buf = tget_line(folder->conn->rsock,
-				NET_TIMEOUT, FALSE)) != NULL) {
+	    && (buf = tget_line(folder->conn->rsock, NET_TIMEOUT)) != NULL) {
 		/* Connection still ok; update and use it */
 
 		folder->size = (unsigned) atoi(buf + strlen(POP3_OK));
@@ -1084,7 +1096,7 @@ char *response;
 
 	/* Read the line and check status */
 
-        if ((buf = tget_line(conn->rsock, NET_TIMEOUT, FALSE)) == NULL) {
+        if ((buf = tget_line(conn->rsock, NET_TIMEOUT)) == NULL) {
 		/* Error reading the line; handle the error */
 
 		emsgl("Error reading ", conn->server,
@@ -1119,18 +1131,31 @@ char *response;
 }
 /****************************************************************************/
 /*VARARGS 1*/
+#ifdef __STDC__
+static int netsend(CONNECTION *conn, char *arg, ...)
+#else /* ! __STDC__ */
 static int netsend(conn, va_alist)
 CONNECTION *conn;
 va_dcl
+#endif /* ! __STDC__ */
 {
 	/* Send a line of text to a server */
 
-	char *arg, *line;
+	char *line;
 	va_list arglist;
 
+#ifndef __STDC__
+	char *arg;
+#endif /* ! __STDC__ */
+
 	/* Initialise the varargs handling */
 
+#ifdef __STDC__
+	va_start(arglist, arg);
+#else /* ! __STDC__ */
 	va_start(arglist);
+	arg = va_arg(arglist, char *);
+#endif /* ! __STDC__ */
 
 	/* Initialise the line's text */
 
@@ -1139,9 +1164,15 @@ va_dcl
 
 	/* Now loop through adding the arguments to the line */
 
-	while ((arg = va_arg(arglist, char *)) != NULL) {
+	while (arg != NULL) {
+		/* Add this argument to the line */
+
 		line = xrealloc(line, strlen(line) + strlen(arg) + 1);
 		(void) strcat(line, arg);
+
+		/* Move to the next argument */
+
+		arg = va_arg(arglist, char *);
 	}
 
 	/* Send the line and check for an error */
@@ -1165,34 +1196,86 @@ va_dcl
 }
 /****************************************************************************/
 #ifdef READ_VIA_POP3
-static char *netslurp(conn, hdrs)
+static char *netslurp(conn, unfold)
 CONNECTION *conn;
-int hdrs;
+int unfold;
 {
-	/* Slurp in one logical line from a multiline response */
+	/* 
+	 * Slurp in one logical line from a multiline response.
+	 * Handle folding headers lines in the response text.
+	 *
+	 * Assumes that the routine will be repeatedly called
+	 * to read the entire contents of the response, so we
+	 * can read the next line to check for a fold.
+	 */
 
+	static char *next_line = NULL;
+	static int last_line_blank = FALSE;
 	char *buf, *newbuf, *b, *n;
+	int in_fold = TRUE;
 
-	/* Read the line and check status */
+	/* Read the line if we don't already have it */
 
-        if ((buf = tget_line(conn->rsock, NET_TIMEOUT, hdrs)) == NULL) {
+	if (next_line != NULL) {
+		/* Just use the line we've already read */
+
+		buf = next_line;
+		next_line = NULL;
+	} else if ((buf = tget_line(conn->rsock, NET_TIMEOUT)) == NULL) {
 		/* Error reading the line; handle the error */
 
 		emsgl("Error reading ", conn->server,
 		      ": ", strerror(net_error), NULL);
-		netclose(conn);
 		return(NULL);
 	}
 
 	/* Check for a termination line */
 
 	if (!strcmp(buf, ".\r\n") || !strcmp(buf, ".\n")) {
+		/* Now make sure we end a message with a blank line */
+
+		if (!last_line_blank) {
+			/* Save the terminator and return a blank line */
+
+			next_line = buf;
+			last_line_blank = TRUE;
+			return(xstrdup("\n"));
+		}
+
 		/* Clean up and return null */
 
 		free(buf);
 		return(NULL);
 	}
 
+	/* Now handle reading folded header lines into a single line */
+
+	while (unfold && in_fold && is_header(buf)) {
+		/* Read the next line into the static buffer */
+
+		if ((next_line =
+		     tget_line(conn->rsock, NET_TIMEOUT)) == NULL) {
+			/* Error reading the line; handle the error */
+
+			emsgl("Error reading ", conn->server,
+			      ": ", strerror(net_error), NULL);
+			free(buf);
+			return(NULL);
+		}
+
+		/* Is this line folded? */
+
+		if (in_fold = (*next_line == ' ' || *next_line == '\t')) {
+			/* Append the next line to the buffer */
+
+			buf = xrealloc(buf, strlen(buf)
+				       + strlen(next_line) + 1);
+			(void) strcat(buf, next_line);
+
+			free(next_line);
+		}
+	}
+
 	/* Allocate the space for the line */
 
 	n = newbuf = xmalloc(strlen(buf) + 2);
@@ -1217,17 +1300,19 @@ int hdrs;
 	}
 	*n = '\0';
 	free(buf);
-		
+
 	/* Handle quoted dots in the text we read */
 
-	if (*newbuf == '.' && strcmp(newbuf, ".\n")) {
-		/* Quoted dot, strip the first dot */
-
+	if (*newbuf == '.' && *(newbuf + 1) == '.') {
 		buf = xstrdup(newbuf + 1);
 		free(newbuf);
 		newbuf = buf;
 	}
 
+	/* We got a line, is it blank? */
+
+	last_line_blank = is_blank(buf);
+
 	/* Return the canonicalised buffer */
 
 	return(newbuf);
@@ -1269,11 +1354,11 @@ FILE *fp;
 }
 #endif /* MTA_IS_SMTP */
 /****************************************************************************/
-char *tget_line(fp, delay, hdrs)
+char *tget_line(fp, delay)
 FILE *fp;
-int delay, hdrs;
+int delay;
 {
-	/* Get a line from fp using func, with a timeout of delay seconds */
+	/* Get a line from fp with a timeout of delay seconds */
 
 	char *line;
 	ATIMER tbuf;
@@ -1299,7 +1384,7 @@ int delay, hdrs;
 
 	/* Read the line and save the error number */
 
-	line = get_line(fp, hdrs);
+	line = get_line(fp);
 	net_error = (line == NULL) ? errno : 0;
 
 	/* Now reset the signal handling and return */
