@@ -36,7 +36,7 @@
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: readline.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: readline.c,v 2.1 1997/09/07 00:38:47 malc Exp $";
 static char *ReadlineId = READLINEID;
 #endif /* ! lint */
 
@@ -1397,10 +1397,9 @@ char *new_point;
 
 	while (old_point < point) {
 		if (!(mbuf_modes & M_PASSWORD)) {
-			(void) fputs(strchar(*old_point, xpos,
-					     FALSE), stdout);
+			(void) fputs(strchar(*old_point, xpos), stdout);
 		}
-		xpos += charlen(*old_point++, xpos, FALSE);
+		xpos += charlen(*old_point++, xpos);
 	}
 
 	return;
@@ -1817,7 +1816,7 @@ int col, clear;
 
 	/* Redraw the line, if not in password mode */
 
-	while (col + (len = charlen(*pos, col, FALSE)) < tcols() - 1) {
+	while (col + (len = charlen(*pos, col)) < tcols() - 1) {
 		/* Check if this is the current position */
 
 		if (pos == point) {
@@ -1836,7 +1835,7 @@ int col, clear;
 		/* Print the character if not in password mode */
 
 		if (!(mbuf_modes & M_PASSWORD)) {
-			(void) fputs(strchar(*pos++, col, FALSE), stdout);
+			(void) fputs(strchar(*pos++, col), stdout);
 		}
 		col += len;
 	}
@@ -1880,7 +1879,7 @@ static int scroll()
 	for (p = text + offset; p < text + textlen; p++) {
 		/* Update the column and last character */
 
-		if ((col = col + charlen(*p, col, FALSE)) > tcols() - 2) {
+		if ((col = col + charlen(*p, col)) > tcols() - 2) {
 			last = (p - 1);
 			break;
 		}
@@ -1898,7 +1897,7 @@ static int scroll()
 		col = xpos = startx = strlen(prompt);
 
 		for (p = text; p <= point && col < tcols() - 1; p++) {
-			col += charlen(*p, col, FALSE);
+			col += charlen(*p, col);
 		}
 
 		/* Set the scroll offset for the line */
@@ -1910,7 +1909,7 @@ static int scroll()
 			p = point;
 
 			while (p != text && col < tcols() / 2) {
-				col += charlen(*--p, startx, FALSE);
+				col += charlen(*--p, startx);
 			}
 			offset = (p != text) ? p - text : 1;
 		} else {
@@ -1939,13 +1938,13 @@ int pos;
 
 	p = text + offset;
 	nchars = startx;
-	len = charlen(*p, nchars, FALSE);
+	len = charlen(*p, nchars);
 
 	/* Search along until we reach the character */
 
 	while (nchars + len < pos) {
 		nchars += len;
-		len = charlen(*++p, nchars, FALSE);
+		len = charlen(*++p, nchars);
 	}
 
 	return(len);
