@@ -1,5 +1,5 @@
 /* Sendmail.c - Af routines to send mail outwards.
-   Copyright (C) 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1990 - 2003 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,48 +19,53 @@
 #include <stdio.h>
 #include <fcntl.h>
 #include <ctype.h>
-#include <varargs.h>
+#include <errno.h>
 #include "af.h"
 #include "keyseq.h"
 #include "functions.h"
 #include "sendmail.h"
 #include "variable.h"
-#include "complete.h"
-#include "misc.h"
 #include "io.h"
+#include "mime.h"
+#include "misc.h"
 #include STRING_HDR
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif /* HAVE_UNISTD_H */
 
+#ifdef __STDC__
+#include <stdarg.h>
+#else /* ! __STDC__ */
+#include <varargs.h>
+#endif /* ! __STDC__ */
+
 /****************************************************************************/
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: sendmail.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: sendmail.c,v 2.6 2003/10/27 23:16:20 malc Exp $";
 static char *SendmailId = SENDMAILID;
 #endif /* ! lint */
 
 /****************************************************************************/
 /* Global function declarations */
 
-extern char *xmalloc(), *xrealloc(), *xstrdup(), *tempnam();
-extern char *vstrcat(), *ttyname(), *strerror(), *strdate();
-extern char *unmessage(), *alias(), *get_vtext(), *get_ecstr();
-extern char *utos(), *get_hdr(), *get_line(), *copy_stdin_mail();
-extern int wait(), isatty(), strcasecmp(), strncasecmp();
-extern int mklower(), write_text(), get_vval(), confirm();
-extern int get_key(), shellout();
+extern char *xmalloc(), *xrealloc(), *xstrdup(), *vstrcat();
+extern char *strerror(), *get_vtext(), *utos(), *get_header();
+extern char *c_contype(), *c_encoding(), *content_type();
+extern char *get_temp_file();
+extern int wait(), strcasecmp(), strncasecmp(), confirm();
+extern int select_key(), shellout(), get_vval();
+extern int write_composition(), save_composition();
+extern int edit_composition(), compose_mail();
+extern int attach_messages();
 extern unsigned sleep(), save_atimer(), restore_atimer();
-extern void free(), _exit(), free_hdrs(), typeout(), showtext();
-extern void msg(), msgl(), emsgl(), cmsg(), tbeep();
-extern void reset_signals();
-extern HEADER *init_hdrs(), *form_stdin_mail();
-extern HEADER *form_hdrs(), *form_mail();
-extern HEADER *posting_hdrs(), *saving_hdrs();
-extern DATEZONE *date_now();
-extern CLIST *fn_complete();
+extern void free(), _exit(), typeout(), showtext(), msg();
+extern void emsgl(), cmsg(), show_decoded_text();
+extern void reset_signals(), add_posting_headers();
+extern void free_composition();
+extern COMPOSITION *init_composition();
 
 #ifdef MTA_IS_SMTP
 extern int smtp_deliver();
@@ -68,7 +73,7 @@ extern char *get_addr();
 #endif /* MTA_IS_SMTP */
 
 #ifndef MTA_MMDF_FORMAT
-extern char *get_addr(), *strudate();
+extern char *get_addr();
 #endif /* ! MTA_MMDF_FORMAT */
 
 #ifdef MTA_NEEDS_ARGS
@@ -77,17 +82,18 @@ extern char **addr_args();
 
 /* Local function declarations */
 
-int cp_hdrs();
-static char *get_sigfile(), **mailargs();
-static int do_send(), reset_stdin(), ask_copy();
-static int cp_text(), exec_mta(), post();
-static int save_outbound();
+static char **mailargs();
+static int do_send(), exec_mta(), post();
 static unsigned mail_size();
-static void summarize_mail(), list_mail();
+static void summarise_mail(), list_mail();
 static void spell_check_mail();
 
 #ifndef MTA_IS_SMTP
+#ifdef __STDC__
+static char **argvec(char *, char *, ...);
+#else /* ! __STDC__ */
 static char **argvec();
+#endif /* ! __STDC__ */
 #endif /* ! MTA_IS_SMTP */
 
 #ifdef MTA_NEEDS_ARGS
@@ -105,13 +111,33 @@ extern int errno;
 extern int user_quit;
 
 /****************************************************************************/
-void send_mail(to, subject)
-char *to, *subject;
+int send_mail(to, cc, bcc, subject, contype, multipart)
+char *to, *cc, *bcc, *subject, *contype;
+int multipart;
 {
 	/* Simple send out of edited file */
 
-	(void) do_send(to, NULL, subject, NULL, T_SEND);
-	return;
+	return(do_send(NULL, to, cc, bcc, subject, contype, NULL,
+		       NULL, SM_SEND | ((multipart) ? SM_MPART : 0)));
+}
+/****************************************************************************/
+int send_file(filnam, to, ask_ctype)
+char *filnam, *to;
+int ask_ctype;
+{
+	/* Simple send out of (edited) file */
+
+	char *contype;
+
+	/* First see if we can derive a content-type from the file name */
+
+	contype = content_type(filnam);
+	ask_ctype = (ask_ctype || contype != NULL);
+
+	/* Now actually send the mail */
+
+	return(do_send(filnam, to, NULL, NULL, NULL, contype, NULL,
+		       NULL, SM_SEND | ((ask_ctype) ? SM_MIME : 0)));
 }
 /****************************************************************************/
 int send_reply(to, cc, orig_msg)
@@ -120,7 +146,8 @@ MESSAGE *orig_msg;
 {
 	/* Send mail in reply to some source */
 
-	return(do_send(to, cc, orig_msg->subject, orig_msg, T_REPLY));
+	return(do_send(NULL, to, cc, NULL, orig_msg->subject,
+		       NULL, orig_msg, NULL, SM_REPLY));
 }
 /****************************************************************************/
 int send_forward(to, orig_msg)
@@ -129,7 +156,8 @@ MESSAGE *orig_msg;
 {
 	/* Send existing message, possibly editing */
 
-	return(do_send(to, NULL, orig_msg->subject, orig_msg, T_FORWARD));
+	return(do_send(NULL, to, NULL, NULL, orig_msg->subject,
+		       NULL, orig_msg, NULL, SM_FORWARD));
 }
 /****************************************************************************/
 int send_bounce(to, orig_msg)
@@ -138,120 +166,77 @@ MESSAGE *orig_msg;
 {
 	/* Bounce mail - send with original headers */
 
-	return(do_send(to, NULL, NULL, orig_msg, T_BOUNCE));
+	return(do_send(NULL, to, NULL, NULL, NULL,
+		       NULL, orig_msg, NULL, SM_BOUNCE));
 }
 /****************************************************************************/
-void quiet_mail(to, subject)
+int send_attached(to, subject, attachments, all_headers)
 char *to, *subject;
+MESSAGE **attachments;
+int all_headers;
+{
+	/* Send existing messages as an attachment */
+
+	return(do_send(NULL, to, NULL, NULL, subject, NULL, NULL, attachments,
+		       SM_ATTACH | ((all_headers) ? SM_ALL_HEADERS : 0)));
+}
+/****************************************************************************/
+int quiet_mail(to, cc, bcc, subject, contype)
+char *to, *cc, *bcc, *subject, *contype;
 {
 	/* Send from standard input with no interaction */
 
-	(void) do_send(to, NULL, subject, NULL, T_SILENT);
-	return;
+	return(do_send(NULL, to, cc, bcc, subject,
+		       contype, NULL, NULL, SM_SILENT));
 }
 /****************************************************************************/
-static int do_send(to, cc, subject, orig_msg, mail_type)
-char *to, *cc, *subject;
-MESSAGE *orig_msg;
-int mail_type;
+static int do_send(filnam, to, cc, bcc, subject, contype,
+		   orig_msg, attachments, mail_flags)
+char *filnam, *to, *cc, *bcc, *subject, *contype;
+MESSAGE *orig_msg, **attachments;
+int mail_flags;
 {
 	/* Form a mail message, and send it after confirmation */
 
-	char *tfile, *etfile, *sigfile;
-	char *spellcheck;
-	int fd, copy, status;
-	HEADER *hdrs;
+	char *prompt, *spellcheck;
+	int status;
+	COMPOSITION *comp;
 
-	/* Get a temporary file */
+	/* Initialise the composition */
 
-	if ((tfile = tempnam(TFILEDIR, TFILEPFX)) == NULL) {
-		emsgl("Can't create temporary file ", strerror(errno), NULL);
-		return(FALSE);
-	}
-	if ((fd = creat(tfile, 0666)) < 0) {
-		emsgl("Can't create temporary file ", strerror(errno), NULL);
-		free(tfile);
-		return(FALSE);
-	}
-	(void) close(fd);
+	if ((comp = init_composition(to, cc, bcc, subject, contype, filnam,
+				     orig_msg, mail_flags)) == NULL) {
+		/* The user quit the composition or we got an error */
 
-	/* Initialise the headers */
-
-	hdrs = init_hdrs(to, cc, subject, orig_msg, mail_type);
-
-	/* If stdin isn't a tty then read it into the temporary file */
-
-	if (!isatty(fileno(stdin)) &&
-	    ((etfile = copy_stdin_mail()) == NULL || !reset_stdin() ||
-	     (hdrs = form_stdin_mail(tfile, etfile, hdrs,
-				     mail_type)) == NULL)) {
-		/* Failed to copy the input or reset the tty */
-
-		if (etfile != NULL) {
-			(void) unlink(etfile);
-			free(etfile);
-		}
-		(void) unlink(tfile);
-		free(tfile);
-		free_hdrs(hdrs);
-		return(FALSE);
-	}
-
-	/* Form the default headers for the message */
-
-	if ((hdrs = form_hdrs(hdrs, mail_type)) == NULL) {
-		(void) unlink(tfile);
-		free(tfile);
-		return(FALSE);
-	}
-
-	/* See if we want to copy or edit the message */
-		
-	copy = (mail_type == T_REPLY) ? ask_copy() : orig_msg != NULL;
-		
-	/* Check if the user quit */
-
-	if (user_quit) {
-		(void) unlink(tfile);
-		free(tfile);
-		free_hdrs(hdrs);
-		return(FALSE);
-	}
-
-	/* Check what signature file to use */
-
-	if ((sigfile = get_sigfile(mail_type)) == NULL && user_quit) {
-		(void) unlink(tfile);
-		free(tfile);
-		free_hdrs(hdrs);
-		return(FALSE);
-	}
-
-	/* If we are copying text then do so */
-
-	if (copy && !cp_text(orig_msg, tfile, mail_type)) {
-		(void) unlink(tfile);
-		free(tfile);
-		free_hdrs(hdrs);
 		return(FALSE);
 	}
 
 	/* Do any initial edit of the outgoing text and get the headers */
 
-	if ((hdrs = form_mail(tfile, sigfile, hdrs,
-			      mail_type, TRUE)) == NULL) {
-		(void) unlink(tfile);
-		free(tfile);
+	if (!comp->bounce && !comp->file && !comp->silent
+	    && !edit_composition(comp, get_vval(V_EDIT_IHDRS), TRUE)) {
+		/* Composition failed somehow */
+
+		free_composition(comp);
+		return(FALSE);
+	}
+
+	/* Attach any specified messages now */
+
+	if (attachments != NULL &&
+	    !attach_messages(comp, attachments,
+			     (mail_flags & SM_ALL_HEADERS))) {
+		/* Failed to attach the messages */
+
+		free_composition(comp);
 		return(FALSE);
 	}
 
 	/* Post silent mail now */
 
-	if (mail_type == T_SILENT) {
-		(void) post(tfile, hdrs, FALSE);
-		(void) unlink(tfile);
-		free(tfile);
-		free_hdrs(hdrs);
+	if (comp->silent) {
+		(void) post(comp, FALSE);
+		free_composition(comp);
 		return(TRUE);
 	}
 
@@ -259,241 +244,103 @@ int mail_type;
 
 	spellcheck = get_vtext(V_SPELLCHECK);
 
-	/* Summarise the message's headers */
+	/* Summarise the composition's headers */
 
-	summarize_mail(hdrs);
+	summarise_mail(comp);
+
+	/* Form the prompt for the user */
+
+	prompt = vstrcat("Send, Edit, Attachments, ", (spellcheck == NULL) ?
+			 "" : "Check Spelling, ", "List or Forget?", NULL);
 
 	/* Loop until the user selects send or forget */
 
 	for (;;) {
-		/* Prompt for what to do next */
-
-		msgl("Send, Edit, ", (spellcheck == NULL) ? "" :
-		     "Check Spelling, ", "List or Forget?", NULL);
-
 		/* Now handle the user response */
 
-		switch (mklower(get_key())) {
-		case 's':
+		switch (select_key(prompt, (spellcheck != NULL) ?
+				   SC_MAIL_OPTS : MAIL_OPTS, TRUE)) {
+		case SEND_MESSAGE:
 			/* Send the mail to the MTA */
 
-			status = post(tfile, hdrs, TRUE);
-			(void) unlink(tfile);
-			free(tfile);
-			free_hdrs(hdrs);
+			status = post(comp, TRUE);
+			free(prompt);
+			free_composition(comp);
 			return(status);
 
-		case 'e':
+		case EDIT_MESSAGE:
 			/* Edit the outgoing message again */
 
-			if ((hdrs = form_mail(tfile, NULL, hdrs, mail_type,
-					      FALSE)) == NULL) {
-				(void) unlink(tfile);
-				free(tfile);
+			if (!edit_composition(comp, V_TRUE, TRUE)) {
+				free_composition(comp);
 				return(FALSE);
 			}
-			summarize_mail(hdrs);
+			summarise_mail(comp);
 			break;
 
-		case 'c':
+		case ATTACHMENTS:
+			/* Drop into compose mode on the message */
+
+			if (!compose_mail(comp)) {
+				free_composition(comp);
+				return(FALSE);
+			}
+			summarise_mail(comp);
+			break;
+
+		case CHECK_SPELLING:
 			/* Spell check the outgoing message */
 
 			if (spellcheck != NULL) {
-				spell_check_mail(spellcheck, tfile, hdrs);
-				summarize_mail(hdrs);
-			} else {
-				tbeep();
+				spell_check_mail(comp, spellcheck);
+				summarise_mail(comp);
 			}
 			break;
 
-		case 'l':
+		case LIST_MESSAGE:
 			/* List the message to typeout */
 
-			list_mail(tfile, hdrs);
-			summarize_mail(hdrs);
+			list_mail(comp);
+			summarise_mail(comp);
 			break;
 
-		case 'f':
+		case FORGET_MESSAGE:
 			/* Forget about the message */
 
 			msg("(Mail forgotten)");
-			(void) unlink(tfile);
-			free(tfile);
-			free_hdrs(hdrs);
+
+			/* Fall through into quit case */
+
+		case EOF:
+			free_composition(comp);
 			return(FALSE);
-
-		default:
-			/* Invalid user response */
-
-			tbeep();
-			break;
 		}
 	}
 	/*NOTREACHED*/
 }
 /****************************************************************************/
-static int reset_stdin()
+static void summarise_mail(comp)
+COMPOSITION *comp;
 {
-	/* Reset stdin to the terminal */
-
-	char *term;
-
-	/* Get the name of the terminal from stdout */
-
-	term = ttyname(fileno(stdout));
-
-	/* Reopen stdin to the terminal */
-
-	if (freopen(term, "r", stdin) == NULL) {
-		emsgl("Can't open ", term, ": ", strerror(errno), NULL);
-		return(FALSE);
-	}
-
-	/* All went ok */
-
-	return(TRUE);
-}
-/****************************************************************************/
-static int ask_copy()
-{
-	/* Ask the user whether to copy message on reply */
-
-	int yes;
-
-	switch (get_vval(V_COPY)) {
-	case V_FALSE:
-		return(FALSE);
-	case V_TRUE:
-		return(TRUE);
-	}
-
-	/* Ask if we want to copy */
-
-	yes = confirm("Copy message? ", FALSE);
-	return(yes);
-}
-/****************************************************************************/
-static char *get_sigfile(mail_type)
-int mail_type;
-{
-	/* Determine and return the user's signature file */
-
-	char *sigfile, *prompt;
-	char *base_prompt = "Signature file: ";
-
-	/* Get the users's default signature file */
-
-	sigfile = get_vtext(V_SIGFILE);
-
-	/* Check if a signature file is defined or required */
-
-	if (sigfile == NULL || mail_type == T_BOUNCE) {
-		return(NULL);
-	}
-
-	/* If ask: is not set then just return sigfile */
-
-	if (strncasecmp(sigfile, V_ASK_SIGFILE, strlen(V_ASK_SIGFILE))) {
-		return(sigfile);
-	}
-
-	/* Get the default signature file */
-
-	sigfile = sigfile + strlen(V_ASK_SIGFILE);
-
-	/* If we are mailing silently then return the default signature */
-
-	if (mail_type == T_SILENT) {
-		return(sigfile);
-	}
-
-	/* Ask the user which signature file */
-
-	prompt = xstrdup(base_prompt);
-	sigfile = get_ecstr(NULL, prompt, sigfile, fn_complete, C_CAUTIOUS);
-	free(prompt);
-
-	return(sigfile);
-}
-/****************************************************************************/
-static int cp_text(orig_msg, tfile, mail_type)
-MESSAGE *orig_msg;
-char *tfile;
-int mail_type;
-{
-	/* Copy the text of orig_msg into tfile */
-	
-	char *hlist, *prefix, *preface;
-	int  hskip, status;
-	FILE *fp;
-
-	/* Open the file */
-
-	if ((fp = fopen(tfile, "w")) == NULL) {
-		emsgl("Can't open ", tfile, ": ", strerror(errno), NULL);
-		return(FALSE);
-	}
-
-	/* Decide which headers to copy */
-
-	hskip = (mail_type == T_BOUNCE) ? HS_FROM : HS_COPY;
-	hlist = (mail_type == T_BOUNCE) ? NULL : get_vtext(V_TOCOPY);
-
-	/* Determine the prefix for the copy */
-
-	prefix = (mail_type == T_REPLY || mail_type == T_FORWARD)
-		? get_vtext(V_COPY_PFX) : NULL;
-
-	/* Preface the copied text if required */
-
-	if (mail_type != T_BOUNCE && (preface = get_vtext(V_PREFACE)) != NULL
-	    && (preface = unmessage(orig_msg, preface, FOLD_WIDTH)) != NULL
-	    && (fputs(preface, fp) == EOF || fputc('\n', fp) == EOF)) {
-		/* Error writing copy preface */
-
-		emsgl("Error writing temporary file: ",
-		      strerror(errno), NULL);
-		(void) fclose(fp);
-		return(FALSE);
-	}
-
-	/* Write the text */
-
-	if (status = write_text(fp, orig_msg, prefix, hlist, hskip, FALSE)) {
-		/* Error writing copied text */
-
-		emsgl("Error writing temporary file: ",
-		      strerror(status), NULL);
-		(void) fclose(fp);
-		return(FALSE);
-	}
-
-	/* Close the file and return ok */
-
-	(void) fclose(fp);
-	return(TRUE);
-}
-/****************************************************************************/
-static void summarize_mail(hdrs)
-HEADER *hdrs;
-{
-	/* Summarize the mail; to be sent via showtext */
+	/* Summarise the mail; to be sent via showtext */
 
 	HEADER *h;
 
-	/* First Let the user know what they can see */
+	/* First let the user know what they can see */
 
 	showtext("Summary of message headers: \n\n");
 
 	/* Then list the headers via showtext */
 
-	for (h = hdrs; !user_quit && h != NULL; h = h->next) {
+	for (h = comp->headers; !user_quit && h != NULL; h = h->next) {
 		/* Should this header be listed? */
 
-		if (h->edit && h->hdr_text != NULL) {
-			showtext(h->hdr_name);
+		if (h->edit && h->text != NULL) {
+			/* Show the header */
+
+			showtext(h->name);
 			showtext(" ");
-			showtext(h->hdr_text);
+			showtext(h->text);
 			showtext("\n");
 		}
 	}
@@ -505,46 +352,48 @@ HEADER *hdrs;
 	return;
 }
 /****************************************************************************/
-static void list_mail(tfile, hdrs)
-char *tfile;
-HEADER *hdrs;
+static void list_mail(comp)
+COMPOSITION *comp;
 {
 	/* List the mail to be sent to typeout */
 
-	char buf[BUFSIZ];
-	FILE *fp;
+	char *ctype, *enc, *slash;
+	int textual;
 	HEADER *h;
 
+	/* Find out if the file is textual, and how it's encoded */
+
+	ctype = c_contype(get_header(comp, CONTENT_TYPE));
+	slash = (ctype != NULL) ? strchr(ctype, '/') : NULL;
+	textual = (ctype == NULL || slash == NULL ||
+		   !strncasecmp(ctype, TEXT_TYPE, slash - ctype));
+	enc = c_encoding(get_header(comp, C_T_ENCODING));
+	enc = (enc != NULL || textual) ? enc : xstrdup(BASE64);
+	free(ctype);
+
 	/* First list the headers */
 
-	for (h = hdrs; !user_quit && h != NULL; h = h->next) {
+	for (h = comp->headers; !user_quit && h != NULL; h = h->next) {
 		/* Show the header if it is set */
 
-		if (h->hdr_text != NULL) {
-			typeout(h->hdr_name);
+		if (h->text != NULL) {
+			/* Show the header line */
+
+			typeout(h->name);
 			typeout(" ");
-			typeout(h->hdr_text);
+			typeout(h->text);
 			typeout("\n");
 		}
 	}
 
-	/* Now list the message, if any */
-
-	if (!user_quit && (fp = fopen(tfile, "r")) != NULL) {
-		/* Show the blank line */
+	/* Now list the decoded message body */
 
+	if (!user_quit) {
+		/* Show a blank line and then the decoded body */
 		typeout("\n");
-
-		/* And the the message text */
-
-		while (!user_quit && fgets(buf, BUFSIZ, fp) != NULL) {
-			typeout(buf);
-		}
-
-		/* Now close the file */
-
-		(void) fclose(fp);
+		show_decoded_text(comp->body, enc, textual);
 	}
+	free(enc);
 
 	/* End the typeout and reset the quit flag */
 
@@ -553,25 +402,51 @@ HEADER *hdrs;
 	return;
 }
 /****************************************************************************/
-void spell_check_mail(spellcheck, tfile, hdrs)
-char *spellcheck, *tfile;
-HEADER *hdrs;
+static void spell_check_mail(comp, spellcheck)
+COMPOSITION *comp;
+char *spellcheck;
 {
-	/* Interactively spell-check an outgoing message */
+	/* Interactively spell-check a message */
 
-	char *cmd;
+	char *tfile, *cmd;
 	ATIMER tbuf;
 
+	/* Get the name of a temporary file to use */
+
+	if ((tfile = tempnam(TFILEDIR, TFILEPFX)) == NULL) {
+		emsgl("Can't create temporary file: ", strerror(errno), NULL);
+	}
+
+	/* Write the decoded message to a file and spell check it */
+
+	if (!save_composition(comp, tfile, CS_BODY | CS_EDIT)) {
+		/* Couldn't write the temp file, update the display */
+
+		summarise_mail(comp);
+
+		/* And pause so the user can see the message */
+
+		(void) save_atimer(&tbuf);
+		(void) sleep(ECHO_DELAY);
+		(void) restore_atimer(&tbuf);
+
+		/* Clean up and fail */
+
+		(void) unlink(tfile);
+		free(tfile);
+		return;
+	}
+
 	/* Build the command to run the spell checker */
 
 	cmd = vstrcat(spellcheck, " ", tfile, NULL);
 
 	/* Now run the interactive spell checker */
 
-	if (!shellout(cmd, FALSE)) {
+	if (!shellout(cmd, FALSE, FALSE, NULL, NULL)) {
 		/* Spell check failed, update the display */
 
-		summarize_mail(hdrs);
+		summarise_mail(comp);
 
 		/* And pause so the user can see the message */
 
@@ -582,13 +457,14 @@ HEADER *hdrs;
 
 	/* Clean up and return */
 
+	(void) unlink(tfile);
+	free(tfile);
 	free(cmd);
 	return;
 }
 /****************************************************************************/
-static int post(mailfile, hdrs, verbose)
-char *mailfile;
-HEADER *hdrs;
+static int post(comp, verbose)
+COMPOSITION *comp;
 int verbose;
 {
 	/* Send mail via the MTA */
@@ -597,13 +473,12 @@ int verbose;
 	char *ofolder, *prompt;
 	int status, othreshold;
 	unsigned mlines;
-	FILE *fp;
 
 	/* Get the folder and threshold for saving outbound messages */
 
 	ofolder = get_vtext(V_OUTBOUND);
 	othreshold = get_vval(V_THRESHOLD);
-	mlines = mail_size(mailfile);
+	mlines = mail_size(comp);
 
 	/* Check if we want to save a long message */
 
@@ -628,41 +503,31 @@ int verbose;
 
 	msg("Sending mail...");
 
-	/* Open the mail body file */
-
-	if ((fp = fopen(mailfile, "r")) == NULL) {
-		emsgl("Can't open temporary file: ", strerror(errno), NULL);
-		return(FALSE);
-	}
-
 	/* Set up the command we're going to execute */
 
-	argv = mailargs(hdrs);
+	argv = mailargs(comp);
 
 	/* Add any headers needed before we send */
 
-	hdrs = posting_hdrs(hdrs);
+	add_posting_headers(comp);
 
 	/* Now submit the mail by the required method */
 
 #ifdef MTA_IS_SMTP
 	/* Submit the mail to a remote host via SMTP */
 
-	status = smtp_deliver(get_vtext(V_SMTP_HOST),
-			get_addr(), argv, fp, hdrs);
+	status = smtp_deliver(comp, get_vtext(V_SMTP_HOST),
+			      get_addr(), argv);
 #else /* ! MTA_IS_SMTP */
 	/* Exec the MTA and pipe the mail into it */
 
-	status = exec_mta(argv, fp, hdrs);
+	status = exec_mta(comp, argv);
 #endif /* MTA_IS_SMTP */
 
-	/* Close the mail body file */
-
-	(void) fclose(fp);
-
 	/* Save the message to the outbound folder if required */
-	
-	if (status && save_outbound(ofolder, mailfile, hdrs)) {
+
+	if (status && (ofolder == NULL
+		       || save_composition(comp, ofolder, CS_MBOX))) {
 		cmsg(" Done");
 	}
 
@@ -675,14 +540,12 @@ int verbose;
 	return(status);
 }
 /****************************************************************************/
-static int exec_mta(argv, mail_fp, hdrs)
+static int exec_mta(comp, argv)
+COMPOSITION *comp;
 char **argv;
-FILE *mail_fp;
-HEADER *hdrs;
 {
 	/* Actually send the mail, by spawning the MTA */
 
-	char buf[BUFSIZ];
 	int fds[2], fd;
 	int status = 0;
 	FILE *fp;
@@ -698,7 +561,7 @@ HEADER *hdrs;
 
 	switch(fork()) {
 	case -1:						/* Failed */
-		emsgl("Mail delivery failed: ", strerror(status), NULL);
+		emsgl("Mail delivery failed: ", strerror(errno), NULL);
 		return(FALSE);
 
 	case 0:							/* Child */
@@ -733,10 +596,7 @@ HEADER *hdrs;
 
 		/* Write the message headers and body to the pipe */
 
-		(void) cp_hdrs(fp, hdrs, TRUE);
-		while (fgets(buf, BUFSIZ, mail_fp) != NULL) {
-			(void) fputs(buf, fp);
-		}
+		(void) write_composition(comp, fp, NULL, CS_SAVE);
 
 		/* Close the pipe and wait for the child */
 
@@ -750,235 +610,28 @@ HEADER *hdrs;
 	/*NOTREACHED*/
 }
 /****************************************************************************/
-static unsigned mail_size(mailfile)
-char *mailfile;
+static unsigned mail_size(comp)
+COMPOSITION *comp;
 {
-	/* Return the number of lines in mailfile */
+	/* Return the number of lines the composition's body */
 
-	char buf[LINESIZE + 1];
 	unsigned lines = 0;
-	FILE *fp;
+	TEXTLINE *t;
 
-	/* Open the mail body file */
+	/* Count the lines in the body */
 
-	if ((fp = fopen(mailfile, "r")) == NULL) {
-		return(0);
-	}
-
-	/* Scan the file for lines */
-
-	while (fgets(buf, LINESIZE, fp) != NULL) {
+	for (t = comp->body; t != NULL; t = t->next) {
 		lines++;
 	}
 
-	/* Close the file and return the number of lines */
+	/* And return the number of lines */
 
-	(void) fclose(fp);
 	return(lines);
 }
 /****************************************************************************/
-static int save_outbound(outbound, mailfile, hdrs)
-char *outbound, *mailfile;
-HEADER *hdrs;
-{
-	/* Save a copy of the message to any outbound folder */
-
-	char *fromline, *buf;
-	int status;
-	FILE *fp, *ofp;
-
-	/* If no outbound folder then return TRUE */
-
-	if (outbound == NULL) {
-		return(TRUE);
-	}
-
-	/* Open the mail body file */
-
-	if ((fp = fopen(mailfile, "r")) == NULL) {
-		emsgl("Error opening temporary file: ",
-		      strerror(errno), NULL);
-		return(FALSE);
-	}
-
-	/* Open the outbound folder to append */
-
-	if ((ofp = fopen(outbound, "a")) == NULL) {
-		emsgl("Can't open ", outbound, ": ",
-		      strerror(errno), NULL);
-		(void) fclose(fp);
-		return(FALSE);
-	}
-
-	/* Add any headers needed before we save */
-
-	hdrs = saving_hdrs(hdrs, mailfile);
-
-#ifdef MTA_MMDF_FORMAT
-	/* Write a header delimiter to the outbound folder */
-
-	if (fputs(MMDF_DELIM, ofp) == EOF) {
-		emsgl("Error writing ", outbound,
-		      ": ", strerror(errno), NULL);
-		(void) fclose(fp);
-		(void) fclose(ofp);
-		return(FALSE);
-	}
-#endif /* ! MTA_MMDF_FORMAT */
-
-	/* Write a From line to the outbound folder */
-
-	fromline = vstrcat(MFROM, get_addr(), " ", 
-			   strudate(date_now()), "\n", NULL);
-
-	if (fputs(fromline, ofp) == EOF) {
-		emsgl("Error writing ", outbound,
-		      ": ", strerror(errno), NULL);
-		free(fromline);
-		(void) fclose(fp);
-		(void) fclose(ofp);
-		return(FALSE);
-	}
-	free(fromline);
-
-	/* Append the headers to the outbound folder */
-
-	if (status = cp_hdrs(ofp, hdrs, TRUE)) {
-		emsgl("Error writing ", outbound,
-		      ": ", strerror(status), NULL);
-		(void) fclose(fp);
-		(void) fclose(ofp);
-		return(FALSE);
-	}
-
-	/* And the message text */
-
-	while ((buf = get_line(fp, FALSE)) != NULL) {
-#ifndef MTA_MMDF_FORMAT
-		/* Quote Delimiters in the message body */
-
-		if (!strncmp(buf, MFROM, strlen(MFROM))
-		    && fputs(FROM_QUOTE, ofp) == EOF) {
-			emsgl("Error writing ", outbound,
-			      ": ", strerror(errno), NULL);
-			free(buf);
-			(void) fclose(fp);
-			(void) fclose(ofp);
-			return(FALSE);
-		}
-#endif /* ! MTA_MMDF_FORMAT */
-
-		/* Now write the line to the folder */
-
-		if (fputs(buf, ofp) == EOF) {
-			emsgl("Error writing ", outbound,
-			      ": ", strerror(errno), NULL);
-			free(buf);
-			(void) fclose(fp);
-			(void) fclose(ofp);
-			return(FALSE);
-		}
-
-		/* And free the buffer */
-
-		free(buf);
-	}
-
-	/* Close the mail body file */
-
-	(void) fclose(fp);
-
-	/* Append a delimiter to the folder */
-
-#ifdef MTA_MMDF_FORMAT
-	if (fputs(MMDF_DELIM, ofp) == EOF) {
-		emsgl("Error writing ", outbound,
-		      ": ", strerror(errno), NULL);
-		(void) fclose(ofp);
-		return(FALSE);
-	}
-#else /* ! MTA_MMDF_FORMAT */
-	if (putc('\n', ofp) == EOF) {
-		emsgl("Error writing ", outbound,
-		      ": ", strerror(errno), NULL);
-		(void) fclose(ofp);
-		return(FALSE);
-	}
-#endif /* ! MTA_MMDF_FORMAT */
-
-	/* Close the outbound folder and return */
-
-	(void) fclose(ofp);
-	return(TRUE);
-}
-/****************************************************************************/
-int cp_hdrs(fp, hdrs, send)
-FILE *fp;
-HEADER *hdrs;
-int send;
-{
-	/* Write the headers contained in hdrs to the file pointer fp */
-
-	HEADER *h;
-	
-	/* Loop over the headers */
-
-	for (h = hdrs; h != NULL; h = h->next) {
-		/* Do we need to send this header? */
-
-		if (send && (h->hdr_text != NULL)
-		    || !send && h->edit && h->show) {
-#ifdef NO_MTA_CC
-			/* Handle the Bcc header when sending */
-
-			if (send && (!strcasecmp(h->hdr_name, BCC) ||
-				     !strcasecmp(h->hdr_name, RESENT_BCC))) {
-				/* Might want an empty Bcc: header */
-
-				if (get_hdr(TO, hdrs) == NULL &&
-				    get_hdr(RESENT_TO, hdrs) == NULL
-				    && get_hdr(CC, hdrs) == NULL &&
-				    get_hdr(RESENT_CC, hdrs) == NULL
-				    && (fputs(h->hdr_name, fp) == EOF
-					|| putc('\n', fp) == EOF)) {
-					/* Error writing the header */
-
-					return(errno);
-				}
-
-				/* Skip the rest of the header text */
-
-				continue;
-			}
-#endif /* NO_MTA_CC */
-			/* Put the header name and text */
-
-			if (fputs(h->hdr_name, fp) == EOF
-			    || putc(' ', fp) == EOF) {
-				return(errno);
-			}
-			if (h->hdr_text != NULL && 
-			    fputs(h->hdr_text, fp) == EOF) {
-				return(errno);
-			}
-			if (putc('\n', fp) == EOF) {
-				return(errno);
-			}
-		}
-	}
-
-	/* Output a blank line following the headers */
-
-	if (putc('\n', fp) == EOF) {
-		return(errno);
-	}
-
-	return(0);
-}
-/****************************************************************************/
 /*ARGSUSED*/
-static char **mailargs(hdrs)
-HEADER *hdrs;
+static char **mailargs(comp)
+COMPOSITION *comp;
 {
 	/*
 	 * Form the shell command line to exec MAILPROG.
@@ -1007,7 +660,7 @@ HEADER *hdrs;
 #ifdef MTA_NEEDS_ARGS
 	/* Find the headers specifying destination */
 
-	users = set_addrs(hdrs);
+	users = set_addrs(comp);
 
 	/* Add the destinations to the argument vector */
 
@@ -1022,9 +675,13 @@ HEADER *hdrs;
 /****************************************************************************/
 #ifndef MTA_IS_SMTP
 /*VARARGS 1*/
+#ifdef __STDC__
+static char **argvec(char *prog, char *arg, ...)
+#else /* ! __STDC__ */
 static char **argvec(prog, va_alist)
 char *prog;
 va_dcl
+#endif /* ! __STDC__ */
 {
 	/*
 	 * Build an argument vector for prog and the arguments.
@@ -1036,14 +693,22 @@ va_dcl
 	 * NB: This function does not handle quoting or backslash escapes.
 	 */
 
-	char *line, **argv = NULL;
-	char *arg, *space;
+	char *line, **argv = NULL, *space;
 	int argno;
 	va_list arglist;
 
+#ifndef __STDC__
+	char *arg;
+#endif /* ! __STDC__ */
+
 	/* Initialise vararg handling */
 
+#ifdef __STDC__
+	va_start(arglist, arg);
+#else /* ! __STDC__ */
 	va_start(arglist);
+	arg = va_arg(arglist, char *);
+#endif /* ! __STDC__ */
 	
 	/* Copy prog into line */
 
@@ -1052,10 +717,16 @@ va_dcl
 
 	/* Now loop through the arguments, adding them to line */
 
-	while ((arg = va_arg(arglist, char *)) != NULL) {
+	while (arg != NULL) {
+		/* Add this argument to the line */
+
 		line = xrealloc(line, strlen(line) + strlen(arg) + 2);
 		(void) strcat(line, " ");
 		(void) strcat(line, arg);
+
+		/* Move to the next argument */
+
+		arg = va_arg(arglist, char *);
 	}
 
 	/* Clean up the varargs handling */
@@ -1115,10 +786,10 @@ va_dcl
 #endif /* ! MTA_IS_SMTP */
 /****************************************************************************/
 #ifdef MTA_NEEDS_ARGS
-static char *set_addrs(hdrs)
-HEADER *hdrs;
+static char *set_addrs(comp)
+COMPOSITION *comp;
 {
-	/* Form a list of destinations from hdrs and return it */
+	/* Form a list of destinations from composition's headers */
 
 	char *users = NULL, *new_users;
 	int resent = FALSE, new_resent;
@@ -1126,7 +797,7 @@ HEADER *hdrs;
 
 	/* Loop through the headers handling each one */
 
-	for (h = hdrs; h != NULL; h = h->next) {
+	for (h = comp->headers; h != NULL; h = h->next) {
 		/* Initialise for this header */
 
 		new_users = NULL;
@@ -1134,23 +805,23 @@ HEADER *hdrs;
 
 		/* Is it a Resent-To: or To: header? */
 
-		if (!strcasecmp(h->hdr_name, RESENT_TO)) {
-			new_users = h->hdr_text;
+		if (!strcasecmp(h->name, RESENT_TO)) {
+			new_users = h->text;
 			new_resent = TRUE;
-		} else if (!resent && !strcasecmp(h->hdr_name, TO)) {
-			new_users = h->hdr_text;
+		} else if (!resent && !strcasecmp(h->name, TO)) {
+			new_users = h->text;
 			new_resent = FALSE;
 
 #ifdef NO_MTA_CC
 		/* The MTA doesn't handle Cc or Bcc, do it ourselves */
 
-		} else if (!strcasecmp(h->hdr_name, RESENT_CC)
-			   || !strcasecmp(h->hdr_name, RESENT_BCC)) {
-			new_users = h->hdr_text;
+		} else if (!strcasecmp(h->name, RESENT_CC)
+			   || !strcasecmp(h->name, RESENT_BCC)) {
+			new_users = h->text;
 			new_resent = TRUE;
-		} else if (!resent && (!strcasecmp(h->hdr_name, CC) ||
-				       !strcasecmp(h->hdr_name, BCC))) {
-			new_users = h->hdr_text;
+		} else if (!resent && (!strcasecmp(h->name, CC) ||
+				       !strcasecmp(h->name, BCC))) {
+			new_users = h->text;
 			new_resent = FALSE;
 #endif /* NO_MTA_CC */
 		}
