@@ -1,5 +1,5 @@
 /* Typeout.c - Typeout handling for af.
-   Copyright (C) 1991, 1992, 1993, 1994, 1995, 1996, 1997 Malc Arnold.
+   Copyright (C) 1991 - 2002 Malc Arnold.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
 
 #include <stdio.h>
 #include <ctype.h>
+#include <errno.h>
 #include <regex.h>
 #include "af.h"
 #include "keyseq.h"
@@ -26,13 +27,15 @@
 #include "variable.h"
 #include "mode.h"
 #include "complete.h"
+#include "typeout.h"
 #include STRING_HDR
 
 /****************************************************************************/
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: typeout.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: typeout.c,v 2.5 2002/09/08 21:26:22 malc Exp $";
+static char *TypeoutId = TYPEOUTID;
 #endif /* ! lint */
 
 /****************************************************************************/
@@ -45,11 +48,12 @@ extern char *get_vtext(), *strkey(), *strchar();
 extern char *utos();
 extern int atoi(), get_vval(), is_blank();
 extern int charlen(), tlines(), tcols();
+extern int vtupdate();
 extern unsigned cmodes();
-extern void free(), free_forms(), free_mtext();
-extern void msg(), msgl(), cmsg(), emsg(), emsgl();
-extern void clearmsg(), display(), interactive();
-extern void vtupdate(), vtredraw(), hide_cursor();
+extern void free(), free_forms(), msg(), msgl();
+extern void cmsg(), emsg(), emsgl(), clearmsg();
+extern void alldisplay(), interactive();
+extern void vtredraw(), hide_cursor();
 extern void type_show(), type_display();
 extern FORM *exec_type_key();
 extern ARGUMENT *form_or_arg();
@@ -58,13 +62,14 @@ extern CLIST *fn_complete();
 /* Local function declarations */
 
 unsigned to_lines();
-void typeout();
+void typebin(), showbin(), column();
 static char *indent();
-static int scroll(), search();
+static int scroll(), search(), section();
 static void show_to_file(), init_typeout(), end_typeout();
 static void set_point(), set_search(), trim_blank_lines();
-static MSG_TEXT *show_text(), *format_text();
-static MSG_TEXT *add_text(), *find_line();
+static void free_type_lines();
+static TYPE_LINE *show_line(), *format_line();
+static TYPE_LINE *add_line(), *find_line();
 
 /****************************************************************************/
 /* Import the system error number */
@@ -85,8 +90,10 @@ extern int user_quit;
 static char *type_file = NULL;		/* File for typeout redirection */
 static FILE *type_fp = NULL;		/* Pointer to redirection file */
 static int type_error = 0;		/* Error while writing typeout */
-static int type_iso = FALSE;		/* Show ISO characters in typeout? */
-static MSG_TEXT *type_lines = NULL;	/* List of lines we're typing out */
+static int type_sectioned = FALSE;	/* Are sections currently active? */
+static int type_showing = FALSE;	/* This typeout is in show mode */
+static TYPE_LINE *type_lines = NULL;	/* First line in current section */
+static TYPE_LINE *type_last = NULL;	/* Last line in current section */
 static char *type_search = NULL;	/* Expression we're searching for */
 static regex_t *type_expr = NULL;	/* Compiled search expression */
 static int type_repeat = FALSE;		/* Repeat search in typeout? */
@@ -96,9 +103,12 @@ static int type_point = 0;		/* Number of first line on screen */
 static int type_mark = 0;		/* Number of current mark line */
 static int type_tenths = 0;		/* New position within typeout */
 
+static int type_sections = 0;		/* Which section to show next */
 static int type_needed = 0;		/* Lines needed to fill window */
 static int type_more = FALSE;		/* Need to read an extra line? */
 static int type_end = FALSE;		/* Read to end of typeout? */
+static int type_have_prev_sec = FALSE;	/* Is there a previous section? */
+static int type_have_next_sec = FALSE;	/* Is there a next section? */
 static int type_indent = FALSE;		/* Are we indenting lines? */
 
 static unsigned old_modes = M_NULL;	/* Modes active before typeout */
@@ -167,16 +177,52 @@ char *output_type;
 	return(TRUE);
 }
 /****************************************************************************/
-void iso_typeout()
+void sectioned_typeout(active, have_prev, have_next)
+int active, have_prev, have_next;
 {
-	/* Make the next typeout show iso 8-bit characters */
+	/*
+	 * Activate or deactivate sectioned typeout.  This allows you to
+	 * call sectioned_typeout, whereby a whole set of typeouts get
+	 * merged into a single, coherent, typeout session.  In addition
+	 * to the obvious effect, calling this routine also makes typeout
+	 * more persistent; avoiding redraws between sections.
+	 *
+	 * This routine should be called before typeout is in use, and
+	 * before each section of the typeout.
+	 */
+
+	/* Redraw after the typeout if required */
+
+	if (!(type_sectioned = active) && cwin != NULL) {
+		alldisplay(NULL);
+	}
+
+	/* Clear the number of sections to move */
+
+	type_sections = 0;
+
+	/* Set the next and previous section flags */
+
+	type_have_prev_sec = have_prev;
+	type_have_next_sec = have_next;
+
+	/* That's all there is */
 
-	type_iso = TRUE;
 	return;
 }
 /****************************************************************************/
 void typeout(text)
 char *text;
+{
+	/* Display non-binary text via typeout mode */
+
+	typebin(text, (text != NULL) ? strlen(text) : 0);
+	return;
+}
+/****************************************************************************/
+void typebin(text, len)
+char *text;
+size_t len;
 {
 	/* 
 	 * Display text via typeout mode.  The overall structure
@@ -188,18 +234,18 @@ char *text;
 	 * Typeout is terminated when text is equal to NULL.
 	 */
 
-	static int type_active = FALSE;
+	static int typeout_active = FALSE;
 	FORM *status;
 
 	/* Recursive use of typeout must fail */
 
-	if (type_active && text != NULL && !user_quit) {
+	if (typeout_active && text != NULL && !user_quit) {
 		/* Let the user know what's happened */
 
 		emsg("Command attempted to use typeout while in typeout");
 		user_quit = TRUE;
 		return;
-	} else if (type_active || user_quit) {
+	} else if (typeout_active || user_quit) {
 		/* Ignore the typeout */
 
 		return;
@@ -208,38 +254,45 @@ char *text;
 	/* Let typeout to a file look after itself */
 
 	if (type_fp != NULL) {
-		show_to_file(text);
+		show_to_file(text, len);
 		return;
 	}
 
 	/* Initialise typeout if required */
 
-	if (type_lines == NULL && !type_needed
-	    && !type_more && text != NULL) {
-		init_typeout(TRUE);
+	if (type_lines == NULL && !type_sections &&
+	    !type_needed && !type_more && text != NULL) {
+		init_typeout(FALSE);
 	}
 
 	/* Now add the text to the list and display it */
 
-	type_lines = show_text(type_lines, text);
-	if (type_end) {
+	if (!type_sections) {
+		type_lines = show_line(type_lines, text, len);
+	}
+
+	/* Trim unwanted blank lines from the list */
+
+	if (type_end && !type_sections) {
 		trim_blank_lines();
 	}
 
 	/* Set up ready for processing commands */
 
-	if (type_lines != NULL && !type_needed && !type_more) {
-		/* Set up modes and point */
+	if (type_lines != NULL && !type_sections
+	    && !type_needed && !type_more) {
+		/* Set up the initial point */
 
 		set_point(type_point, 0);
 	}
 
 	/* Now process user input if required */
 
-	while (type_lines != NULL && !type_needed && !type_more) {
+	while (type_lines != NULL && !type_sections
+	       && !type_needed && !type_more) {
 		/* Disable typeout while the command executes */
 
-		type_active = TRUE;
+		typeout_active = TRUE;
 
 		/* Execute the command and check for quit */
 
@@ -250,7 +303,7 @@ char *text;
 		/* Now clean up and reenable typeout */
 
 		free_forms(status);
-		type_active = FALSE;
+		typeout_active = FALSE;
 	}
 
 	return;
@@ -258,6 +311,16 @@ char *text;
 /****************************************************************************/
 void showtext(text)
 char *text;
+{
+	/* Display text via typeout show mode */
+
+	showbin(text, (text != NULL) ? strlen(text) : 0);
+	return;
+}
+/****************************************************************************/
+void showbin(text, len)
+char *text;
+size_t len;
 {
 	/* 
 	 * Display text via typeout show mode, where only the
@@ -273,13 +336,14 @@ char *text;
 
 	/* Initialise typeout if required */
 
-	if (type_lines == NULL && !type_needed && !type_more) {
-		init_typeout(FALSE);
+	if (type_lines == NULL && !type_needed
+	    && !type_more && text != NULL) {
+		init_typeout(TRUE);
 	}
 
 	/* Now add the text to the list and display it */
 
-	type_lines = show_text(type_lines, text);
+	type_lines = show_line(type_lines, text, len);
 
 	/* Now handle displaying the first page */
 
@@ -287,8 +351,8 @@ char *text;
 		/* Display the page and end typeout */
 
 		(void) cmodes(M_TYPEOUT | M_SHOW);
-		set_point(type_point, 0);
-		vtupdate(TRUE);
+		type_display(type_lines);
+		(void) vtupdate(TRUE);
 		hide_cursor();
 		end_typeout();
 	}
@@ -301,24 +365,13 @@ char *name, *entry;
 {
 	/* Print name followed by a tabulated entry to typeout */
 
-	char *key;
-	int len = 0;
+	/* Set up the name as the first column */
 
-	/* Print the name of the table line */
-
-	typeout(name);
-
-	/* Calculate the length of the name */
-
-	for (key = name; *key != '\0'; key++) {
-		len += charlen(*key, len, type_iso);
-	}
+	column(name);
 
 	/* And print the entry if non-null */
 
 	if (entry != NULL) {
-		type_indent = TRUE;
-		typeout(indent(len));
 		typeout(entry);
 	}
 	typeout("\n");
@@ -326,6 +379,31 @@ char *name, *entry;
 	return;
 }
 /****************************************************************************/
+void column(text)
+char *text;
+{
+	/* Use text as the base column for a table entry */
+
+	char *key;
+	int len = 0;
+
+	/* Print the text of the base column */
+
+	typeout(text);
+
+	/* Calculate the length of the name */
+
+	for (key = text; *key != '\0'; key++) {
+		len += charlen(*key, len);
+	}
+
+	/* Now set up the typeout indent */
+
+	type_indent = TRUE;
+	typeout(indent(len));
+	return;
+}
+/****************************************************************************/
 int error_in_typeout()
 {
 	/* Return any error encountered by the last typeout */
@@ -359,6 +437,13 @@ unsigned to_position()
 	return(type_point);
 }
 /****************************************************************************/
+unsigned to_sections()
+{
+	/* Return the number of sections requested by the user */
+
+	return(type_sections);
+}
+/****************************************************************************/
 int to_reading()
 {
 	/* Return whether we are still reading typeout */
@@ -463,30 +548,6 @@ FORM *forms;
 }
 /****************************************************************************/
 /*ARGSUSED*/
-FORM *to_next(seq, arg, forms)
-KEYSEQ *seq;
-ARGUMENT *arg;
-FORM *forms;
-{
-	/* Scroll up a line if possible */
-
-	int nlines, back;
-
-	/* Convert any form to an argument */
-
-	arg = form_or_arg(forms, arg);
-
-	/* Check for any argument */
-
-	nlines = (arg != NULL) ? arg->value : 1;
-	back = (arg != NULL && arg->negative);
-
-	/* Now do the scrolling and return status */
-
-	return(scroll(nlines, back, TRUE) ? c_t() : c_errored());
-}
-/****************************************************************************/
-/*ARGSUSED*/
 FORM *to_prev(seq, arg, forms)
 KEYSEQ *seq;
 ARGUMENT *arg;
@@ -511,6 +572,30 @@ FORM *forms;
 }
 /****************************************************************************/
 /*ARGSUSED*/
+FORM *to_next(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* Scroll up a line if possible */
+
+	int nlines, back;
+
+	/* Convert any form to an argument */
+
+	arg = form_or_arg(forms, arg);
+
+	/* Check for any argument */
+
+	nlines = (arg != NULL) ? arg->value : 1;
+	back = (arg != NULL && arg->negative);
+
+	/* Now do the scrolling and return status */
+
+	return(scroll(nlines, back, TRUE) ? c_t() : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
 FORM *to_start(seq, arg, forms)
 KEYSEQ *seq;
 ARGUMENT *arg;
@@ -572,6 +657,54 @@ FORM *forms;
 	return(c_t());
 }
 /****************************************************************************/
+/*ARGSUSED*/
+FORM *to_prev_sec(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* Move to the previous section in the typeout */
+
+	int nsecs, back;
+
+	/* Convert any form to an argument */
+
+	arg = form_or_arg(forms, arg);
+
+	/* Check for any argument */
+
+	nsecs = (arg != NULL) ? arg->value : 1;
+	back = (arg == NULL || !arg->negative);
+
+	/* Set the sections to move and return exit status if ok */
+
+	return(section(nsecs, back) ? NULL : c_errored());
+}
+/****************************************************************************/
+/*ARGSUSED*/
+FORM *to_next_sec(seq, arg, forms)
+KEYSEQ *seq;
+ARGUMENT *arg;
+FORM *forms;
+{
+	/* Move to the next section in the typeout */
+
+	int nsecs, back;
+
+	/* Convert any form to an argument */
+
+	arg = form_or_arg(forms, arg);
+
+	/* Check for any argument */
+
+	nsecs = (arg != NULL) ? arg->value : 1;
+	back = (arg != NULL && arg->negative);
+
+	/* Set the sections to move and return exit status if ok */
+
+	return(section(nsecs, back) ? NULL : c_errored());
+}
+/****************************************************************************/
 FORM *to_goto(seq, arg, forms)
 KEYSEQ *seq;
 ARGUMENT *arg;
@@ -654,13 +787,15 @@ FORM *forms;
 		return(c_errored());
 	}
 
-	/* Swap point and mark and redraw screen */
+	/* Swap the point and the mark */
 
 	old_point = type_point;
 	type_point = type_mark;
 	type_mark = old_point;
-	type_display(find_line(type_point));
 
+	/* And redraw the display */
+
+	type_display(find_line(type_point));
 	return(c_t());
 }
 /****************************************************************************/
@@ -704,22 +839,30 @@ FORM *forms;
 	if (!type_end) {
 		cmsg(" read to date");
 	}
+
+	/* And return success */
+
 	return(c_t());
 }
 /****************************************************************************/
-static void init_typeout(clear_echo_area)
-int clear_echo_area;
+static void init_typeout(showing)
+int showing;
 {
-	/* Initialise the typeout position variables */
+	/* Initialise the typeout and section variables */
 
-	type_lines = NULL;
+	type_showing = showing;
+	type_lines = type_last = NULL;
 	type_search = NULL;
 	type_expr = NULL;
 	type_count = type_tenths = 0;
 	type_point = type_mark = 0;
-	type_end = type_indent = FALSE;
+
+	/* Set up the reading details */
+
+	type_sections = 0;
 	type_needed = tlines() - 1;
-	type_more = FALSE;
+	type_end = type_more = FALSE;
+	type_indent = FALSE;
 
 	/* If we weren't interactive before, we are now */
 
@@ -738,7 +881,7 @@ int clear_echo_area;
 
 	/* Clear the echo area if required */
 
-	if (clear_echo_area) {
+	if (!showing) {
 		clearmsg();
 	}
 	return;
@@ -746,20 +889,20 @@ int clear_echo_area;
 /****************************************************************************/
 static void end_typeout()
 {
-	/* Clean up when typeout exits */
+	/* Clean up when typeout exits or handle end-of-section */
 
-	WINDOW *w = cwin;
+	if (type_lines != NULL) {
+		/* Free the typeout's lines and any partial line */
 
-	/* Free the line list */
+		free_type_lines(type_lines);
+		free_type_lines(format_line(NULL, NULL, 0));
+	}
 
-	free_mtext(type_lines);
-	type_lines = NULL;
+	/* Clean up the current values */
+
+	type_lines = type_last = NULL;
 	type_needed = 0;
-	type_more = type_iso = FALSE;
-
-	/* Ensure that any partial line is cleared */
-
-	free_mtext(format_text(NULL, NULL, 0));
+	type_more = FALSE;
 
 	/* Restore the old commands and buffer modes */
 
@@ -769,28 +912,25 @@ static void end_typeout()
 
 	/* Redraw the windows and return */
 
-	if (w != NULL) {
-		do {
-			display(w);
-			w = w->next;
-		} while (w != cwin);
+	if (!type_sectioned && cwin != NULL) {
+		alldisplay(NULL);
 	}
-
 	return;
 }
 /****************************************************************************/
-static MSG_TEXT *show_text(list, text)
-MSG_TEXT *list;
+static TYPE_LINE *show_line(list, text, len)
+TYPE_LINE *list;
 char *text;
+size_t len;
 {
 	/* Format text and add it to the line list */
 
 	int line = type_count + 1;
-	MSG_TEXT *lines, *l;
+	TYPE_LINE *lines, *l;
 
 	/* Format the text into the list of lines */
 
-	lines = format_text(list, text);
+	lines = format_line(list, text, len);
 
 	/* Now show the lines if required */
 
@@ -805,7 +945,7 @@ char *text;
 		/* Do we need to read an extra line? */
 
 		type_more = (!type_needed && l->next == NULL
-					&& is_blank(l->line));
+			     && is_blank(l->line));
 	}
 
 	/* Check for end-of-typeout */
@@ -840,17 +980,19 @@ char *text;
 	return((list == NULL) ? lines : list);
 }
 /****************************************************************************/
-static void show_to_file(text)
+static void show_to_file(text, len)
 char *text;
+size_t len;
 {
 	/* Just write the text to the file and return */
 
 	static char *separator = "\n";
-	MSG_TEXT *lines, *l;
+	TYPE_LINE *lines, *l;
 
 	/* Format the text or a separator into a list of lines */
 
-	lines = format_text(NULL, (text != NULL) ? text : separator);
+	lines = (text != NULL) ? format_line(NULL, text, len) :
+		format_line(NULL, separator, strlen(separator));
 
 	/* Now append the lines to the file */
 
@@ -867,7 +1009,7 @@ char *text;
 
 	/* Free the list of formatted lines */
 
-	free_mtext(lines);
+	free_type_lines(lines);
 
 	/* Handle confirmation at end-of-typeout */
 
@@ -882,17 +1024,18 @@ char *text;
 
 	if (text == NULL) {
 		(void) fclose(type_fp);
-		type_fp = NULL;
 		free(type_file);
+		type_fp = NULL;
 		type_file = NULL;
 	}
 
 	return;
 }
 /****************************************************************************/
-static MSG_TEXT *format_text(list, text)
-MSG_TEXT *list;
+static TYPE_LINE *format_line(list, text, len)
+TYPE_LINE *list;
 char *text;
+size_t len;
 {
 	/*
 	 * Format text into a linked list of lines.  Partial
@@ -904,14 +1047,14 @@ char *text;
 	static char *line = NULL;
 	static int col = 0;
 	char *cstr, wrap, *p, *q;
-	MSG_TEXT *first = NULL;
+	TYPE_LINE *first = NULL;
 
 	/* Handle partial lines at end-of-typeout */
 
 	if (text == NULL && col > 0) {
 		/* Add the partial line and reset */
 
-		first = add_text(list, first, line, col);
+		first = add_line(list, first, line, col);
 		list = (list == NULL) ? first : list;
 		line = NULL;
 		col = 0;
@@ -931,10 +1074,10 @@ char *text;
 
 	/* Now copy the text into the line */
 
-	for (p = text; *p != '\0'; p++) {
+	for (p = text; p < text + len; p++) {
 		/* Get the printable text of the character */
 
-		cstr = strchar(*p, col, type_iso);
+		cstr = strchar(*p, col);
 
 		/* Add non-newline characters to the line */
 
@@ -949,7 +1092,7 @@ char *text;
 
 				/* Add the line and reallocate it */
 
-				first = add_text(list, first, line, col);
+				first = add_line(list, first, line, col);
 				list = (list == NULL) ? first : list;
 				line = xmalloc(tcols());
 
@@ -973,7 +1116,7 @@ char *text;
 		if (*p == '\n') {
 			/* Add the line and reallocate it */
 
-			first = add_text(list, first, line, col);
+			first = add_line(list, first, line, col);
 			list = (list == NULL) ? first : list;
 			line = xmalloc(tcols());
 
@@ -989,20 +1132,24 @@ char *text;
 	return(first);
 }
 /****************************************************************************/
-static MSG_TEXT *add_text(list, first, line, len)
-MSG_TEXT *list, *first;
+static TYPE_LINE *add_line(list, first, line, len)
+TYPE_LINE *list, *first;
 char *line;
 int len;
 {
-	/* Add a line to the list of typeout lines */
+	/*
+	 * Add a line to the list of typeout lines.  Also has the
+	 * job of keeping the type_last pointer up to date.
+	 */
 
-	static MSG_TEXT *last = NULL;
-	MSG_TEXT *node;
+	TYPE_LINE *node;
 
 	/* Terminate and resize the line */
 
-	line[len] = '\0';
-	line = xrealloc(line, len + 1);
+	if (line != NULL) {
+		line[len] = '\0';
+		line = xrealloc(line, len + 1);
+	}
 
 	/* Update the typeout line count */
 
@@ -1010,16 +1157,20 @@ int len;
 
 	/* Allocate and fill the new node */
 
-	node = (MSG_TEXT *) xmalloc(sizeof(MSG_TEXT));
+	node = (TYPE_LINE *) xmalloc(sizeof(TYPE_LINE));
 	node->line = line;
-	node->next = NULL;
+	node->prev = node->next = NULL;
 
 	/* Append the line to the list */
 
 	if (list != NULL) {
-		last->next = node;
+		node->prev = type_last;
+		type_last->next = node;
 	}
-	last = node;
+
+	/* And update the last line */
+
+	type_last = node;
 
 	/* Return the updated first flag */
 
@@ -1092,7 +1243,7 @@ FORM *forms;
 	int expr_flags, status;
 	size_t errlen;
 	regex_t *rexpr;
-	MSG_TEXT *t;
+	TYPE_LINE *l;
 
 	/* Form the prompt for getting the search expression */
 
@@ -1155,11 +1306,11 @@ FORM *forms;
 
 	/* Search the lines for a match */
 
-	for (t = type_lines; t != NULL; t = t->next) {
+	for (l = type_lines; l != NULL; l = l->next) {
 		/* Check if this is our line */
 
 		if ((back && line < type_point || !back && line > type_point)
-		    && !regexec(rexpr, t->line, 0, NULL, 0)) {
+		    && !regexec(rexpr, l->line, 0, NULL, 0)) {
 			/* This line matches the expression */
 
 			match = (back || !match) ? line : match;
@@ -1176,13 +1327,36 @@ FORM *forms;
 	return(match != 0 || !type_end);
 }
 /****************************************************************************/
+static int section(nsecs, back)
+int nsecs, back;
+{
+	/* Move to another section within typeout */
+
+	/* Check that we actually have sections to move to */
+
+	if (back && (!type_sectioned || type_showing
+		     || !type_have_prev_sec)) {
+		emsg("First section");
+		return(FALSE);
+	} else if (!back && (!type_sectioned || type_showing
+			     || !type_have_next_sec)) {
+		emsg("Last section");
+		return(FALSE);
+	}
+
+	/* Now set the number of sections to move */
+
+	type_sections = (back) ? -nsecs : nsecs;
+	return(TRUE);
+}
+/****************************************************************************/
 static void set_point(point, tenths)
 int point, tenths;
 {
 	/* Set up point and check if more lines are needed */
 
 	int row = 0, last;
-	MSG_TEXT *t;
+	TYPE_LINE *l;
 
 	/* We need to handle pending typeout tenths */
 
@@ -1225,8 +1399,8 @@ int point, tenths;
 
 		/* Print as many lines as we can */
 
-		for (t = find_line(type_point); t != NULL; t = t->next) {
-			type_show(t->line, row++);
+		for (l = find_line(type_point); l != NULL; l = l->next) {
+			type_show(l->line, row++);
 		}
 	} else {
 		/* We just need to display the lines */
@@ -1259,8 +1433,7 @@ regex_t *rexpr;
 		/* Update the point and mark */
 
 		msgl("Found ", (expr != NULL) ? expr : type_search, NULL);
-		type_mark = (repeat || type_repeat)
-			? type_mark : type_point;
+		type_mark = (repeat || type_repeat) ? type_mark : type_point;
 		set_point(point, 0);
 	} else {
 		emsgl("No match found for ", (expr != NULL)
@@ -1292,43 +1465,93 @@ static void trim_blank_lines()
 {
 	/* Trim trailing blank lines from the typeout list */
 
-	int line = 1, lastpos = 0;
-	MSG_TEXT *t, *last = NULL;
+	int blank_lines = 0;
+	TYPE_LINE *last = NULL;
 
 	/* Find the last non-blank line in the list */
 
-	for (t = type_lines; t != NULL; t = t->next) {
-		last = (!is_blank(t->line)) ? t : last;
-		lastpos = (!is_blank(t->line)) ? line : lastpos;
-		line++;
+	for (last = type_last; last != NULL &&
+	     is_blank(last->line); last = last->prev) {
+		/* Update the count of lines to delete */
+
+		blank_lines++;
 	}
 
-	/* Trim the lines if required */
+	/* Trim any blank lines we found */
 
-	if (last != NULL && last->next != NULL) {
-		free_mtext(last->next);
+	if (last != NULL) {
+		/* Remove the blank lines */
+
+		free_type_lines(last->next);
 		last->next = NULL;
-		type_count = lastpos;
+
+		/* And update the globals */
+
+		type_last = last;
+		type_count -= blank_lines;
 	}
 	return;
 }
 /****************************************************************************/
-static MSG_TEXT *find_line(pos)
+static TYPE_LINE *find_line(pos)
 int pos;
 {
 	/* Find the line at row pos in typeout */
 
-	int line = 1;
-	MSG_TEXT *t;
+	int line, back;
+	TYPE_LINE *first, *l;
+
+	/* Is it quicker to search backwards for pos? */
+
+	back = (pos > type_count / 2);
+
+	/* Initialise the loop position counters */
+
+	first = (back) ? type_last : type_lines;
+	line = (back) ? type_count : 1;
 
 	/* Loop over the lines until we reach pos */
 
-	for (t = type_lines; t != NULL && line < pos; t = t->next) {
-		line++;
+	for (l = first; l != NULL && line != pos;
+	     l = (back) ? l->prev : l->next) {
+		/* And update the line position */
+
+		line = (back) ? line - 1 : line + 1;
 	}
 
 	/* And return the node from the list */
 
-	return(t);
+	return(l);
+}
+/****************************************************************************/
+static void free_type_lines(list)
+TYPE_LINE *list;
+{
+	/* Free the list of lines we're typing out */
+
+	TYPE_LINE *next;
+
+	/* Don't use recursion here for speed */
+
+	while (list != NULL) {
+		/* Save the next line */
+
+		next = list->next;
+	
+		/* Free this line */
+
+		if (list->line != NULL) {
+			free(list->line);
+		}
+		free(list);
+
+		/* And move on */
+
+		list = next;
+	}
+
+	/* That's all folks */
+
+	return;
 }
 /****************************************************************************/
