@@ -26,13 +26,14 @@
 #include "tags.h"
 #include "complete.h"
 #include "varlist.h"
+#include "mime.h"
 #include STRING_HDR
 
 /****************************************************************************/
 /* RCS info */
 
 #ifndef lint
-static char *RcsId = "$Id: variable.c,v 2.0 1997/09/06 12:00:58 malc Rel $";
+static char *RcsId = "$Id: variable.c,v 2.1 1997/09/07 00:38:47 malc Exp $";
 static char *VariableId = VARIABLEID;
 #endif /* ! lint */
 
@@ -44,19 +45,22 @@ extern char *getenv(), *get_dstr(), *get_estr();
 extern char *get_dcstr(), *get_ecstr(), *get_host();
 extern char *get_home(), *get_realname(), *expand();
 extern char *alias(), *canonical(), *realname();
-extern char *a_strerror(), *strkey(), *strform();
-extern char *formtext(), *strcanon(), *strtags();
-extern char *unmessage(), *utos();
+extern char *charset(), *encoding(), *strkey();
+extern char *a_strerror(), *strcanon(), *utos();
+extern char *strform(), *formtext(), *strtags();
+extern char *unmessage();
 extern int atoi(), strcasecmp(), strncasecmp();
-extern int get_key(), set_length(), retextual();
+extern int get_key(), set_length();
 extern unsigned alarm();
 extern void free(), msg(), cmsg(), emsg();
 extern void emsgl(), typeout(), table();
-extern void display(), free_tlist();
+extern void alldisplay(), free_tlist();
+extern void display_iso_chars(), remime();
 extern FORM *atomstr(), *list(), *add_form();
 extern TAG_LIST *taglist();
 extern CLIST *fn_complete(), *sort_complete();
-extern CLIST *add_clist();
+extern CLIST *charset_complete(), *add_clist();
+extern CLIST *encoding_complete();
 
 /* Local function declarations */
 
@@ -310,6 +314,26 @@ char *prompt;
 	return(get_ecstr(NULL, prompt, var->text, sort_complete, C_STRICT));
 }
 /****************************************************************************/
+static char *vg_charset(var, prompt)
+VARIABLE *var;
+char *prompt;
+{
+	/* Return a value suitable for setting a charset variable */
+
+	return(get_ecstr(NULL, prompt, var->text,
+			 charset_complete, C_PERMISSIVE));
+}
+/****************************************************************************/
+static char *vg_encoding(var, prompt)
+VARIABLE *var;
+char *prompt;
+{
+	/* Return a value suitable for setting an encoding variable */
+
+	return(get_ecstr(NULL, prompt, var->text,
+			 encoding_complete, C_STRICT));
+}
+/****************************************************************************/
 /*ARGSUSED*/
 static char *vg_char(var, prompt)
 VARIABLE *var;
@@ -375,7 +399,7 @@ FORM *form;
 {
 	/* Check a string form and convert it to a string */
 
-	char *nullstring = "";
+	static char *nullstring = "";
 	int key;
 
 	/* The form must be a string or nil */
@@ -747,6 +771,66 @@ char *text;
 	return(TRUE);
 }
 /****************************************************************************/
+static int vs_charset(var, text)
+VARIABLE *var;
+char *text;
+{
+	/* Set the variable to the charset in text */
+
+	char *cset;
+
+	/* Can't unset the variable */
+
+	if (text == NULL) {
+		emsgl("Can't unset ", var->name, NULL);
+		return(FALSE);
+	}
+
+	/* Canonicalise the charset */
+
+	if ((cset = charset(text)) == NULL) {
+		emsg(a_strerror());
+		return(FALSE);
+	}
+
+	/* Set the variable as a string */
+
+	(void) vs_string(var, cset);
+	free(cset);
+
+	return(TRUE);
+}
+/****************************************************************************/
+static int vs_encoding(var, text)
+VARIABLE *var;
+char *text;
+{
+	/* Set the variable to the encoding in text */
+
+	char *enc;
+
+	/* Can't unset the variable */
+
+	if (text == NULL) {
+		emsgl("Can't unset ", var->name, NULL);
+		return(FALSE);
+	}
+
+	/* Canonicalise the encoding */
+
+	if ((enc = encoding(text)) == NULL) {
+		emsg(a_strerror());
+		return(FALSE);
+	}
+
+	/* Set the variable as a string */
+
+	(void) vs_string(var, enc);
+	free(enc);
+
+	return(TRUE);
+}
+/****************************************************************************/
 static int vs_char(var, text)
 VARIABLE *var;
 char *text;
@@ -919,10 +1003,10 @@ char *text;
 
 	(void) vs_string(var, text);
 
-	/* And redisplay any active window */
+	/* And redisplay any active windows */
 
 	if (cwin != NULL) {
-		display(cwin);
+		alldisplay(cwin->buf);
 	}
 	return(TRUE);
 }
@@ -933,8 +1017,6 @@ char *text;
 {
 	/* Set the header-line-format variable var to text */
 
-	WINDOW *w;
-
 	/* Can't unset the text */
 
 	if (text == NULL) {
@@ -954,11 +1036,8 @@ char *text;
 
 	/* And update any active windows */
 
-	if ((w = cwin) != NULL) {
-		do {
-			display(w);
-			w = w->next;
-		} while (w != cwin);
+	if (cwin != NULL) {
+		alldisplay(NULL);
 	}
 
 	return(TRUE);
@@ -970,8 +1049,6 @@ char *text;
 {
 	/* Set the mode-line-format variable var to text */
 
-	WINDOW *w;
-
 	/* Can't unset the text */
 
 	if (text == NULL) {
@@ -991,11 +1068,8 @@ char *text;
 
 	/* And update any active windows */
 
-	if ((w = cwin) != NULL) {
-		do {
-			display(w);
-			w = w->next;
-		} while (w != cwin);
+	if (cwin != NULL) {
+		alldisplay(NULL);
 	}
 
 	return(TRUE);
@@ -1007,8 +1081,6 @@ char *text;
 {
 	/* Update the show-dates-in-local-times variable */
 
-	WINDOW *w;
-
 	/* Update the variable as a boolean */
 
 	if (!vs_boolean(var, text)) {
@@ -1019,11 +1091,8 @@ char *text;
 
 	/* And update any active windows */
 
-	if ((w = cwin) != NULL) {
-		do {
-			display(w);
-			w = w->next;
-		} while (w != cwin);
+	if (cwin != NULL) {
+		alldisplay(NULL);
 	}
 
 	return(TRUE);
@@ -1035,21 +1104,44 @@ char *text;
 {
 	/* Set the viewable-charsets variable var to text */
 
-	WINDOW *w;
+	char *cset, *end;
+	int iso = FALSE;
+	size_t len;
 
 	/* Update the variable as a string */
 
 	(void) vs_string(var, text);
 
-	/* And reset the MIME flag on messages */
+	/* Loop through checking charset against viewed */
 
-	if ((w = cwin) != NULL && retextual(cwin->buf)) {
-		/* Redisplay all the windows */
+	cset = text;
+	while (!iso && cset != NULL && *cset != '\0') {
+		/* Find the end of the viewable charset */
 
-		do {
-			display(w);
-			w = w->next;
-		} while (w != cwin);
+		end = strchr(cset, ':');
+
+		/* How long is the current charset? */
+
+		len = (end != NULL) ? end - cset : strlen(cset);
+
+		/* Does this charset match us-ascii? */
+
+		iso = (strncasecmp(cset, US_ASCII, len) != 0);
+
+		/* Update the loop counter */
+
+		cset = (end != NULL) ? end + 1 : NULL;
+	}
+
+	/* Now set whether the display can handle iso characters */
+
+	display_iso_chars(iso);
+
+	/* Update any messages' MIME flag and redraw windows */
+
+	if (cwin != NULL) {
+		remime(cwin->buf);
+		alldisplay(NULL);
 	}
 
 	return(TRUE);
@@ -1255,6 +1347,8 @@ int headings;
 
 	VARIABLE *v;
 
+	/* Loop over the available variables */
+
 	for (v = variables; !user_quit && v->name != NULL; v++) {
 		/* Should this variable be listed? */
 
